theory AbstrState = JVM:

text{*We define an abstract type of equivalence class
identifiers.*}

typedecl Colour

consts freshID:: "Colour set \<Rightarrow> Colour"
axioms fresh_ID_fresh:"freshID X \<notin> X"

lemma freshfresh:"\<lbrakk>freshID X : Y; Y\<subseteq> X\<rbrakk> \<Longrightarrow> ANY"
by (insert fresh_ID_fresh, fast)

datatype TP = intTP | classTP Class

types VarContext = "(Var,Colour) AssList"

types OBJECT = "Class \<times> ((Field,Colour) AssList)"

types HP = "(Colour,OBJECT)AssList"

types STATE = "Colour list \<times> VarContext \<times> HP"

constdefs SD_sum::"STATE \<Rightarrow> HP \<Rightarrow> STATE \<Rightarrow> bool"
"SD_sum \<Sigma> H \<Pi> \<equiv> 
    (case \<Sigma> of (S,\<Gamma>,H0) \<Rightarrow>
    (\<exists> HH . \<Pi> = (S,\<Gamma>,HH) \<and> AL_sum H0 H HH \<and> AL_dom H0 \<inter> AL_dom H = {}))"

lemma SD_sum_emp1:"SD_sum \<Sigma> emp \<Sigma>"
(*<*)
apply (simp add: SD_sum_def)
apply (case_tac \<Sigma>, clarsimp)
apply (rule, rule AL_sum_emp2)
apply (simp add: AL_dom_emp)
done
(*>*)

constdefs AbsFieldsAt::"HP \<Rightarrow> Colour \<Rightarrow> Field \<Rightarrow> Colour set"
"AbsFieldsAt H d f \<equiv> {c. \<exists> C F. H\<down>d = Some (C, F) \<and> F\<down>f = Some c}"

constdefs AbsFieldsOf::"HP \<Rightarrow> Colour \<Rightarrow> Colour set"
"AbsFieldsOf H d \<equiv> {c. \<exists> f . c : AbsFieldsAt H d f}" 

constdefs AbsFields::"HP \<Rightarrow> Colour set"
"AbsFields H \<equiv> {c. \<exists> d . c : AbsFieldsOf H d}"

lemma AbsFields_expand: "AbsFields H = { c . \<exists> d C F f. H\<down>d = Some(C,F) \<and> F\<down>f = Some c}"
apply (simp add: AbsFields_def AbsFieldsOf_def AbsFieldsAt_def) apply auto done

lemma AbsFieldsD: "c : AbsFields H \<Longrightarrow> (\<exists> d C F f. H\<down>d = Some(C,F) \<and> F\<down>f = Some c)"
by (simp add: AbsFields_expand)

lemma AbsFieldsI:"\<lbrakk>H\<down>c = Some (D, G); G\<down>f = Some d\<rbrakk> \<Longrightarrow> d \<in> AbsFields H"
by (simp add: AbsFields_expand, fast)

lemma AbsFields_extensional: "\<forall> x c. (H\<down>x = Some c) = (HH\<down>x = Some c) \<Longrightarrow> AbsFields H = AbsFields HH"
by (simp add: AbsFields_def AbsFieldsOf_def AbsFieldsAt_def)

lemma AbsFields_insert_freshLoc: "cl \<notin> AL_dom H \<Longrightarrow> AbsFields H[cl\<mapsto>(C, emp)] = AbsFields H"
apply (simp add: AbsFields_def AbsFieldsOf_def AbsFieldsAt_def)
apply (rule, clarsimp)
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) apply(simp add: AL_emp1)
  apply clarsimp apply fast
apply clarsimp
  apply (case_tac "cl=d", clarsimp) apply (drule AL_domI) apply simp
  apply (rule, rule, rule, rule, rule)
    apply (erule AL_update5) apply assumption
    apply assumption
done

lemma AbsFields_emp: "AbsFields emp = {}"
by (simp add:AbsFields_expand AL_emp1)

lemma AbsFields_delete: "AbsFields (delete L c) \<subseteq> AbsFields L"
apply (simp add: AbsFields_expand, clarsimp)
apply (drule AL_delete6, clarsimp) 
apply fast
done

lemma AbsFields_sum: "AL_sum H K HK \<Longrightarrow> AbsFields HK \<subseteq> AbsFields H \<union> AbsFields K"
apply (simp add: AbsFields_expand, clarsimp)
apply (drule AL_sum_split, assumption)
apply (erule disjE)
  apply (rule_tac x=d in exI, clarsimp) apply (rule, assumption)
apply clarsimp
done

lemma AL_sum_AbsFields: "AL_sum H K HK \<Longrightarrow> AbsFields HK = AbsFields H \<union> AbsFields K"
apply rule apply (erule AbsFields_sum)
apply safe
  apply (drule AbsFieldsD, clarsimp) apply (rule AbsFieldsI) apply (erule AL_sum3, assumption) apply assumption
  apply (drule AbsFieldsD, clarsimp) apply (rule AbsFieldsI) apply (erule AL_sum4, assumption) apply assumption
done

lemma AL_sum_absFields: "\<lbrakk>AL_sum H Ha K; c \<in> AbsFields H\<rbrakk> \<Longrightarrow> c : AbsFields K"
apply (simp add: AbsFields_expand, clarsimp)
apply (drule AL_sum3, assumption)
apply (rule_tac x=d in exI, clarsimp)
apply (rule, assumption)
done

constdefs HP_good::"HP \<Rightarrow> Colour set \<Rightarrow> bool"
"HP_good H X \<equiv> \<forall> c C F . H\<down>c = Some(C,F) \<longrightarrow> (F:AL \<and> AL_cod F \<subseteq> X)"

lemma HP_goodI:"\<forall> c C F . H\<down>c = Some(C,F) \<longrightarrow> (F:AL \<and> AL_cod F \<subseteq> X) \<Longrightarrow> HP_good H X"
by (simp add: HP_good_def)

lemma HP_good_emp:"HP_good emp X"
by (simp add: HP_good_def emp_def)

lemma HP_goodD: "\<lbrakk>HP_good H X; H\<down>c = Some(C,F)\<rbrakk> \<Longrightarrow> F:AL \<and> AL_cod F \<subseteq> X"
by (simp add: HP_good_def)

lemma HP_good_subset:"\<lbrakk>HP_good H X; contained K H; X\<subseteq>Y\<rbrakk> \<Longrightarrow> HP_good K Y"
apply (simp add: HP_good_def, clarsimp)
apply (drule contained3) apply assumption apply fast 
done

lemma HP_good_AbsFields: "HP_good H X \<Longrightarrow> AbsFields H \<subseteq> X"
apply (simp add: HP_good_def AbsFields_def AbsFieldsOf_def AbsFieldsAt_def)
apply clarsimp
apply (erule_tac x=d in allE, clarsimp)
apply (drule_tac L=F in AL_codI) apply fast
done

lemma HP_good_MP_contained:
"\<lbrakk>HP_good H (AL_dom N); contained N M\<rbrakk> \<Longrightarrow> HP_good H (AL_dom M)"
apply (simp add: HP_good_def, clarsimp)
apply (erule_tac x=c in allE, clarsimp)
  apply (drule SetTriv) apply assumption
  apply (erule contained6) apply assumption
done

lemma HP_good_MP_extensional:
"\<lbrakk>HP_good H (AL_dom N); \<forall>x a . (N\<down>x = Some a) = (NN\<down>x = Some a)\<rbrakk> \<Longrightarrow> HP_good H (AL_dom NN)"
apply (simp add: HP_good_def) apply clarsimp
apply (erule_tac x=c in allE, clarsimp)
apply (drule SetTriv, assumption)
apply (drule AL_domD, clarsimp)
apply (erule AL_domI)
done

lemma HP_good_insert:
"\<lbrakk>HP_good H (AL_dom N); cl \<notin> AL_dom N; AL_dom H \<subseteq> AL_dom N\<rbrakk> \<Longrightarrow> HP_good H[cl\<mapsto>(C, emp)] (AL_dom N[cl\<mapsto>(classTP C)])"
apply (simp add: HP_good_def, clarsimp)
apply (drule AL_upd_split)
apply (erule disjE, clarsimp)
  apply (rule, rule AL_EMP)
  apply (simp add: AL_cod_emp)
apply (erule_tac x=c in allE, clarsimp)
  apply (drule SetTriv) apply assumption
  apply (simp add: AL_dom_upd1)
done

constdefs DOM::"STATE \<Rightarrow> Colour set"
"DOM \<Sigma> \<equiv> (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow> (set S) \<union> (AL_cod \<Gamma>) \<union> (AL_dom H) \<union> AbsFields H)"

lemma DOM_1:"DOM ([c], emp, H) \<subseteq> DOM (c # S, \<Gamma>, H)" 
by (simp add: DOM_def AL_cod_emp, fast)

lemma DOM_emptyOpStack_subset: "DOM ([], \<Gamma>, H) \<subseteq> DOM (S, \<Gamma>, H)"
apply (simp add: DOM_def, fast) 
done

lemma DOM_split_equals:"DOM (c # S, \<Gamma>, H) = DOM ([c], emp, H) \<union> DOM (S, \<Gamma>, emp)"
by (simp add: DOM_def AL_dom_emp AL_cod_emp AbsFields_emp, fast)

lemma DOM_cons: "DOM (S, \<Gamma>, H) \<subseteq> DOM (c # S, \<Gamma>, H)"
by (simp add: DOM_def, fast)


lemma DOM_HP_extensional:"\<forall> a obj. (H\<down>a = Some obj) = (HH\<down>a = Some obj) \<Longrightarrow> DOM (S, \<Gamma>, H) = DOM (S, \<Gamma>, HH)"
apply (subgoal_tac " AL_dom HH \<subseteq> AL_dom H")
prefer 2 apply (simp add: AL_dom_def)
apply (subgoal_tac " AL_dom H \<subseteq> AL_dom HH")
prefer 2 apply (simp add: AL_dom_def)
apply (frule AbsFields_extensional)
apply (simp add: DOM_def)
done

(**********************************Administrative maps******************************)
types MP = "(Colour,TP) AssList"

constdefs ObjIDs::"MP \<Rightarrow> Colour set"
"ObjIDs N \<equiv> {c. \<exists>C . N\<down>c = Some (classTP C)}"

lemma ObjIDs_in_Dom: "ObjIDs N \<subseteq> AL_dom N" apply (simp add: ObjIDs_def AL_dom_def) apply fast done

lemma AL_sum_split1_ObjID: "\<lbrakk>AL_sum N L M;  x \<in> ObjIDs M; x \<in> AL_dom N\<rbrakk> \<Longrightarrow> x: ObjIDs N"
apply (simp add: ObjIDs_def, clarsimp)
apply (drule AL_domD, clarsimp)
apply (drule AL_sum3, assumption)
apply clarsimp
done

lemma contained_ObjID_Inv: "\<lbrakk>contained N M;  x \<in> ObjIDs M; x \<in> AL_dom N\<rbrakk> \<Longrightarrow> x: ObjIDs N"
apply (simp add: ObjIDs_def, clarsimp)
apply (drule AL_domD, clarsimp)
apply (drule contained3, assumption)
apply clarsimp
done

lemma MP_sum_emp_ObjIDs: "AL_sum N emp M \<Longrightarrow> ObjIDs M = ObjIDs N"
apply (simp add: ObjIDs_def)
apply (rule, clarsimp)
  apply (drule AL_sum_split, assumption)
  apply (erule disjE, clarsimp)
  apply (simp add: AL_emp1)
apply clarsimp
  apply (drule AL_sum3, assumption)
  apply clarsimp
done

lemma ObjIDs_insertObj: "ObjIDs N[c\<mapsto>(classTP C)] = ObjIDs N \<union> {c}"
apply (simp add: ObjIDs_def)
apply (rule, clarsimp)
  apply (drule_tac a=c in AL_update3a, assumption) apply clarsimp
apply clarsimp apply (simp add: AL_update1)
  apply clarsimp apply (case_tac "x=c", simp add: AL_update1)
                 apply (rule, erule AL_update5a) apply assumption
done

lemma ObjIDs_extensional: "\<forall>x a . (N\<down>x = Some a) = (NN\<down>x = Some a) \<Longrightarrow>  ObjIDs N =  ObjIDs NN"
by (simp add: ObjIDs_def)

lemma ObjIDs_insertInt: "ObjIDs NN[c\<mapsto>intTP] \<subseteq> ObjIDs NN"
apply (simp add: ObjIDs_def, clarsimp)
  apply (drule AL_upd_split, clarsimp)
done

lemma contained_ObjIds_subset: "contained N M \<Longrightarrow> ObjIDs N \<subseteq> ObjIDs M"
apply (simp add: ObjIDs_def, clarsimp)
apply (drule contained3, assumption)
apply fast
done

lemma AL_sum_ObjIDs_left: "AL_sum N M L \<Longrightarrow> ObjIDs N \<subseteq> ObjIDs L"
apply (simp add: AL_sum_def, (erule conjE)+)
apply (erule contained_ObjIds_subset)
done

constdefs State_good::"STATE \<Rightarrow> MP \<Rightarrow> bool"
"State_good \<Sigma> N \<equiv> 
   case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow> \<Gamma>:AL \<and> H:AL \<and> HP_good H (AL_dom N)"

lemma Staate_good_Tail: "State_good (c # S, \<Gamma>, H) N \<Longrightarrow> State_good (S, \<Gamma>, H) N"
by (simp add: State_good_def)

lemma State_good_delete: "State_good (S, \<Gamma>, L) N \<Longrightarrow> State_good (S, \<Gamma>, delete L c) N"
apply (simp add: State_good_def, clarsimp)
apply (rule, erule AL_DEL)
apply (erule HP_good_subset) apply (rule contained_delete) apply simp
done

lemma State_good_insert:
  "\<lbrakk>State_good (S, \<Gamma>, H) N; cl \<notin> AL_dom N; AL_dom H \<subseteq> AL_dom N\<rbrakk> 
  \<Longrightarrow> State_good (cl # S, \<Gamma>, H[cl\<mapsto>(C, emp)]) N[cl\<mapsto>(classTP C)]"
apply (simp add: State_good_def, clarsimp)
apply (rule, erule_tac M=H and x=cl in AL_CONS) apply fast apply fastsimp
  apply (erule HP_good_insert) apply assumption+
done

lemma State_good_MP_contained:
  "\<lbrakk>State_good (T, \<Delta>, K) N; contained N M\<rbrakk> \<Longrightarrow> State_good (T, \<Delta>, K) M"
apply (simp add: State_good_def, clarsimp)
apply (erule HP_good_MP_contained) apply assumption
done

lemma State_good_MP_extensional: 
"\<lbrakk>State_good (S, \<Gamma>, H) N; \<forall> x c. (N\<down>x = Some c) = (NN\<down>x = Some c)\<rbrakk> \<Longrightarrow> State_good (S, \<Gamma>, H) NN"
apply (simp add: State_good_def, clarsimp)
apply (erule HP_good_MP_extensional, assumption)
done

constdefs State_welldef::"STATE \<Rightarrow> MP \<Rightarrow> bool"
"State_welldef \<Sigma> N \<equiv>
     (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow>
        (set S) \<union> (AL_cod \<Gamma>) \<union> (AL_dom H) \<union> AbsFields H \<subseteq> AL_dom N \<and>
     ((set S) \<union> (AL_cod \<Gamma>) \<union> AbsFields H) \<inter> ObjIDs N \<subseteq> AL_dom H \<and>
     (AL_dom H) \<subseteq> ObjIDs N \<and>
     State_good \<Sigma> N \<and> N:AL)"

lemma State_welldef_contract:
"State_welldef \<Sigma> N =
     (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow>
        DOM \<Sigma> \<subseteq> AL_dom N \<and>
     ((set S) \<union> (AL_cod \<Gamma>) \<union> AbsFields H) \<inter> ObjIDs N \<subseteq> AL_dom H \<and>
     (AL_dom H) \<subseteq> ObjIDs N \<and>
     State_good \<Sigma> N \<and> N:AL)"
apply (case_tac \<Sigma>, clarsimp)
apply (simp add: State_welldef_def DOM_def)
done

lemma State_welldef_D_AL:"State_welldef \<Sigma> N \<Longrightarrow> N:AL"
apply (simp add: State_welldef_def)
apply (case_tac \<Sigma>, clarsimp)
done

lemma State_welldef_HPextensional:"\<lbrakk>State_welldef (S,\<Gamma>,H) N; \<forall> x c. (H\<down>x = Some c) = (HH\<down>x = Some c); HH:AL\<rbrakk>
    \<Longrightarrow> State_welldef (S,\<Gamma>,HH) N"
apply (subgoal_tac " AL_dom HH \<subseteq> AL_dom H")
prefer 2 apply (simp add: AL_dom_def)
apply (subgoal_tac " AL_dom H \<subseteq> AL_dom HH")
prefer 2 apply (simp add: AL_dom_def)
apply (frule AbsFields_extensional)
apply (simp add: State_welldef_def State_good_def HP_good_def)
done

lemma State_welldef_D_good_AL: "State_welldef \<Sigma> N \<Longrightarrow> State_good \<Sigma> N \<and>N:AL"
apply (simp add: State_welldef_def) 
apply (case_tac \<Sigma>, clarsimp)
done

lemma State_welldef_D_Heap_ObjIDs: "State_welldef (S,\<Gamma>,H) N \<Longrightarrow> AL_dom H \<subseteq> ObjIDs N" 
by (simp add: State_welldef_def)

lemma State_welldef_fresh:
  "\<lbrakk> State_welldef (S, \<Gamma>, H) N;
     c \<notin> AL_dom N\<rbrakk>
  \<Longrightarrow> State_welldef (S, \<Gamma>, H) N[c\<mapsto>tp]"
apply (simp add: State_welldef_def, clarsimp)
apply (simp add: AL_dom_upd1)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply rule apply (rule SetTriv101) prefer 2 apply assumption
           apply (rule, clarsimp) apply (simp add: ObjIDs_def, clarsimp)
             apply (drule AL_upd_split)
             apply (rotate_tac -1, erule disjE, clarsimp)
               apply (subgoal_tac "c:AL_dom N", simp) apply (erule thin_rl) apply (erule disjE, erule SetTriv, assumption)
                 apply (erule disjE, erule SetTriv, assumption) apply (erule SetTriv, assumption) 
           apply clarsimp 
apply (rule, erule_tac Y="ObjIDs N" in SetTriv101) apply (simp add: ObjIDs_def, clarsimp) 
         apply (rule, rule AL_update5, assumption) apply (drule AL_domI) apply fast 
apply rule apply (simp add: State_good_def, clarsimp)
           apply (erule HP_good_subset) apply (rule contained_refl) apply (rule AL_dom_upd)
apply (erule AL_UPD)
done

lemma State_welldef_ConsCons:
  "\<lbrakk> State_welldef (S, \<Gamma>, H) NN;
     AL_sum N Na NN; c \<notin> AL_dom N; c \<notin> AL_dom Na\<rbrakk>
  \<Longrightarrow> State_welldef (c # S, \<Gamma>, H) NN[c\<mapsto>intTP]"
apply (simp add: State_welldef_def, clarsimp)
apply (simp add: AL_dom_upd1)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply (rule, clarsimp)
   apply (case_tac "x=c", clarsimp) apply (simp add: ObjIDs_def AL_update1) 
   apply clarsimp
   apply (subgoal_tac "x \<in> ObjIDs NN", fast)
    apply (simp add: ObjIDs_def, clarsimp) apply (drule AL_update3a) apply assumption apply clarsimp
apply (rule, clarsimp)
  apply (case_tac "x=c", clarsimp) 
    apply (drule_tac x=c in SetTriv, assumption) apply (simp add: AL_sum_def)
  apply (drule_tac x=x and Y="ObjIDs NN" in SetTriv, assumption) 
    apply (simp add: ObjIDs_def) apply clarsimp apply (rule, erule AL_update5a) apply assumption
apply rule apply (simp add: State_good_def, clarsimp)
           apply (erule HP_good_subset) apply (rule contained_refl) apply (rule AL_dom_upd)
apply (erule AL_UPD)
done

lemma State_welldef_ConsCons2:
  "\<lbrakk> State_welldef (S, \<Gamma>, H) NN;
     AL_sum N Na NN; c \<notin> AL_dom N; c \<notin> AL_dom Na\<rbrakk>
  \<Longrightarrow> State_welldef (S, \<Gamma>, H) NN[c\<mapsto>intTP]"
apply (simp add: State_welldef_def, clarsimp)
apply (simp add: AL_dom_upd1)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply (rule, fast)
apply (rule, clarsimp)
   apply (case_tac "x=c", clarsimp) apply (simp add: ObjIDs_def AL_update1) 
   apply (subgoal_tac "x \<in> ObjIDs NN", fast)
    apply (simp add: ObjIDs_def, clarsimp) apply (drule AL_update3a) apply assumption apply clarsimp
apply (rule, clarsimp)
  apply (case_tac "x=c", clarsimp) 
    apply (drule_tac x=c in SetTriv, assumption) apply (simp add: AL_sum_def)
  apply (drule_tac x=x and Y="ObjIDs NN" in SetTriv, assumption) 
    apply (simp add: ObjIDs_def) apply clarsimp apply (rule, erule AL_update5a) apply assumption
apply rule apply (simp add: State_good_def, clarsimp)
           apply (erule HP_good_subset) apply (rule contained_refl) apply (rule AL_dom_upd)
apply (erule AL_UPD)
done

lemma State_welldef_Store_delete: "State_welldef (T, M, K) N \<Longrightarrow> State_welldef (T, delete M x , K) N"
apply (simp add: State_welldef_def, clarsimp)
apply (subgoal_tac "AL_cod (delete M x) \<subseteq> AL_cod M") prefer 2 apply (rule AL_cod_delete)
apply rule apply (erule SetTriv101) apply assumption
apply rule apply (rule SetTriv101) prefer 2 apply assumption apply fast
apply (simp add: State_good_def, erule conjE) apply (erule AL_DEL)
done

lemma State_welldef_HPextensional:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; \<forall> x c . (H\<down>x = Some c) = (K\<down>x = Some c); K:AL\<rbrakk> \<Longrightarrow> State_welldef (S, \<Gamma>, K) N"
apply (frule AbsFields_extensional)
apply (simp add: State_welldef_def AL_dom_def, clarsimp)
apply (simp add: State_good_def)
apply (simp add: HP_good_def)
done

lemma State_welldef_MPextensional:
 "\<lbrakk>State_welldef \<Sigma> N; \<forall> x c. (N\<down>x = Some c) = (NN\<down>x = Some c); NN \<in> AL\<rbrakk>
    \<Longrightarrow> State_welldef \<Sigma> NN"
apply (subgoal_tac " AL_dom N \<subseteq> AL_dom NN")
prefer 2 apply (simp add: AL_dom_def)
apply (simp add: State_welldef_def)
apply (case_tac \<Sigma>, clarsimp)
apply (rule, erule SetTriv101, assumption) 
apply (rule, erule SetTriv101, assumption) 
apply (rule, erule SetTriv101, assumption) 
apply (rule, erule SetTriv101, assumption) 
apply (frule ObjIDs_extensional, clarsimp)
apply (erule State_good_MP_extensional, clarsimp)
done

(*generalises Invoke_State_welldef_preserved1*)
lemma State_welldef_merge:
  "\<lbrakk> State_welldef (S, \<Gamma>, H) N; State_welldef (T, \<Delta>, K) NN; 
        contained N NN; AL_dom H \<subseteq> AL_dom K\<rbrakk>
   \<Longrightarrow> State_welldef (S, \<Gamma>, K) NN"
apply (simp add: State_welldef_def, clarsimp) 
apply (rule, erule SetTriv101)  apply (erule contained5)
apply (rule, erule SetTriv101) apply (erule contained5)
apply (rule, clarsimp)
  apply (erule disjE, frule_tac x=x in SetTriv, assumption)
    apply (subgoal_tac "x:ObjIDs N")
      apply (rule SetTriv) prefer 2 apply assumption
      apply (rule_tac X="(set S \<union> AL_cod \<Gamma> \<union> AbsFields H) \<inter> ObjIDs N" in SetTriv) apply fastsimp apply assumption
      apply (simp add: ObjIDs_def, clarsimp) apply (drule contained2, assumption) apply assumption apply clarsimp
  apply (erule disjE, frule_tac x=x in SetTriv, assumption)
    apply (subgoal_tac "x:ObjIDs N")
      apply (rule SetTriv) prefer 2 apply assumption
      apply (rule_tac X="(set S \<union> AL_cod \<Gamma> \<union> AbsFields H) \<inter> ObjIDs N" in SetTriv) apply fastsimp apply assumption
      apply (simp add: ObjIDs_def, clarsimp) apply (drule contained2, assumption) apply assumption apply clarsimp
  apply (rule_tac X="(set T \<union> AL_cod \<Delta> \<union> AbsFields K) \<inter> ObjIDs NN" in SetTriv) apply fast apply assumption
apply (simp add: State_good_def)
done

lemma State_welldef_RET: "State_welldef (c # S, \<Gamma>, H) N \<Longrightarrow> State_welldef ([c], emp, H) N"
apply (simp add: State_welldef_def, clarsimp)
apply (simp add: add: AL_cod_emp) 
apply rule apply (rule SetTriv101) prefer 2 apply assumption apply fast
apply (simp add: State_good_def)
apply (rule AL_EMP)
done

lemma State_welldef_Tail: "State_welldef (c # S, \<Gamma>, H) N \<Longrightarrow> State_welldef (S, \<Gamma>, H) N"
apply(simp add: State_welldef_def, clarsimp)
apply rule apply fast
apply (erule Staate_good_Tail)
done
lemma State_welldef_insertL:
  "\<lbrakk> State_welldef (S, \<Gamma>, H) N; cl \<notin> AL_dom N\<rbrakk>
  \<Longrightarrow> State_welldef (cl # S, \<Gamma>, H[cl\<mapsto>(C, emp)]) (N[cl\<mapsto>(classTP C)])"
apply (simp add: State_welldef_def, clarify)
apply (rule, rule AL_dom_upd2)
apply (rule, erule AL_dom_upd3) 
apply (rule, erule AL_dom_upd3) 
apply (rule, erule AL_dom_upd4) 
apply (subgoal_tac "cl \<notin> AL_dom H") prefer 2 apply fast
apply (drule_tac C=C in AbsFields_insert_freshLoc) apply clarsimp
apply (rule, erule SetTriv101) apply (rule AL_dom_upd)
apply (simp add: ObjIDs_insertObj)
apply (rule, rule AL_dom_upd2)
apply rule apply (erule SetTriv101) apply (rule AL_dom_upd)
apply (simp add: AL_dom_upd1) 
apply (rule, fast)
apply (rule, erule State_good_insert, assumption, assumption)
apply (erule_tac M=N in AL_CONS) apply assumption apply simp
done

lemma State_welldef_MP_contained:
  "\<lbrakk>State_welldef (T, \<Delta>, K) N; contained N M; M:AL\<rbrakk> \<Longrightarrow> State_welldef (T, \<Delta>, K) M"
apply (simp add: State_welldef_def, clarsimp)
apply (rule, erule SetTriv101) apply (erule contained5)
apply (rule, erule SetTriv101) apply (erule contained5)
apply (rule, erule SetTriv101) apply (erule contained5)
apply (rule, erule SetTriv101) apply (erule contained5)
apply rule
  apply (simp add: ObjIDs_def)
  apply rule
  apply (subgoal_tac "x : (set T \<union> AL_cod \<Delta> \<union> AbsFields K) \<and> x: {c. \<exists>C . M\<down>c = Some (classTP C)}")
  prefer 2 apply clarsimp
  apply (rotate_tac -2, erule thin_rl)
  apply (erule conjE)
  apply (subgoal_tac "x:AL_dom N") prefer 2 apply clarsimp
                                            apply (erule disjE, erule SetTriv, assumption) 
                                            apply (erule disjE, erule SetTriv, assumption) 
                                            apply (erule SetTriv, assumption) 
  apply clarsimp
  apply (drule contained7, assumption) apply assumption
  apply (rule_tac X="(set T \<union> AL_cod \<Delta> \<union> AbsFields K) \<inter> {c. \<exists>C . N\<down>c = Some (classTP C)}" in SetTriv)
    apply clarsimp
  apply assumption
 apply (rule, erule_tac Y="ObjIDs N" in SetTriv101) 
           apply (erule contained_ObjIds_subset)
 apply (erule State_good_MP_contained) apply assumption
done

lemma State_welldef_OPS_lookup_dom:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; n < length S\<rbrakk> \<Longrightarrow> (S ! n) : AL_dom N"
apply (simp add: State_welldef_def, clarsimp)
apply (subgoal_tac "S!n : set S", fast)
apply simp
done

lemma State_welldef_OPS_lookup:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; n < length S\<rbrakk> \<Longrightarrow> \<exists> tp . N\<down>(S ! n) = Some tp"
apply (drule State_welldef_OPS_lookup_dom, assumption)
apply (drule AL_domD, clarsimp)
done

lemma State_welldef_Store_lookup_dom:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; \<Gamma>\<down>x = Some c\<rbrakk> \<Longrightarrow> c:AL_dom N"
apply (simp add: State_welldef_def, clarsimp)
apply (drule AL_codI) apply (erule SetTriv, assumption)
done

lemma State_welldef_Store_lookup:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; \<Gamma>\<down>x = Some c\<rbrakk> \<Longrightarrow> \<exists> tp . N\<down>c = Some tp"
apply (drule State_welldef_Store_lookup_dom, assumption)
apply (drule AL_domD, clarsimp)
done

lemma State_welldef_Heap_lookup_dom:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; H\<down>c = Some (D, G)\<rbrakk>
       \<Longrightarrow> c : AL_dom N"
apply (simp add: State_welldef_def, clarsimp)
apply (rule SetTriv, erule AL_domI, assumption)
done

lemma State_welldef_Heap_lookup:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; H\<down>c = Some (D, G)\<rbrakk>
       \<Longrightarrow> \<exists> tp . N\<down>c = Some tp"
apply (drule State_welldef_Heap_lookup_dom, assumption)
apply (drule AL_domD) apply clarsimp
done

lemma State_welldef_AbsField_lookup_dom:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; d:AbsFields H\<rbrakk>
       \<Longrightarrow> d : AL_dom N"
apply (simp add: State_welldef_def, clarsimp)
apply (erule SetTriv, assumption)
done
lemma State_welldef_AbsField_lookup:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; d:AbsFields H\<rbrakk>
       \<Longrightarrow> \<exists> tp . N\<down>d = Some tp"
apply (drule State_welldef_AbsField_lookup_dom, assumption)
apply (drule AL_domD) apply clarsimp
done

lemma State_welldef_lookup_AbsField:
"State_welldef (S, \<Gamma>, H) N \<Longrightarrow> AbsFields H \<subseteq> AL_dom N"
by (simp add: State_welldef_def)

lemma State_welldef_Field_lookup_dom:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; H\<down>c = Some (D, G); G\<down>f = Some d\<rbrakk>
       \<Longrightarrow> d : AL_dom N"
apply (simp add: State_welldef_def, clarsimp)
apply (drule AbsFieldsI, assumption)
apply (erule SetTriv, assumption)
done
lemma State_welldef_Field_lookup:
"\<lbrakk>State_welldef (S, \<Gamma>, H) N; H\<down>c = Some (D, G); G\<down>f = Some d\<rbrakk>
       \<Longrightarrow> \<exists> tp . N\<down>d = Some tp"
apply (drule State_welldef_Field_lookup_dom, assumption, assumption)
apply (drule AL_domD) apply clarsimp
done

lemma State_welldef_lookupHeap:"State_welldef (S, \<Gamma>, H) N \<Longrightarrow> AL_dom H \<subseteq> AL_dom N" 
by (simp add: State_welldef_def) 

lemma State_welldef_lookupHeap1:"\<lbrakk>State_welldef (S, \<Gamma>, H) N; c: AL_dom H\<rbrakk> \<Longrightarrow> c : AL_dom N" 
apply (erule SetTriv) apply (erule State_welldef_lookupHeap)
done

lemma State_welldef_OPS_lookup_set_dom:
"State_welldef (S, \<Gamma>, H) N \<Longrightarrow> set S \<subseteq> AL_dom N"
by (simp add: State_welldef_def)

lemma State_welldef_Store_lookup_cod_dom:
"State_welldef (S, \<Gamma>, H) N \<Longrightarrow> AL_cod \<Gamma> \<subseteq> AL_dom N"
by (simp add: State_welldef_def)

lemma State_welldef_DOM_ALdom: "State_welldef \<Sigma> N \<Longrightarrow> DOM \<Sigma> \<subseteq> AL_dom N"
apply (simp add: State_welldef_def DOM_def)
apply (case_tac \<Sigma>, clarsimp)
apply (erule disjE, fast)
apply (erule disjE, fast)
apply (erule disjE, fast)
apply fast
done

end