(*File: svn/mobius/Formalisation/LMU_BCL/AssocLists.thy
  Author: L Beringer & M Hofmann, LMU Munich
  Date: 04/04/2006
  Purpose: Representation of partial maps by association lists         
*)
(*<*)
theory AssocLists = Trivialities:
(*>*)

section{*Preliminaries*}
subsection{*Association lists \label{sec:preliminaries}*}

text{*Finite maps are used frequently, both in the representation of
syntax and in the program logics.  For the present development, the
following data type of association lists suffices.*}

types ('a,'b) AssList = "('a \<times> 'b) list"

consts lookup::"('a, 'b) AssList \<Rightarrow> 'a \<Rightarrow> ('b option)"
          	("_\<down>_" [1000,1000] 90)
primrec
"lookup [] l = None"
"lookup (h # t) l = (if fst h = l then Some(snd h) else lookup t l)"

(*<*)
lemma AL_lookup1[rule_format]:"\<forall> a b . (L\<down>a = Some b) \<longrightarrow> ((a,b) mem L)"
by (induct L, simp_all)
(*by (induct L, simp_all add: set_mem_eq)*)

lemma AL_Triv1:"a=b \<Longrightarrow> L\<down>a = L\<down>b" by simp

lemma AL_Triv2: "\<lbrakk>L\<down>a = X; L\<down>a = Y\<rbrakk> \<Longrightarrow> X=Y" by simp

lemma AL_Triv3:"\<lbrakk>L=M ; M\<down>b = X\<rbrakk> \<Longrightarrow> L\<down>b = X" by clarsimp

lemma AL_Triv4:"\<lbrakk>L=M ; L\<down>b = X\<rbrakk> \<Longrightarrow> M\<down>b = X" by clarsimp

(*>*)

text{*The statement following the type declaration of @{text lookup}
indicates that we may use the infix notation @{text "L\<down>a"} for the
lookup operation henceforth, and asserts some precedence for
bracketing. In a similar way, numerous shorthands are introduced for
various operations throughout this document.*}

consts delete::"('a, 'b) AssList \<Rightarrow> 'a \<Rightarrow> ('a, 'b) AssList"
primrec
"delete [] a = []"
"delete (h # t) a = (if (fst h) = a then delete t a else (h # (delete t a)))"

(*<*)
lemma AL_delete1: "(delete L a) \<down> a = None"
by (induct L, auto)

lemma AL_delete2[rule_format]:"b \<noteq> a \<longrightarrow> (delete l a)\<down>b = l\<down>b"
by (induct l, simp, simp)

lemma AL_delete2a[rule_format]:"a \<noteq> b \<longrightarrow> (delete l a)\<down>b = l\<down>b"
by (induct l, simp, simp)

lemma AL_delete3: "L\<down>a = None \<Longrightarrow> delete L a = L"
apply (induct L)
apply simp
apply clarsimp
apply (subgoal_tac "L\<down>a = None", clarsimp)
(*apply (subgoal_tac "list\<down>a = None", clarsimp)*)
apply (case_tac "aa=a", clarsimp) apply clarsimp
done

lemma AL_delete4[simp]: "length(delete t a) < Suc(length t)"
by (induct t, simp+) 

lemma AL_delete5[rule_format]:"\<lbrakk>b \<noteq> a; l\<down>b = x\<rbrakk> \<Longrightarrow> (delete l a)\<down>b = x"
apply (drule_tac l=l in AL_delete2) apply simp done

lemma AL_delete5b[rule_format]:"\<lbrakk>a \<noteq> b; l\<down>b = x\<rbrakk> \<Longrightarrow> (delete l a)\<down>b = x"
apply (drule_tac l=l in AL_delete2a) apply simp done

lemma AL_delete6: "(delete L x)\<down>a = Some X \<Longrightarrow> x\<noteq>a \<and> L\<down>a=Some X"
apply (case_tac "x=a", simp add: AL_delete1) 
apply (frule_tac l=L in AL_delete2a) apply simp
done

lemma AL_delete_commutative: "delete (delete M c) x = delete (delete M x) c"
apply (induct M)
apply simp
apply simp
done

lemma AL_delete_idempotent: "delete M x = delete (delete M x) x"
by (induct M, auto)

lemma mem_delete1[rule_format]: "(x, y) mem delete M x \<longrightarrow> Q"
by (induct M, auto)

lemma mem_delete2[rule_format]: "(z, y) mem delete M x \<longrightarrow> x \<noteq> xa \<longrightarrow> (z,y) mem M"
by (induct M, auto)

lemma set_map_fst_delete: "set (map fst (delete M x)) = set (map fst M) - {x}"
apply (induct M)
apply simp
apply clarsimp
  apply (rule, clarsimp) apply (erule set_map_fst_set) apply assumption
  apply clarsimp apply (erule set_map_fst_set) apply assumption
done

lemma map_fst_set_delete: "fst ` set (delete M x) = (fst ` set M) - {x}"
apply (induct M)
apply simp
apply clarsimp 
  apply (rule, clarsimp) apply (erule set_map_fst_set) apply assumption
  apply clarsimp apply (erule set_map_fst_set) apply assumption 
done

lemma not_set_mem_snd_lookup_Q[rule_format]: "xa \<notin> snd ` set M \<longrightarrow> M\<down>a = Some xa \<longrightarrow> Q"
by (induct M, auto)

(*>*)

constdefs upd::"('a, 'b) AssList \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> ('a, 'b) AssList" 
          ("_[_\<mapsto>_]" [1000,1000,1000] 1000)
"upd L a b == (a,b) # (delete L a)"
(*<*)

lemma AL_update1: "L[a\<mapsto>b]\<down>a = Some b"
by (simp add: upd_def)

lemma AL_update1a: "a=c \<Longrightarrow> L[a\<mapsto>b]\<down>c = Some b" by (simp add: AL_update1)

lemma AL_update2: "a \<noteq> b \<Longrightarrow> L[a\<mapsto>v]\<down>b = L\<down>b"
by (induct L, simp_all add: upd_def)
lemma AL_update2a: "b \<noteq> a \<Longrightarrow> L[a\<mapsto>v]\<down>b = L\<down>b"
by (rule AL_update2, fast) 

lemma AL_update3: "\<lbrakk>L[a\<mapsto>v]\<down>b = X; a \<noteq> b\<rbrakk> \<Longrightarrow> L\<down>b=X"
by (subgoal_tac "L[a\<mapsto>v]\<down>b =L\<down>b", clarsimp,erule AL_update2) 

lemma AL_update3a: "\<lbrakk>L[a\<mapsto>v]\<down>b = X; b \<noteq> a\<rbrakk> \<Longrightarrow> L\<down>b=X"
by (erule AL_update3, fast)

lemma AL_update4:"\<lbrakk> L\<down>b = Some X; a \<noteq> b\<rbrakk>\<Longrightarrow> L[a\<mapsto>v]\<down>b = Some X"
by (subgoal_tac "L[a\<mapsto>v]\<down>b = L\<down>b", clarsimp, erule AL_update2)
lemma AL_update4a:"\<lbrakk> L\<down>b = Some X; b \<noteq> a\<rbrakk>\<Longrightarrow> L[a\<mapsto>v]\<down>b = Some X"
by (erule AL_update4, fast) 

lemma AL_update5: "\<lbrakk>L\<down>b = M; a \<noteq> b\<rbrakk> \<Longrightarrow> L[a\<mapsto>X]\<down>b = M"
apply (subgoal_tac "L[a\<mapsto>X]\<down>b = L\<down>b", simp)
apply (erule AL_update2)
done
lemma AL_update5a: "\<lbrakk>L\<down>b = M; b \<noteq> a\<rbrakk> \<Longrightarrow> L[a\<mapsto>X]\<down>b = M"
by (erule AL_update5, fast) 

lemma AL_upd_split:"N[a\<mapsto>b]\<down>c = Some d \<Longrightarrow> (c=a \<and> d=b) \<or> (a \<noteq> c \<and> N\<down>c = Some d)"
apply (simp add: upd_def)
apply clarsimp
apply (case_tac "a=c",clarsimp) apply clarsimp
apply (subgoal_tac "(delete N a)\<down>c = N\<down>c", clarsimp) apply (rule AL_delete2) apply fast
done

lemma AL_delete_upd_same: "delete (M[x\<mapsto>y]) x = delete M x"
apply (induct M) apply (simp add: upd_def)
apply (simp add: upd_def) 
done

lemma AL_delete_upd_diff[rule_format]: "z \<noteq> x \<longrightarrow> delete (M[x\<mapsto>y]) z = (delete M z)[x\<mapsto>y]"
apply (induct M) apply (simp add: upd_def)
apply (simp add: upd_def) 
apply clarsimp
done

lemma AL_delete_upd_diff2[rule_format]: "x \<noteq> z \<Longrightarrow> delete (M[x\<mapsto>y]) z = (delete M z)[x\<mapsto>y]"
apply (rule AL_delete_upd_diff) apply fast
done

lemma AL_delete7:"delete M[x\<mapsto>y] x = delete M x"
apply (induct M)
apply (simp add: upd_def)
apply (simp add: upd_def)
done

lemma AL_delete8:"(delete fM[fc\<mapsto>y] fc)[fc\<mapsto>x] = (delete fM fc)[fc\<mapsto>x]"
apply (induct fM)
apply (simp add: upd_def)
apply (simp add: upd_def)
done

lemma AL_upd_idempotent:"MM[b\<mapsto>a][b\<mapsto>a] = MM[b\<mapsto>a]"
apply (simp add: upd_def)
apply (subgoal_tac "delete MM b = delete (delete MM b) b",clarsimp)
apply (rule AL_delete_idempotent)
done

(*>*)

constdefs emp::"('a, 'b)AssList"
"emp == []"

(*<*)
lemma AL_emp1: "emp\<down>a = None"
by (simp add: emp_def)

constdefs AL_isEmp::"('a,'b) AssList \<Rightarrow> bool"
"AL_isEmp l == (\<forall> a . l\<down>a = None)"

lemma AL_delete_emp: "delete emp x = emp" by (simp add: emp_def)

(*>*)

constdefs contained::"('a,'b) AssList \<Rightarrow> ('a,'b) AssList \<Rightarrow> bool"
"contained L M == (\<forall> a b . L\<down>a = Some b \<longrightarrow> M\<down>a = Some b)"

consts AL_Size::"('a,'b) AssList \<Rightarrow> nat"
       ("|_|" [1000] 1000)
recdef AL_Size "measure (\<lambda> l . length l)"
"AL_Size [] = 0"
"AL_Size (h # t) = Suc (AL_Size (delete t (fst h)))"
(*<*)
lemma AL_Size_Zero: "|L| = 0 \<Longrightarrow> None = L\<down>a"
by (induct L, simp, simp)

lemma AL_Size_Suc: "\<forall> n . |L| = Suc n \<longrightarrow> (\<exists> a b . L\<down>a = Some b)"
apply (induct L)
apply clarsimp 
apply clarsimp
apply (rule_tac x=a in exI, simp)
done

lemma AL_Size_UpdateSuc:"L\<down>a = None \<Longrightarrow> |L[a\<mapsto>b]| = Suc |L|"
apply (induct L)
apply (simp add: upd_def)
apply (simp add: upd_def)
apply clarsimp
apply (case_tac "aa=a", clarsimp) 
apply clarsimp
apply (drule AL_delete3) apply clarsimp
done

lemma AL_Size_UpdateSuc1:"\<lbrakk>L\<down>a = None; |L| = n\<rbrakk>\<Longrightarrow> |L[a\<mapsto>b]| = Suc n"
apply (drule AL_Size_UpdateSuc) apply simp done

lemma AL_Size_SucSplit[rule_format]:
"\<forall> n . |L| = Suc n \<longrightarrow> 
  (\<exists> a b M . |M| = n \<and> M\<down>a = None \<and> L\<down>a = Some b \<and> (\<forall> c . c\<noteq> a \<longrightarrow> M\<down>c = L\<down>c))"
apply (induct L)
apply clarsimp
apply clarsimp
apply (rule_tac x=a in exI)  apply simp
apply (rule_tac x="delete L a" in exI, clarsimp)
  apply rule  apply (rule AL_delete1)
apply clarsimp apply (erule AL_delete2)
done
(*>*)

constdefs AL_dom::"('a,'b) AssList \<Rightarrow> 'a set"
"AL_dom L \<equiv> {a . L\<down>a \<noteq> None}"

lemma AL_domI:"H\<down>d = Some x \<Longrightarrow> d \<in> AL_dom H" by (simp add: AL_dom_def)
lemma AL_domD:" d \<in> AL_dom H \<Longrightarrow> \<exists> x . H\<down>d = Some x" by (simp add: AL_dom_def)
lemma AL_dom_emp: "AL_dom emp ={}" by (simp add: AL_dom_def AL_emp1)

lemma AL_notdomD: "x \<notin> AL_dom N \<Longrightarrow> N\<down>x = None"
apply (case_tac "N\<down>x", clarsimp) apply (drule AL_domI) apply simp
done
lemma AL_notdomD_delete: "x \<notin> AL_dom N \<Longrightarrow> delete N x = N"
apply (rule AL_delete3) apply (erule AL_notdomD) done

lemma AL_dom_upd:"AL_dom L \<subseteq> AL_dom L[a\<mapsto>b]"
apply (simp add: AL_dom_def)
apply clarsimp
apply (case_tac "a=x", clarsimp) apply (rule, rule AL_update1)
apply (rule, erule AL_update4) apply assumption
done

lemma AL_dom_upd1:"AL_dom L[a\<mapsto>b] = AL_dom L \<union> {a}"
apply (simp add: AL_dom_def)
apply rule
  apply clarsimp apply (drule AL_update3) apply fast apply (rule, assumption)
  apply clarsimp
    apply rule apply (rule, rule AL_update1)
    apply (rule, clarsimp) apply (case_tac "a=x", clarsimp) apply (rule, rule AL_update1) apply (rule, erule AL_update4) apply assumption
done

lemma AL_dom_upd2:"a \<in> AL_dom N[a\<mapsto>b]"
by (simp add: AL_dom_upd1)

lemma AL_dom_upd3:"X \<subseteq> AL_dom N \<Longrightarrow> X \<subseteq> AL_dom N[a\<mapsto>b]"
apply (simp add: AL_dom_upd1)
apply fast
done

lemma AL_dom_upd4:"AL_dom X \<subseteq> AL_dom N \<Longrightarrow> AL_dom X[a \<mapsto> k] \<subseteq> AL_dom N[a\<mapsto>b]"
apply (simp add: AL_dom_upd1)
apply fast
done

lemma AL_dom_delete: "AL_dom (delete N x) = AL_dom N - {x}"
apply (induct N)
apply (simp add: AL_dom_def)
apply (simp add: AL_dom_def)
apply (rule, clarsimp)
  apply (rule, rule,clarsimp) apply clarsimp
apply clarsimp
  apply (rule, rule,clarsimp) 
    apply (case_tac "xa=x", clarsimp) apply (simp add: AL_delete1)
    apply clarsimp apply (drule_tac b=xa and l=N in AL_delete2) apply clarsimp
  apply clarsimp
    apply (case_tac "xa=x", clarsimp) apply (simp add: AL_delete1)
    apply (drule_tac b=xa and l=N in AL_delete2) apply clarsimp
done

lemma AL_upd_delete_upd:"N = M[x\<mapsto>y] \<Longrightarrow> N=(delete M x)[x\<mapsto>y]"
apply (simp add: upd_def) apply (insert AL_delete_idempotent) apply fast done

lemma AL_delete_upd_upd:"(delete M x)[x\<mapsto>y] = M[x\<mapsto>y]" 
apply (simp add: upd_def) apply (insert AL_delete_idempotent [of M x]) apply fastsimp done

lemma upd_upd: "M[x\<mapsto>a][x\<mapsto>b] = M[x\<mapsto>b]"
apply (simp add: upd_def)
apply (subgoal_tac "delete M x = delete (delete M x) x",simp) 
apply (rule AL_delete_idempotent)
done

lemma AL_updupd:"L[a\<mapsto>b]=M \<Longrightarrow> L[a\<mapsto>c][a\<mapsto>b]=M"
apply (simp add: upd_def)
apply (subgoal_tac "delete (delete L a) a = delete L a", clarsimp)
apply (erule thin_rl) apply (insert AL_delete_idempotent [of L a]) apply simp
done

constdefs AL_cod::"('a,'b) AssList \<Rightarrow> 'b set"
"AL_cod L \<equiv> {b . \<exists> a . L\<down>a = Some b}"

lemma AL_codI:"L\<down>a = Some b \<Longrightarrow> b:AL_cod L" by (simp add:AL_cod_def, fast) 
lemma AL_codD:"b:AL_cod L \<Longrightarrow> \<exists> a . L\<down>a = Some b" by (simp add:AL_cod_def) 
lemma AL_cod_emp: "AL_cod emp ={}" by (simp add: AL_cod_def AL_emp1)

lemma AL_cod_upd1:"AL_cod L[a\<mapsto>b] \<subseteq> AL_cod L \<union> {b}"
apply (simp add: AL_cod_def)
  apply clarsimp apply (drule AL_upd_split) apply clarsimp apply (rule, assumption)
done

lemma AL_cod_upd2: "y:AL_cod M[x\<mapsto>y]"  apply (simp add: AL_cod_def) apply (rule, rule AL_update1) done

lemma AL_cod_upd3: "x \<notin> AL_dom L \<Longrightarrow> AL_cod L \<subseteq> AL_cod L[x\<mapsto>c]"
apply (simp add: upd_def AL_cod_def, clarsimp)
apply (rule_tac x=a in exI, rule)
apply clarsimp apply (drule AL_domI) apply fast
apply clarsimp apply (drule_tac l=L in AL_delete2a) apply simp
done

lemma AL_cod_delete: "AL_cod (delete M x) \<subseteq> AL_cod M"
apply (induct M)
apply (simp add: AL_cod_def)
apply (simp add: AL_cod_def)
apply clarsimp
  apply (rule, clarsimp)
    apply (drule AL_delete6) apply clarsimp apply (rule_tac x=aa in exI, clarsimp)
  apply clarsimp apply (case_tac "a=aa", clarsimp) apply (rule_tac x=aa in exI, simp)
                 apply clarsimp apply (drule AL_delete6) apply clarsimp apply (rule_tac x=aa in exI, clarsimp)
done

lemma AL_cod_split: "x \<in> AL_cod L[a\<mapsto>b] \<Longrightarrow> x:AL_cod L \<or> x=b"
apply (simp add: AL_cod_def, clarsimp)
apply (drule AL_upd_split) apply fast 
done

constdefs AL_sum::"('a,'b) AssList \<Rightarrow> ('a,'b) AssList \<Rightarrow> ('a,'b) AssList \<Rightarrow> bool"
"AL_sum N M NM \<equiv> AL_dom NM = (AL_dom N) \<union> (AL_dom M) \<and> contained N NM \<and> contained M NM"

(*<*)
lemma AL_sum_member:"\<lbrakk>x \<in> AL_dom H; AL_sum H H1 HH\<rbrakk> \<Longrightarrow> x \<in> AL_dom HH"
by (simp add: AL_sum_def)

lemma AL_sum_entry:"\<lbrakk>H\<down>x = Some a; AL_sum H H1 HH\<rbrakk> \<Longrightarrow> HH\<down>x = Some a"
by (simp add: AL_sum_def contained_def) 

lemma AL_sum_entryR:"\<lbrakk>H1\<down>x = Some a; AL_sum H H1 HH\<rbrakk> \<Longrightarrow> HH\<down>x = Some a"
by (simp add: AL_sum_def contained_def) 

lemma AL_sum_split:
"\<lbrakk>AL_sum L M N; N\<down>c = Some x\<rbrakk> \<Longrightarrow> L\<down>c = Some x \<or> M\<down>c = Some x"
apply (simp add: AL_sum_def) apply clarsimp
apply (simp add: contained_def)
apply (subgoal_tac "c:AL_dom L \<or> c:AL_dom M")
prefer 2 apply (simp add: AL_dom_def) apply fast
apply (erule disjE) apply (simp add: AL_dom_def) apply clarsimp
apply (simp add: AL_dom_def) apply clarsimp
done

lemma AL_sum_exists[rule_format]: "\<forall> N . AL_dom N \<inter> AL_dom M = {} \<longrightarrow> (\<exists> NM . AL_sum N M NM)"
apply (induct M)
  apply clarsimp apply (rule_tac x=N in exI) apply (simp add: AL_sum_def AL_dom_def contained_def) 
  apply clarsimp apply (erule_tac x=N in allE) apply (erule impE) apply (simp add: AL_dom_def) apply fastsimp 
apply (erule exE)
  apply (rule_tac x="(a,b) # x" in exI)
  apply (simp add: AL_sum_def)
  apply clarsimp
  apply rule apply (subgoal_tac "AL_dom ((a, b) # x) = insert a (AL_dom x)", clarsimp) apply (subgoal_tac "AL_dom ((a, b) # M) = insert a (AL_dom M)", clarsimp) apply (erule thin_rl)+  apply (simp add: AL_dom_def) apply fast
    apply (erule thin_rl)+  apply (simp add: AL_dom_def) apply fast
apply rule apply (simp add: contained_def) apply clarsimp
     apply (subgoal_tac "a \<notin> AL_dom N")
     prefer 2 apply (subgoal_tac "AL_dom ((a, b) # M) = insert a (AL_dom M)", clarsimp) apply (erule thin_rl)+
               apply (simp add: AL_dom_def) apply fast 
     apply (simp add: AL_dom_def)
apply (simp add: contained_def) 
done

lemma AL_sum_sum: "\<lbrakk>AL_sum H H1 HH; AL_sum HH Ha bb; AL_sum H1 Ha Ha1\<rbrakk> \<Longrightarrow> AL_sum H Ha1 bb"
apply (simp add: AL_sum_def)
apply (rule, clarsimp)
  apply (rule, clarsimp) 
    apply (rule, clarsimp)
    apply (rule, clarsimp) apply clarsimp
  apply clarsimp
    apply (rule, clarsimp)
    apply (rule, clarsimp) apply clarsimp
apply (simp add: contained_def)
apply (rule, clarsimp)
  apply (subgoal_tac "a : AL_dom H1 \<or> a : AL_dom Ha")
    apply (erule disjE)
      apply (subgoal_tac "a : AL_dom HH", simp add: AL_dom_def, clarsimp) apply fast
    apply (simp add: AL_dom_def, clarsimp)
apply (subgoal_tac "a : AL_dom Ha1", fast) 
  apply (unfold AL_dom_def) apply clarify apply (rule, assumption)
done

lemma AL_sum_aux:
  "\<lbrakk>af\<down>ca = Some x; AL_sum N Na af; contained N NN; AL_sum NN Na xb\<rbrakk> \<Longrightarrow> xb\<down>ca = Some x"
apply (drule AL_sum_split, assumption)
apply (subgoal_tac "NN\<down>ca = Some x \<or> Na\<down>ca = Some x")
  apply (simp add: AL_sum_def contained_def) apply fast
apply (simp add: contained_def)
apply fast
done

lemma AL_dom_subset:"\<lbrakk>x \<in> X; X \<subseteq> AL_dom M\<rbrakk> \<Longrightarrow> \<exists> y . M\<down>x = Some y"
by (simp add: AL_dom_def, fast)

lemma AL_sum_exists2[rule_format]: "\<lbrakk>AL_dom N1 \<subseteq> AL_dom N; AL_dom N \<inter> AL_dom M = {}\<rbrakk> \<Longrightarrow> (\<exists> NM . AL_sum N1 M NM)"
apply (rule AL_sum_exists) apply fast
done

lemma contained_AL_dom: "\<lbrakk>contained NN Nd; AL_dom Nd \<inter> AL_dom Na = {}\<rbrakk> \<Longrightarrow>  AL_dom NN \<inter> AL_dom Na = {}"
apply (subgoal_tac "AL_dom NN \<subseteq>AL_dom Nd", fast)
apply (simp add: AL_dom_def contained_def)  apply fast
done

lemma contained_transitive:"\<lbrakk>contained N L;contained L M\<rbrakk> \<Longrightarrow> contained N M"
by (simp add: contained_def)

lemma contained1:"\<lbrakk>AL_dom au = AL_dom ap \<union> AL_dom N; contained ap au; contained N au; contained ap ak; contained N ak\<rbrakk> \<Longrightarrow> contained au ak"
apply (simp add: contained_def)
apply clarsimp
apply (subgoal_tac "(\<exists> y . ap\<down>a = Some y) \<or> (\<exists> y . N\<down>a = Some y)")
  apply (erule disjE, clarsimp) apply clarsimp
apply (drule AL_domI) apply clarsimp
apply (erule disjE)
  apply (erule AL_domD)
apply (drule AL_domD) apply clarsimp
done

lemma contained2: "\<lbrakk> contained N NN; NN\<down>c = Some x; c:AL_dom N\<rbrakk> \<Longrightarrow> N\<down>c = Some x"
apply (simp add: contained_def)
apply (drule AL_domD, clarsimp)
done

lemma contained3: "\<lbrakk> contained N NN; N\<down>c = Some x\<rbrakk> \<Longrightarrow> NN\<down>c = Some x"
by (simp add: contained_def)

lemma contained4:
  "\<lbrakk> ao\<down>c = Some x; AL_dom ao = AL_dom N \<union> AL_dom Na; contained N ao; contained Na ao; contained Na N1; contained N N1\<rbrakk>
         \<Longrightarrow> N1\<down>c = Some x"
    apply (subgoal_tac "c: AL_dom N \<union> AL_dom Na")
    prefer 2 apply (rotate_tac -1, drule AL_domI)  apply simp
    apply clarsimp
    apply (erule disjE)
      apply (drule AL_domD) apply clarsimp
      apply (drule contained3, assumption)
      apply (drule contained3, assumption)
      apply clarsimp
    apply (drule AL_domD, clarsimp)
      apply (drule contained3, assumption)
      apply (drule contained3, assumption)
      apply clarsimp
done

lemma contained5:"contained N M \<Longrightarrow> AL_dom N \<subseteq> AL_dom M"
apply clarsimp
apply (drule AL_domD, clarsimp)
apply (rule AL_domI)
apply (erule contained3) apply assumption
done

lemma contained6:"\<lbrakk>contained N M; x:AL_dom N\<rbrakk> \<Longrightarrow> x:AL_dom M"
apply (drule contained5) apply fast
done

lemma contained7:"\<lbrakk>contained N M; M\<down>x = Some y; x:AL_dom N\<rbrakk> \<Longrightarrow> N\<down>x = Some y"
apply (drule AL_domD, clarsimp)
apply (drule contained3, assumption) apply simp
done

lemma contained_upd: "x \<notin> AL_dom N \<Longrightarrow> contained N N[x\<mapsto>y]" 
apply (simp add: contained_def, clarsimp)
apply (rule AL_update5) apply assumption apply (drule AL_domI) apply fast
done

lemma contained_delete: "contained (delete L c) L"
apply (simp add: contained_def, clarsimp)
apply (drule AL_delete6) apply simp
done

lemma contained_refl: "contained N N" by (simp add: contained_def)

lemma SumSum_lookup: "\<lbrakk>AL_sum H[cl\<mapsto>(C, emp)] Ha H1; AL_sum H Ha bf; H1\<down>d = Some y; d \<in> AL_dom H; cl \<noteq> d\<rbrakk> \<Longrightarrow> bf\<down>d = Some y"
apply (simp add: AL_sum_def, clarsimp)
apply (drule contained2, assumption) 
apply (simp add: AL_dom_upd1)
apply (rotate_tac -1, drule AL_update3) apply assumption
apply (erule contained3) apply assumption
done

lemma AL_sum_dom_split:"\<lbrakk>AL_sum H H1 H2; x:AL_dom H2\<rbrakk> \<Longrightarrow> x:AL_dom H \<or> x:AL_dom H1"
by (simp add: AL_sum_def)

lemma AL_sum3: "\<lbrakk>AL_sum H H1 H2; H\<down>x = Some y\<rbrakk> \<Longrightarrow> H2\<down>x = Some y"
apply (simp add: AL_sum_def, clarsimp)
apply (erule contained3) apply assumption
done

lemma AL_sum3a: "\<lbrakk>AL_sum H H1 H2; x:AL_dom H\<rbrakk> \<Longrightarrow> x:AL_dom H2"
by (simp add: AL_sum_def)

lemma AL_sum4: "\<lbrakk>AL_sum H H1 H2; H1\<down>x = Some y\<rbrakk> \<Longrightarrow> H2\<down>x = Some y"
apply (simp add: AL_sum_def, clarsimp)
apply (erule contained3, assumption) 
done

lemma AL_sum4a: "\<lbrakk>AL_sum H H1 H2; x:AL_dom H1\<rbrakk> \<Longrightarrow> x:AL_dom H2"
by (simp add: AL_sum_def)

lemma AL_sum_upd_sum:
"\<lbrakk>AL_sum H[c\<mapsto>(C, upd F f c1)] Ha H1; H1\<down>d = Some (D, Fa);  
  AL_sum H Ha bf; d \<noteq> c\<rbrakk> \<Longrightarrow> bf\<down>d = Some (D, Fa)"
apply (drule AL_sum_split) apply assumption
apply (erule disjE, drule AL_upd_split) apply clarsimp apply (erule AL_sum3, assumption)
apply (erule AL_sum4, assumption)
done

lemma AL_sum_no_conflicts: "\<lbrakk>AL_sum N L M; N\<down>x = Some y; L\<down>x = Some z\<rbrakk> \<Longrightarrow> y=z"
apply (frule AL_sum3, assumption)
apply (frule AL_sum4, assumption) apply simp
done

lemma AL_sum_extensional:"\<lbrakk>AL_sum N N1 N0; AL_sum N N1 NN0\<rbrakk> \<Longrightarrow> \<forall>x c. (N0\<down>x = Some c) = (NN0\<down>x = Some c)"
apply clarsimp
apply (rule, drule AL_sum_split, assumption)
  apply (erule disjE, erule AL_sum3) apply assumption
  apply (erule AL_sum4) apply assumption
apply (drule AL_sum_split, assumption)
  apply (erule disjE, erule AL_sum3) apply assumption
  apply (erule AL_sum4) apply assumption
done

lemma AL_sum3b: "AL_sum H K HK \<Longrightarrow> AL_dom H \<subseteq> AL_dom HK"
by (rule, erule AL_sum3a, assumption)

lemma AL_sum4b: "AL_sum H K HK \<Longrightarrow> AL_dom K \<subseteq> AL_dom HK"
by (rule, erule AL_sum4a, assumption)

(*>*)

lemma mem_map_snd_set[rule_format]: "(a, x) mem M \<longrightarrow> x \<in> snd ` set M"
by (induct M, auto)

lemma contained_emp:"contained emp N"
by (simp add: contained_def emp_def)

lemma contained_emp2: "contained N emp \<Longrightarrow> N=emp" 
apply (simp add: contained_def)
apply (case_tac N, clarsimp) apply (simp add: emp_def)
apply clarsimp apply (erule_tac x=a in allE, clarsimp) apply (simp add: AL_emp1)
done

lemma contained_sum_sum:"\<lbrakk>contained N M; AL_sum N Nx K; AL_sum M Nx L\<rbrakk> \<Longrightarrow> contained K L"
apply (simp (no_asm) add: contained_def, clarsimp)
apply (drule AL_sum_split, assumption)
apply (erule disjE) apply (erule AL_sum3) apply (erule contained3,assumption)
apply (erule AL_sum4, assumption)
done

lemma contained8: "\<lbrakk>AL_sum N Nx N0;  contained N NN;AL_sum NN Nx N1\<rbrakk> \<Longrightarrow> contained N0 N1"
apply (simp (no_asm) add: contained_def, clarsimp)
apply (drule AL_sum_split, assumption)
apply (erule disjE, drule contained3, assumption) apply (erule AL_sum3, assumption)
apply (erule AL_sum4, assumption)
done

lemma AL_sum_emp1:"AL_sum emp N N"
apply (simp add: AL_sum_def AL_dom_emp)
apply (rule, rule contained_emp)
apply (rule contained_refl)
done

lemma AL_sum_emp2:"AL_sum N emp N"
apply (simp add: AL_sum_def AL_dom_emp)
apply (rule, rule contained_refl)
apply (rule contained_emp)
done

lemma AL_sum_emp3: "AL_sum H K emp \<Longrightarrow> H=emp \<and> K=emp"
apply (simp add: AL_sum_def)
apply (simp add: AL_dom_emp)
apply clarsimp
apply (case_tac H, clarsimp) 
  apply (case_tac K, clarsimp)  apply (simp add: emp_def)
  apply clarsimp apply (subgoal_tac "a:AL_dom ((a, b) # list)", fast) apply (erule thin_rl)+ apply (simp add: AL_dom_def)
apply clarsimp apply (subgoal_tac "a:AL_dom ((a, b) # list)", fast) apply (erule thin_rl)+ apply (simp add: AL_dom_def) 
done

lemma AL_sum_emp4: "AL_sum emp emp H \<Longrightarrow> H=emp"
apply (simp add: AL_sum_def)
apply (case_tac H, clarsimp) apply (simp add: emp_def)
apply clarsimp apply (simp add: AL_dom_def emp_def) apply (erule_tac x=a in allE, clarsimp)
done

lemma AL_sum_upd_upd: 
  "\<lbrakk>AL_sum N M L; c:AL_dom N; AL_dom N \<inter> AL_dom M = {}\<rbrakk> \<Longrightarrow> AL_sum N[c\<mapsto>x] M L[c\<mapsto>x]"
apply (simp add: AL_sum_def AL_dom_upd1) 
             apply rule apply (simp (no_asm) add: contained_def, clarsimp)
                        apply (drule AL_upd_split)
                        apply (erule disjE, clarsimp) apply (rule AL_update1)
                        apply clarsimp apply (rule AL_update5) 
                                         apply (erule contained3, assumption)
                                       apply assumption
             apply (simp (no_asm) add: contained_def, clarsimp)
               apply (drule_tac N=M and NN=L in contained3, assumption)
               apply (erule AL_update5) 
               apply (drule_tac H=M in AL_domI) apply fast 
done

lemma AL_sum_upd_upd2: 
  "\<lbrakk>AL_sum N M L; c\<notin> AL_dom M; AL_dom N \<inter> AL_dom M = {}\<rbrakk> \<Longrightarrow> AL_sum N[c\<mapsto>x] M L[c\<mapsto>x]"
apply (simp add: AL_sum_def AL_dom_upd1) 
             apply rule apply (simp (no_asm) add: contained_def, clarsimp)
                        apply (drule AL_upd_split)
                        apply (erule disjE, clarsimp) apply (rule AL_update1)
                        apply clarsimp apply (rule AL_update5) 
                                         apply (erule contained3, assumption)
                                       apply assumption
             apply (simp (no_asm) add: contained_def, clarsimp)
               apply (drule_tac N=M and NN=L in contained3, assumption)
               apply (erule AL_update5) 
               apply (drule_tac H=M in AL_domI) apply fast 
done

lemma map_add_disj_AL_sum: 
  "\<lbrakk> \<forall>c. c \<in> AL_dom N \<longrightarrow> f' c = f c; \<forall>c. c \<in> AL_dom N \<longrightarrow> g c = f' c; 
     \<forall>c. c \<in> AL_dom M \<longrightarrow> g c = h c; c \<in> AL_dom aj; 
     AL_sum N M aj; AL_dom N \<inter> AL_dom M = {}; dom h = AL_dom M \<rbrakk>
  \<Longrightarrow> g c = (f ++ h) c"
apply (simp add: AL_sum_def, clarsimp)
                     apply (erule disjE) 
                       apply (erule_tac x=c in allE, erule impE, assumption)
                       apply (erule_tac x=c in allE, erule impE, assumption) 
                       apply (case_tac "h c") apply (drule_tac h=h and f=f in map_add_NoneD) apply clarsimp 
                       apply (subgoal_tac "c:AL_dom M") 
                         apply (drule_tac X="AL_dom N" and Y="AL_dom M" and x=c and y=c in SetTriv106) apply assumption+  apply simp 
                       apply (drule_tac m=h in domI, simp) 
                     apply (erule_tac x=c in allE, erule impE, assumption) apply clarsimp 
                       apply (subgoal_tac "c:dom h", drule_tac m=h in domD, clarsimp) 
                       apply simp
done

consts AL::"('a,'b) AssList set"
inductive AL intros
AL_EMP: "emp : AL"
AL_CONS: "\<lbrakk>M : AL; x\<notin>AL_dom M; N=M[x\<mapsto>y]\<rbrakk> \<Longrightarrow> N :AL"

lemma AL_emp:"[]:AL"
apply (subgoal_tac "emp:AL", simp add: emp_def) apply (rule AL_EMP) done

lemma AL_DEL:"M : AL \<Longrightarrow> delete M x :AL"
apply (erule AL.induct)
apply (simp add: emp_def) apply (rule AL_emp)
apply clarsimp 
  apply (case_tac "x=xa", clarsimp) apply (simp add: AL_delete_upd_same)
  apply (erule_tac M="delete M x" and N="delete M[xa\<mapsto>y] x" and x=xa and y=y in AL_CONS) 
    apply (simp add: AL_dom_delete)
  apply (erule AL_delete_upd_diff)
done

lemma AL_UPD:"N \<in> AL \<Longrightarrow> N[c\<mapsto>y] \<in> AL"
apply (rule_tac M="delete N c" and x=c and y=y in AL_CONS) 
               apply (erule AL_DEL) 
               apply (simp add: AL_dom_delete) 
               apply (simp add: upd_def) 
               apply (rule AL_delete_idempotent)
done 

lemma AL_dom_set_map_fst: "M:AL \<Longrightarrow> AL_dom M = set (map fst M)"
apply (erule AL.induct)
apply (simp add: AL_dom_def emp_def)
apply clarsimp apply (simp add: AL_dom_def upd_def) 
  apply (rule, clarsimp)
    apply (case_tac "xa=x", clarsimp)
    apply (erule impE) apply fast
    apply clarsimp 
    apply (frule_tac b=xa and l=M in AL_delete2) apply clarsimp
    apply (subgoal_tac "fst ` set (delete M x) = (fst ` set M) - {x}", clarsimp)  
      apply (subgoal_tac "xa : {a. \<exists>y. M\<down>a = Some y}", clarsimp) apply fast 
    apply (rule map_fst_set_delete) 
  apply clarsimp
    apply (case_tac "a : fst ` set (delete M x)")
    prefer 2 apply (erule set_map_fst_set) apply assumption
    apply (simp add: map_fst_set_delete)
    apply (case_tac "a:{a. \<exists>y. M\<down>a = Some y}")
      apply (subgoal_tac "\<exists> y . M\<down>a = Some y", clarsimp) apply (rule, erule AL_delete5b) apply assumption
      apply fast
    apply clarsimp 
done

lemma AssList_AL: "\<exists> M .  (\<forall> x y .(N\<down>x = Some y) = (M\<down>x = Some y)) \<and> M:AL"
apply (induct N)
apply (rule_tac x="[]" in exI, simp add: AL_emp1) apply (rule AL_emp)
apply clarsimp
  apply (case_tac "\<exists> b . N\<down>a = Some b", clarsimp)
  apply (rule_tac x="(delete M a)[a\<mapsto>b]" in exI)
    apply (rule, clarsimp)
      apply (rule, clarsimp) apply (simp add: AL_update1)
      apply clarsimp apply (simp add: AL_update2) apply (drule_tac l=M in AL_delete2a) apply clarsimp
    apply (rule_tac M="delete M a" and x=a in AL_CONS) apply (erule AL_DEL) apply (simp add: AL_dom_delete) apply simp
  apply (drule_tac x=a and y=b in AL_CONS) apply (simp add: AL_dom_def) apply simp 
    apply (rule_tac x="M[a\<mapsto>b]" in exI, clarsimp)
    apply (rule, clarsimp) apply (simp add: AL_update1)
    apply clarsimp apply (simp add: AL_update2)
done

lemma AL_no_junk:"M:AL \<Longrightarrow> (\<forall> x y . (M\<down>x = Some y) = (x,y) mem M)"
apply (erule AL.induct)
apply (simp add: emp_def)
apply clarsimp
    apply rule apply (erule AL_lookup1) 
    apply (simp add: upd_def)
    apply (rule, clarsimp) apply (case_tac "y=ya", clarsimp) apply clarsimp apply (erule mem_delete1)
    apply clarsimp apply (drule mem_delete2) apply assumption
                   apply (erule_tac x=xa in allE, erule_tac x=ya in allE, clarsimp)
                   apply (erule AL_delete5b) apply assumption
done

lemma AL_AssList:"M:AL \<Longrightarrow> \<exists> N .  (\<forall> x y .(N\<down>x = Some y) = (x,y) mem M)"
apply (rule_tac x=M in exI)
apply (erule AL_no_junk)
done

lemma AL_cod_set_map_snd: "M:AL \<Longrightarrow> AL_cod M = set (map snd M)"
apply (drule AL_no_junk)
apply (simp add: AL_cod_def)
apply (rule, clarsimp) 
  apply (erule_tac x=a in allE, erule_tac x=x in allE, clarsimp)
  apply (erule mem_map_snd_set)
apply clarsimp
  apply (simp add: mem_iff) apply (rule, assumption)
done

lemma AL_sum_exists_AL:"AL_dom N \<inter> AL_dom M = {} \<Longrightarrow> \<exists> NM . AL_sum N M NM \<and> NM:AL"
apply (drule AL_sum_exists) apply (erule exE)
apply (subgoal_tac "\<exists> M .  (\<forall> z y .(x\<down>z = Some y) = (M\<down>z = Some y)) \<and> M:AL", erule exE)
  apply (rule_tac x=Ma in exI, clarsimp)
  apply (simp add: AL_sum_def, clarsimp)
  apply (rule, simp add: AL_dom_def)
  apply (simp (no_asm) add: contained_def, rule)
  apply clarsimp apply (drule contained3, assumption) apply fast
  apply clarsimp apply (drule contained3, assumption) apply fast
apply (rule AssList_AL)
done

lemma contained_extensional:
  "\<lbrakk>contained N M; \<forall>x y. (N\<down>x = Some y) = (NN\<down>x = Some y);
        \<forall>x y. (M\<down>x = Some y) = (MM\<down>x = Some y)\<rbrakk>
  \<Longrightarrow> contained NN MM"
by (simp add: contained_def)

lemma AL_contained_sum_existsAux[rule_format]:
"N:AL \<Longrightarrow> M:AL \<longrightarrow> contained N M \<longrightarrow> 
(\<exists> L . AL_sum N L M \<and> L:AL \<and> AL_dom L= AL_dom M - AL_dom N)"
apply (erule AL.induct)
apply clarsimp
  apply (erule AL.elims, clarsimp)
    apply (rule_tac x=emp in exI) apply (rule, rule AL_sum_emp2) apply (rule, rule AL_EMP) apply (simp add: AL_dom_emp)
  apply (simp add: contained_def, clarsimp) apply (rule_tac x="M[x\<mapsto>y]" in exI)
    apply (rule, rule AL_sum_emp1) apply (rule, rule AL_CONS) prefer 3 apply fastsimp apply assumption apply assumption
    apply (simp add: AL_dom_emp)
apply clarsimp
  apply (erule impE) apply (simp add: contained_def, clarsimp) apply (erule_tac x=a in allE, erule_tac x=b in allE, clarsimp) 
    apply (case_tac "a=x", simp add: AL_update1) apply clarsimp apply (drule AL_domI) apply fastsimp
      apply (subgoal_tac "Ma[x\<mapsto>y]\<down>a = Ma\<down>a", clarsimp) apply (erule AL_update2a)
  apply clarsimp
    apply (subgoal_tac "L\<down>x = Some y")
      apply (rule_tac x="delete L x" in exI)
      apply rule apply (simp add: AL_sum_def, clarsimp)
                 apply (rule, simp add: AL_dom_upd1 AL_dom_delete)
                   apply (drule AL_domI) apply fast
                 apply (simp (no_asm) add: contained_def, clarsimp)
                   apply (erule_tac N=L in contained3) apply (drule AL_delete6) apply simp
      apply (rule, erule AL_DEL)
       apply (simp add: AL_dom_upd1 AL_dom_delete) apply fast
    apply (drule contained3) apply (rule AL_update1)
      apply (subgoal_tac "x:AL_dom L")
       apply (drule AL_domD, clarsimp) apply (drule AL_sum4, assumption) apply clarsimp
      apply (frule AL_domI) apply (simp add: AL_sum_def) 
done

lemma AL_contained_sum_exists[rule_format]:
"contained N M \<Longrightarrow>
\<exists> L . AL_sum N L M \<and> L:AL \<and> AL_dom L= AL_dom M - AL_dom N"
apply (subgoal_tac "\<exists> NN .  (\<forall> x y .(N\<down>x = Some y) = (NN\<down>x = Some y)) \<and> NN:AL")
prefer 2 apply (rule AssList_AL)
apply (subgoal_tac "\<exists> MM .  (\<forall> x y .(M\<down>x = Some y) = (MM\<down>x = Some y)) \<and> MM:AL")
prefer 2 apply (rule AssList_AL)
apply (erule exE)+ apply (erule conjE)+
apply (frule contained_extensional, assumption, assumption)
apply (drule AL_contained_sum_existsAux, assumption, assumption)
apply clarsimp
apply (rule_tac x=L in exI, simp)
apply (subgoal_tac "AL_dom M = AL_dom MM")
apply (subgoal_tac "AL_dom N = AL_dom NN")
apply (simp add: AL_sum_def)
apply (simp add: contained_def)
apply (simp add: AL_dom_def)
apply (simp add: AL_dom_def)
done

lemma AL_sum_upd_sum_extensional:
"\<lbrakk>AL_sum N Nx N0; AL_sum N[c\<mapsto>y] Nx N1; c \<notin> AL_dom Nx\<rbrakk> \<Longrightarrow> (N0[c\<mapsto>y]\<down>x = Some d) = (N1\<down>x = Some d)"
               apply rule apply (drule AL_upd_split)
                          apply (erule disjE, clarsimp) 
                            apply (erule AL_sum3) apply (rule AL_update1)
                          apply clarsimp 
                            apply (drule AL_sum_split, assumption)
                            apply (erule disjE) 
                              apply (erule AL_sum3) apply (erule AL_update5) apply assumption
                              apply (erule AL_sum4, assumption) 
               apply (drule AL_sum_split, assumption)
                 apply (erule disjE) 
                   apply (drule AL_upd_split)
                   apply (erule disjE, clarsimp) apply (rule AL_update1)
                   apply clarsimp apply (rule AL_update5) apply (erule AL_sum3, assumption) apply assumption
                 apply (rule AL_update5) apply (erule AL_sum4, assumption) apply (rotate_tac -1, drule AL_domI) apply fast 
done

consts AL_X::"('a,'b) AssList set"
inductive AL_X intros
AL_X_EMP: "emp : AL_X"
AL_X_CONS: "\<lbrakk>M : AL_X; N=M[x\<mapsto>y]\<rbrakk> \<Longrightarrow> N :AL_X"

lemma AL_X_emp:"[]:AL_X"
apply (subgoal_tac "emp:AL_X", simp add: emp_def) apply (rule AL_X_EMP) done

lemma AL_X_DEL:"M : AL_X \<Longrightarrow> delete M x :AL_X"
apply (erule AL_X.induct)
apply (simp add: emp_def) apply (rule AL_X_emp)
apply clarsimp 
  apply (case_tac "x=xa", clarsimp) apply (simp add: AL_delete_upd_same)
  apply (erule_tac M="delete M x" and N="delete M[xa\<mapsto>y] x" and x=xa and y=y in AL_X_CONS) 
  apply (erule AL_delete_upd_diff)
done

lemma AL_X_dom_set_map_fst: "M:AL_X \<Longrightarrow> AL_dom M = set (map fst M)"
apply (erule AL_X.induct)
apply (simp add: AL_dom_def emp_def)
apply clarsimp apply (simp add: AL_dom_def upd_def) 
  apply (rule, clarsimp)
    apply (case_tac "xa=x", clarsimp)
    apply (erule impE) apply fast
    apply clarsimp 
    apply (frule_tac b=xa and l=M in AL_delete2) apply clarsimp
    apply (subgoal_tac "fst ` set (delete M x) = (fst ` set M) - {x}", clarsimp)  
      apply (subgoal_tac "xa : {a. \<exists>y. M\<down>a = Some y}", clarsimp) apply fast 
    apply (rule map_fst_set_delete) 
  apply clarsimp
    apply (case_tac "a : fst ` set (delete M x)")
    prefer 2 apply (erule set_map_fst_set) apply assumption
    apply (simp add: map_fst_set_delete)
    apply (case_tac "a:{a. \<exists>y. M\<down>a = Some y}")
      apply (subgoal_tac "\<exists> y . M\<down>a = Some y", clarsimp) apply (rule, erule AL_delete5b) apply assumption
      apply fast
    apply clarsimp 
done

lemma AssList_AL_X: "\<exists> M .  (\<forall> x y .(N\<down>x = Some y) = (M\<down>x = Some y)) \<and> M:AL_X"
apply (induct N)
apply (rule_tac x="[]" in exI, simp add: AL_emp1) apply (rule AL_X_emp)
apply clarsimp
  apply (case_tac "\<exists> b . N\<down>a = Some b", clarsimp)
  apply (rule_tac x="M[a\<mapsto>b]" in exI)
    apply (rule, clarsimp)
      apply (rule, clarsimp) apply (simp add: AL_update1)
      apply clarsimp apply (simp add: AL_update2)
    apply (erule AL_X_CONS) apply simp
  apply (drule_tac x=a and y=b in AL_X_CONS) apply simp
    apply (rule_tac x="M[a\<mapsto>b]" in exI, clarsimp)
    apply (rule, clarsimp) apply (simp add: AL_update1)
    apply clarsimp apply (simp add: AL_update2)
done

lemma AL_X_no_junk:"M:AL_X \<Longrightarrow> (\<forall> x y . (M\<down>x = Some y) = (x,y) mem M)"
apply (erule AL_X.induct)
apply (simp add: emp_def)
apply clarsimp
    apply rule apply (erule AL_lookup1) 
    apply (simp add: upd_def)
    apply (rule, clarsimp) apply (case_tac "y=ya", clarsimp) apply clarsimp apply (erule mem_delete1)
    apply clarsimp apply (drule mem_delete2) apply assumption
                   apply (erule_tac x=xa in allE, erule_tac x=ya in allE, clarsimp)
                   apply (erule AL_delete5b) apply assumption
done

lemma AL_X_AssList:"M:AL_X \<Longrightarrow> \<exists> N .  (\<forall> x y .(N\<down>x = Some y) = (x,y) mem M)"
apply (rule_tac x=M in exI)
apply (erule AL_X_no_junk)
done

lemma AL_X_cod_set_map_snd: "M:AL_X \<Longrightarrow> AL_cod M = set (map snd M)"
apply (drule AL_X_no_junk)
apply (simp add: AL_cod_def)
apply (rule, clarsimp) 
  apply (erule_tac x=a in allE, erule_tac x=x in allE, clarsimp)
  apply (erule mem_map_snd_set)
apply clarsimp
  apply (simp add: mem_iff) apply (rule, assumption)
done

lemma AL_Size_UpdateSuc1:"\<lbrakk>L\<down>a = None; |L| = n\<rbrakk>\<Longrightarrow> |L[a\<mapsto>b]| = Suc n"
apply (drule AL_Size_UpdateSuc) apply simp done

(*<*)
end
(*>*)
