(*<*)
theory ListSLRep = ListProgs + ExampleBase:
(*>*)

subsubsection{*Simple representation predicates*}

text{*We first define the representation predicates used in the unary
and relational verifications of Sections 5.2 and 5.3, as outlined in
Section 5.1.*}

consts ListHP :: "Colour \<Rightarrow> ((Colour \<times> Colour) list) \<Rightarrow> HP option"
primrec
"ListHP \<gamma> [] = Some(emp[\<gamma>\<mapsto>(NIL,emp)])"
"ListHP \<gamma> (ht # \<Lambda>) = 
    (case ListHP (snd ht) \<Lambda> of
        None \<Rightarrow> None
      | Some H \<Rightarrow> (if \<gamma> \<in> AL_dom H 
                   then None 
                   else Some(H[\<gamma>\<mapsto>(CONS,emp[HD\<mapsto>(fst ht)][TL\<mapsto>(snd ht)])])))"

(*<*)
lemma ListHP_dom1[rule_format]: "\<forall> \<gamma> H . ListHP \<gamma> \<Lambda> = Some H \<longrightarrow> \<gamma> \<in> AL_dom H"
apply (induct \<Lambda>)
apply clarsimp 
  apply (subgoal_tac "\<gamma> : AL_dom (emp[\<gamma>\<mapsto>(NIL, emp)])") apply fast 
  apply (rule AL_dom_upd2)
apply clarsimp
  apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom aa", clarsimp, clarsimp)
  apply (erule_tac x=b in allE, clarsimp)
  apply (rule AL_dom_upd2)
done

lemma ListHP_dom2[rule_format]: "\<forall> \<gamma> H . ListHP \<gamma> \<Lambda> = Some H \<longrightarrow> set (map snd \<Lambda>) \<subseteq> AL_dom H"
apply (induct \<Lambda>)
apply clarsimp 
apply clarsimp
  apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom aa", clarsimp, clarsimp)
  apply rule apply (drule ListHP_dom1) apply (simp add: AL_dom_upd1)
  apply (erule_tac x=b in allE, clarsimp)apply (simp add: AL_dom_upd1) apply (rule disjI2) apply auto
done

lemma ListHP_dom3[rule_format]: "\<forall> \<gamma> H . ListHP \<gamma> \<Lambda> = Some H \<longrightarrow> \<gamma> \<notin> set (map snd \<Lambda>)"
apply (induct \<Lambda>)
apply simp
apply clarsimp
  apply (rule, clarsimp)
    apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
    apply (case_tac "b \<in> AL_dom aa", clarsimp,clarsimp)
    apply (erule_tac x=b in allE, clarsimp)
    apply (drule ListHP_dom1) apply simp
  apply clarsimp
    apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
    apply (case_tac "ba \<in> AL_dom ab", clarsimp, clarsimp)
    apply (erule_tac x=b in allE, clarsimp)
    apply (drule ListHP_dom2) apply fastsimp 
done

lemma ListHP_dom4[rule_format]: "\<forall> \<gamma> H . ListHP \<gamma> \<Lambda> = Some H \<longrightarrow> AL_dom H \<subseteq> {\<gamma>} \<union> set (map snd \<Lambda>)"
apply (induct \<Lambda>)
apply clarsimp apply (simp add: AL_dom_upd1 AL_dom_emp)
apply clarsimp
  apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom aa", clarsimp, clarsimp)
  apply (erule_tac x=b in allE, clarsimp)
  apply (simp add: AL_dom_upd1)
  apply (case_tac "x = \<gamma>", assumption) apply simp
  apply (subgoal_tac "x : insert b (snd ` set \<Lambda>)", clarsimp) apply fast
done

lemma ListHP_dom:"ListHP \<gamma> \<Lambda> = Some H \<Longrightarrow> AL_dom H = {\<gamma>} \<union> set (map snd \<Lambda>)"
apply rule apply (erule ListHP_dom4)
apply safe apply (erule ListHP_dom1)
  apply (drule ListHP_dom2) apply fast 
done

lemma ListHP_AL[rule_format]: "\<forall> \<gamma> H . ListHP \<gamma> \<Lambda> = Some H \<longrightarrow> H:AL"
apply (induct \<Lambda>)
apply clarsimp apply (rule AL_UPD) apply (rule AL_EMP)
apply clarsimp
  apply (case_tac "ListHP b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom aa", clarsimp) apply clarsimp
  apply (erule_tac x=b in allE, clarsimp)
  apply (erule AL_UPD)
done
(*>*)

consts ListN :: "Colour \<Rightarrow> ((Colour \<times> Colour) list) \<Rightarrow> MP option"
primrec
"ListN \<gamma> [] = Some(emp[\<gamma>\<mapsto>(classTP NIL)])"
"ListN \<gamma> (ht # \<Lambda>) = 
    (case ListN (snd ht) \<Lambda> of 
        None \<Rightarrow> None
      | Some N \<Rightarrow> (if \<gamma> \<in> AL_dom N then None 
                   else Some(N[\<gamma>\<mapsto>(classTP CONS)])))"

(*<*)
lemma ListN_dom1[rule_format]: "\<forall> \<gamma> N . ListN \<gamma> \<Lambda> = Some N \<longrightarrow> \<gamma> \<in> AL_dom N"
apply (induct \<Lambda>)
apply clarsimp 
  apply (rule AL_dom_upd2)
apply clarsimp
  apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom a", clarsimp, clarsimp)
  apply (erule_tac x=b in allE, clarsimp)
  apply (rule AL_dom_upd2)
done

lemma ListN_dom2[rule_format]: "\<forall> \<gamma> N . ListN \<gamma> \<Lambda> = Some N \<longrightarrow> set (map snd \<Lambda>) \<subseteq> AL_dom N"
apply (induct \<Lambda>)
apply clarsimp 
apply clarsimp
  apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom a", clarsimp, clarsimp)
  apply rule apply (drule ListN_dom1) apply (simp add: AL_dom_upd1)
  apply (erule_tac x=b in allE, clarsimp) apply (simp add: AL_dom_upd1) apply (rule disjI2) apply auto
done

lemma ListN_dom3[rule_format]: "\<forall> \<gamma> N . ListN \<gamma> \<Lambda> = Some N \<longrightarrow> \<gamma> \<notin> set (map snd \<Lambda>)"
apply (induct \<Lambda>)
apply simp
apply clarsimp
  apply (rule, clarsimp)
    apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
    apply (case_tac "b \<in> AL_dom a", clarsimp,clarsimp)
    apply (erule_tac x=b in allE, clarsimp)
    apply (drule ListN_dom1) apply simp
  apply clarsimp
    apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
    apply (case_tac "ba \<in> AL_dom aa", clarsimp, clarsimp)
    apply (erule_tac x=b in allE, clarsimp)
    apply (drule ListN_dom2) apply fastsimp 
done

lemma ListN_dom4[rule_format]: "\<forall> \<gamma> N . ListN \<gamma> \<Lambda> = Some N \<longrightarrow> AL_dom N \<subseteq> {\<gamma>} \<union> set (map snd \<Lambda>)"
apply (induct \<Lambda>)
apply clarsimp apply (simp add: AL_dom_upd1 AL_dom_emp)
apply clarsimp
  apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom a", clarsimp, clarsimp)
  apply (erule_tac x=b in allE, clarsimp)
  apply (simp add: AL_dom_upd1)
  apply (case_tac "x = \<gamma>", assumption) apply simp
  apply (subgoal_tac "x : insert b (snd ` set \<Lambda>)", clarsimp) apply fast
done

lemma ListN_dom:"ListN \<gamma> \<Lambda> = Some N \<Longrightarrow> AL_dom N = {\<gamma>} \<union> set (map snd \<Lambda>)"
apply rule apply (erule ListN_dom4)
apply safe apply (erule ListN_dom1)
  apply (drule ListN_dom2) apply fast 
done

lemma ListN_AL[rule_format]: "\<forall> \<gamma> H . ListN \<gamma> \<Lambda> = Some H \<longrightarrow> H:AL"
apply (induct \<Lambda>)
apply clarsimp apply (rule AL_UPD) apply (rule AL_EMP)
apply clarsimp
  apply (case_tac "ListN b \<Lambda>", clarsimp, clarsimp)
  apply (case_tac "\<gamma> \<in> AL_dom a", clarsimp, clarsimp)
  apply (erule_tac x=b in allE, clarsimp)
  apply (erule AL_UPD)
done
(*>*)

constdefs preS::"Colour \<Rightarrow> ((Colour \<times> Colour) list) \<Rightarrow> STATE option"
"preS \<gamma> \<Lambda> \<equiv> case ListHP \<gamma> \<Lambda> of None \<Rightarrow> None
                            | Some H \<Rightarrow> Some ([],emp[this\<mapsto>\<gamma>],H)"

(*<*)
lemma starListHP_dom1: "starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<delta> \<Upsilon>) H \<Longrightarrow> ({\<gamma>} \<union> set (map snd \<Lambda>)) \<inter> ({\<delta>} \<union> set (map snd \<Upsilon>)) = {}"
apply (simp add: starOptDEF) 
apply (erule exE, (erule conjE)+)
apply (erule exE, (erule conjE)+)
apply (drule ListHP_dom)
apply (drule ListHP_dom)
apply clarsimp
done

lemma starListHP_dom2: "starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<delta> \<Upsilon>) H \<Longrightarrow> 
AL_dom H = ({\<gamma>} \<union> set (map snd \<Lambda>)) \<union> ({\<delta>} \<union> set (map snd \<Upsilon>))"
apply (simp add: starOptDEF) 
apply (erule exE, (erule conjE)+)
apply (erule exE, (erule conjE)+)
apply (drule ListHP_dom)
apply (drule ListHP_dom)
apply clarsimp
apply (simp add: AL_sum_def)
done

lemma starListN_dom1: "starOpt (ListN \<gamma> \<Lambda>) (ListN \<delta> \<Upsilon>) N \<Longrightarrow> ({\<gamma>} \<union> set (map snd \<Lambda>)) \<inter> ({\<delta>} \<union> set (map snd \<Upsilon>)) = {}"
apply (simp add: starOptDEF) 
apply (erule exE, (erule conjE)+)
apply (erule exE, (erule conjE)+)
apply (drule ListN_dom)
apply (drule ListN_dom)
apply clarsimp
done

lemma starListN_dom2: "starOpt (ListN \<gamma> \<Lambda>) (ListN \<delta> \<Upsilon>) N \<Longrightarrow> 
AL_dom N = ({\<gamma>} \<union> set (map snd \<Lambda>)) \<union> ({\<delta>} \<union> set (map snd \<Upsilon>))"
apply (simp add: starOptDEF) 
apply (erule exE, (erule conjE)+)
apply (erule exE, (erule conjE)+)
apply (drule ListN_dom)
apply (drule ListN_dom)
apply clarsimp
apply (simp add: AL_sum_def)
done
(*>*)

consts Spec::"nat \<Rightarrow> (STATE \<times> MP \<times> HP \<times> Colour \<times> MP) set"
primrec
"Spec 0 = { (\<Sigma>,N,H,\<gamma>',M) . \<exists> \<gamma> . preS \<gamma> [] = Some \<Sigma> \<and> ListN \<gamma> [] = Some N \<and>
                                 starOpt (ListHP \<gamma> []) (ListHP \<gamma>' []) H \<and>
                                 starOpt (ListN \<gamma> []) (ListN \<gamma>' []) M}"
"Spec (Suc n) = { (\<Sigma>,N,H,\<gamma>',M) . \<exists> \<gamma> \<Lambda> \<Upsilon>. preS \<gamma> \<Lambda> = Some \<Sigma> \<and> ListN \<gamma> \<Lambda> = Some N \<and>
                                     starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<gamma>' \<Upsilon>) H \<and> 
                                     starOpt (ListN \<gamma> \<Lambda>) (ListN \<gamma>' \<Upsilon>) M \<and>
                                     length \<Lambda> = Suc n \<and> length \<Upsilon> = Suc n \<and>
                                     map fst \<Lambda> = map fst \<Upsilon>}"

constdefs U::"Mbody \<Rightarrow> UniCtxt"
"U mb \<equiv> {(ppt,S) . \<exists> n . ppt = (NilCopy,0) \<and> S \<in> Spec 0 \<or> (n>0 \<and> ppt=(mb,0) \<and> S \<in> Spec n)}"

constdefs Sp::"nat \<Rightarrow> (STATE \<times> MP \<times> HP \<times> Colour \<times> MP) set"
"Sp n \<equiv> { (\<Sigma>,N,H,\<gamma>',M) . \<exists> \<gamma> \<Lambda> \<Upsilon>. preS \<gamma> \<Lambda> = Some \<Sigma> \<and> ListN \<gamma> \<Lambda> = Some N \<and>
                                     starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<gamma>' \<Upsilon>) H \<and> 
                                     starOpt (ListN \<gamma> \<Lambda>) (ListN \<gamma>' \<Upsilon>) M \<and>
                                     length \<Lambda> = n \<and> length \<Upsilon> = n \<and>
                                     map fst \<Lambda> = map fst \<Upsilon>}"

lemma SpecSp:"Spec n = Sp n"
apply (simp add: Sp_def)
apply (case_tac n, clarsimp)
apply clarsimp
done

constdefs V::"Mbody \<Rightarrow> UniCtxt"
"V mb \<equiv> {(ppt,S) . \<exists> n . ppt = (NilCopy,0) \<and> S \<in> Sp 0 \<or> (n>0 \<and> ppt=(mb,0) \<and> S \<in> Sp n)}"

constdefs W::"Mbody \<Rightarrow> UniCtxt"
"W mb \<equiv> {(ppt,S) . ppt = (NilCopy,0) \<and> S \<in> Sp 0} \<union> 
        {(ppt,S) . \<exists> n . n>0 \<and> ppt=(mb,0) \<and> S \<in> Sp n}"

lemma VW: "V=W"
by (rule, simp add: V_def W_def, fastsimp)

lemma UW:"U=W"
apply (rule, simp add: U_def W_def SpecSp)
apply rule apply clarsimp apply fast 
done

constdefs RelSp1::"nat \<Rightarrow> (RSD \<times> RSD) set"
"RelSp1 n \<equiv> { (\<phi>,\<psi>) . \<exists> \<gamma> \<Lambda> \<gamma>' \<Upsilon> H \<Sigma> \<Pi> N M. preS \<gamma> \<Lambda> = Some \<Sigma> \<and> ListN \<gamma> \<Lambda> = Some N \<and>
                                     starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<gamma>' \<Upsilon>) H \<and> 
                                     starOpt (ListN \<gamma> \<Lambda>) (ListN \<gamma>' \<Upsilon>) M \<and>
                                     \<Pi> = ([\<gamma>'],emp,H) \<and> \<phi>=(\<Sigma>,N,\<Sigma>) \<and> \<psi>=(\<Pi>,M,\<Pi>)\<and>
                                     length \<Lambda> = n \<and> length \<Upsilon> = n \<and>
                                     map fst \<Lambda> = map fst \<Upsilon>}"

(*constdefs RelU1::Ctxt
"RelU1 \<equiv> {(ppts,S) . \<exists> n . ppts = ((NilCopy,0),(NilCopy,0)) \<and> S \<in> RelSp1 0 \<or> 
                           (n>0 \<and> ppts=((ConsCopy,0),(ConsOpt,0)) \<and> S \<in> RelSp1 n)}"
lemma "RelU1 = RelV1"
by (simp add: RelU1_def RelV1_def, fast)
*)

constdefs G1::Ctxt
"G1 \<equiv> {(ppts,S) . ppts = ((NilCopy,0),(NilCopy,0)) \<and> S \<in> RelSp1 0} \<union>
         {(ppts,S) . \<exists> n . n>0 \<and> ppts=((ConsCopy,0),(ConsOpt,0)) \<and> S \<in> RelSp1 n}"


constdefs RelSp2::"nat \<Rightarrow> (RSD \<times> RSD) set"
"RelSp2 n \<equiv> { (\<phi>,\<psi>) . \<exists> \<gamma> \<Lambda> \<upsilon> \<Upsilon> \<upsilon>' \<Upsilon>' H K \<Sigma> \<Pi> \<Xi> N M Q. preS \<gamma> \<Lambda> = Some \<Sigma> \<and> ListN \<gamma> \<Lambda> = Some N \<and>
                                     starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<upsilon> \<Upsilon>) H \<and> 
                                     starOpt (ListHP \<gamma> \<Lambda>) (ListHP \<upsilon>' \<Upsilon>') K \<and> 
                                     starOpt (ListN \<gamma> \<Lambda>) (ListN \<upsilon> \<Upsilon>) Q \<and>
                                     starOpt (Some Q) (ListN \<upsilon>' \<Upsilon>') M \<and>
                                     \<Pi> = ([\<upsilon>],emp,H) \<and> \<Xi> = ([\<upsilon>'],emp,K) \<and> \<phi>=(\<Sigma>,N,\<Sigma>) \<and> \<psi>=(\<Pi>,M,\<Xi>)\<and>
                                     length \<Lambda> = n \<and> length \<Upsilon> = n \<and> length \<Upsilon>' = n \<and>
                                     map fst \<Lambda> = map fst \<Upsilon> \<and> map fst \<Lambda> = map fst \<Upsilon>'}"

constdefs G2::Ctxt
"G2 \<equiv> {(ppts,S) . ppts = ((NilCopy,0),(NilCopy,0)) \<and> S \<in> RelSp2 0} \<union>
         {(ppts,S) . \<exists> n . n>0 \<and> ppts=((ConsCopy,0),(ConsOpt,0)) \<and> S \<in> RelSp2 n}"

constdefs G3::Ctxt
"G3 \<equiv> {(ppts,S) . ppts = ((NilCopy,0),(NilCopy,0)) \<and> S \<in> RelSp2 0} \<union>
         {(ppts,S) . \<exists> n . n>0 \<and> ppts=((ConsCopy,0),(Cross,0)) \<and> S \<in> RelSp2 n}"

(*<*)
end
(*>*)
