theory InjFinMap = Trivialities:

constdefs inj_finMap ::"('a ~=> 'b) \<Rightarrow> bool"
"inj_finMap f \<equiv> inj_on f (dom f)"

lemma inj_finMapD1: "\<lbrakk>inj_finMap f; f x = Some a; f y = Some b\<rbrakk> \<Longrightarrow> (a=b) = (x=y)" 
apply (simp add: inj_finMap_def)
apply (rule, clarsimp)
  apply (drule_tac x=x and y=y in inj_onD) apply simp apply (simp add: dom_def) apply (simp add: dom_def) apply assumption
apply simp
done

lemma inj_finMapI1:"x \<noteq> y \<Longrightarrow> (\<exists> f . inj_finMap f \<and> f x = None \<and> f y = None)"
apply (rule_tac x=empty in exI)
apply (simp add: inj_finMap_def)
done

constdefs Invv::"('a ~=> 'b) \<Rightarrow> ('b ~=> 'a)"
"Invv f \<equiv> \<lambda> b . if (\<exists> y . f y = Some b) then Some (SOME y. y \<in> dom f & f y = Some b) else None"

lemma InvvD1: "b \<notin> ran f \<Longrightarrow> Invv f b = None" apply (simp add: ran_def Invv_def) done

lemma Invv_f_f: "[| inj_finMap f;  f x = Some a|] ==> Invv f a = Some x"
apply (simp add: Invv_def inj_finMap_def)
apply rule apply clarsimp apply (rule_tac a=x in someI2) apply (simp add: dom_def)
  apply clarsimp apply (drule_tac x=xa and y=x in inj_onD) apply simp 
                   apply (simp add: dom_def) apply (simp add: dom_def) apply assumption
apply (rule, assumption)
done

lemma f_Invv_f: "a : ran f \<Longrightarrow> \<exists> x . Invv f a = Some x \<and> f x = Some a"
apply (simp add: Invv_def)
apply (rule, clarsimp)
apply (fast intro: someI2)
apply (simp add: ran_def)
done

lemma inj_finMap_Invv:"inj_finMap f \<Longrightarrow> inj_finMap (Invv f)"
apply (simp (no_asm) add: inj_finMap_def inj_on_def, clarsimp)
  apply (case_tac "y:ran f") prefer 2 apply (drule_tac b=y in InvvD1) apply clarsimp
  apply (case_tac "x:ran f") prefer 2 apply (drule_tac b=x in InvvD1) apply clarsimp
  apply (drule ranD) apply (drule ranD) apply clarsimp
  apply (frule_tac x=b in Invv_f_f) apply assumption
  apply (drule_tac x=ba in Invv_f_f) apply assumption
  apply clarsimp
done

lemma inj_finMap_dom_invv_ran: "inj_finMap f \<Longrightarrow> dom (Invv f) = ran f"
apply (rule, clarsimp)
  apply (case_tac "x:ran f", clarsimp)
  apply (drule  InvvD1) apply clarsimp
apply rule apply (drule ranD, clarsimp)
  apply (drule Invv_f_f) apply assumption apply simp
done

lemma inj_finMap_ran_invv_dom: "inj_finMap f \<Longrightarrow> ran (Invv f) = dom f"
apply (rule, clarsimp)
  apply (drule ranD, clarsimp)
  apply (case_tac "x:dom f") apply (erule domD)
  apply (case_tac "b:ran f") apply (drule f_Invv_f) apply clarsimp
  apply (drule InvvD1) apply clarsimp
apply rule apply (drule domD, clarsimp)
  apply (drule Invv_f_f) apply assumption apply (erule ranI)
done

constdefs map_compose::"('b ~=> 'c) \<Rightarrow> ('a ~=> 'b) \<Rightarrow> ('a ~=> 'c)"
"map_compose g f \<equiv> \<lambda> a . (case f a of None \<Rightarrow> None | Some b \<Rightarrow> g b)"

lemma map_composeD1:"f a = Some b \<Longrightarrow> map_compose g f a = g b"
by (simp add: map_compose_def)
lemma map_composeD2:"f a = None \<Longrightarrow> map_compose g f a = None"
by (simp add: map_compose_def)

lemma map_composeD3:"map_compose g f a = Some c \<Longrightarrow> \<exists> b . f a = Some b \<and> g b = Some c"
apply (case_tac "f a", clarsimp)
  apply (simp add: map_compose_def)
  apply (simp add: map_compose_def)
done
lemma map_composeD4:"map_compose g f a = None \<Longrightarrow> f a = None \<or> (\<exists> b . f a = Some b \<and> g b = None)"
apply (case_tac "f a", clarsimp) 
  apply (simp add: map_compose_def)
done

constdefs Img::"('a ~=> 'b) \<Rightarrow> ('a set) \<Rightarrow> ('b set)"
"Img f S \<equiv> {x . \<exists> s . s:S \<and> f s = Some x}"

lemma ImgD:"x : Img f S \<Longrightarrow> \<exists> s . s:S \<and> f s = Some x" by (simp add: Img_def)
lemma ImgI:"\<lbrakk>s:S; f s = Some x\<rbrakk> \<Longrightarrow> x : Img f S" apply (simp add: Img_def) apply fast done

lemma Img_emptyset: "Img f {} = {}" by (simp add: Img_def)

lemma Img_Monotone: "X\<subseteq>Y \<Longrightarrow> Img f X \<subseteq> Img f Y"
apply (simp add: Img_def) apply fast done 

lemma Img_ran:"Img f X \<subseteq> ran f" apply (simp add: Img_def ran_def) apply fast done

lemma Img_dom_ran:"Img f (dom f) = ran f"
apply (simp add: Img_def ran_def dom_def) apply auto 
done

lemma Img_union: "Img f (X\<union>Y) = Img f X \<union> Img f Y" by (simp add: Img_def, auto)

lemma Img_inter: "Img f (X\<inter>Y) \<subseteq> Img f X \<inter> Img f Y" by (simp add: Img_def, auto)

lemma Img_inter_inj_on: "inj_on f (X\<union>Y) \<Longrightarrow> Img f X \<inter> Img f Y = Img f (X\<inter>Y)" 
apply (simp add: Img_def) apply auto
apply (drule_tac x=s and y=sa in inj_onD) apply simp apply simp apply simp apply clarsimp apply fast
done

lemma Fun_eqOn_Img:"\<lbrakk>\<forall> c . c:Y \<longrightarrow> g c= f c; X \<subseteq> Y\<rbrakk> \<Longrightarrow> Img f X = Img g X"
by (simp add: Img_def, fastsimp)

lemma InvvD_Some: "Invv f x = Some y \<Longrightarrow> f y = Some x" 
apply (simp add: Invv_def) 
apply (case_tac "\<exists>y. f y = Some x", clarsimp)
prefer 2 apply clarsimp 
apply (rule_tac a=y and Q="\<lambda> c . f c = Some x" in someI2)
  apply simp apply (erule domI)
  apply simp
done

lemma InvvD_None: "Invv f x = None \<Longrightarrow> \<forall> y . f y \<noteq> Some x"
apply (simp add: Invv_def)
apply (case_tac "\<exists>y. f y = Some x", clarsimp)
apply clarsimp 
done

lemma Invv_map_add_disj:
  "\<lbrakk> dom f \<inter> dom g = {}; inj_on f (dom f); inj_on g (dom g); 
     Img f (dom f) \<inter> Img g (dom g) = {}\<rbrakk> 
  \<Longrightarrow> (Invv (f ++ g)) = (Invv f ++ Invv g)"
apply rule
apply (case_tac "Invv (f ++ g) x", clarsimp)
  apply (drule InvvD_None)
  apply (case_tac "(Invv f ++ Invv g) x")
    apply (simp add: map_add_def)
  apply clarsimp
    apply (erule disjE) apply (drule InvvD_Some) apply (erule_tac x=a in allE, clarsimp)
    apply clarsimp apply (drule InvvD_Some)
      apply (case_tac "g a") apply (erule_tac x=a in allE) apply (simp add: map_add_def)
      apply (drule_tac m=f in domI) apply (drule_tac m=g in domI) apply fast
apply clarsimp apply (drule InvvD_Some)
  apply (case_tac "g a", clarsimp)
    apply (case_tac "(Invv g) x")        
      apply (simp add: map_add_def)
      apply (case_tac "(Invv f) x") apply (drule_tac f=f in InvvD_None) apply (erule_tac x=a in allE, clarsimp)
      apply clarsimp apply (drule InvvD_Some)
        apply (erule_tac f=f and x=a and y=aa in inj_onD) apply simp apply (erule domI) apply (erule domI) 
      apply (simp add: map_add_def) apply (drule InvvD_Some)
        apply (subgoal_tac "x:Img f (dom f) \<inter> Img g (dom g)", clarsimp) 
        apply (rotate_tac 3, erule thin_rl, simp add: Img_def)  
        apply (rule, rule_tac x=a in exI, simp) apply (erule domI)
        apply (rule_tac x=aa in exI, simp) apply (erule domI)
  apply clarsimp
    apply (case_tac "(Invv g) x") apply (drule InvvD_None) apply (erule_tac x=a in allE, clarsimp)
    apply (frule InvvD_Some) 
    apply (drule_tac f=g and x=a and y=aa in inj_onD) apply simp apply (erule domI) apply (erule domI) apply clarsimp
done

lemma inj_on_Invv_SelfInverse: "inj_on f (dom f) \<Longrightarrow> Invv (Invv f) = f"
apply rule
  apply (case_tac "Invv (Invv f) x")
    apply (case_tac "f x") apply clarsimp
    apply (frule_tac m=f in ranI) apply (drule f_Invv_f) apply clarsimp
    apply (drule_tac x=x and y=xa in inj_onD, simp) apply (erule domI) apply (erule domI) apply clarsimp
    apply (drule InvvD_None) apply (erule_tac x=a in allE, clarsimp)
  apply (frule InvvD_Some) 
  apply (frule_tac f=f in InvvD_Some) apply clarsimp
done

lemma Invv_map_add_disj_Invv:
"\<lbrakk>dom f \<inter> dom g = {}; inj_on f (dom f); inj_on g (dom g); Img f (dom f) \<inter> Img g (dom g) = {}\<rbrakk> 
\<Longrightarrow> Invv ((Invv f ++ Invv g)) = f ++ g"
apply (subgoal_tac "Img f (dom f) = ran f") prefer 2 apply (simp add: Img_def dom_def ran_def) apply fast
apply (subgoal_tac "Img g (dom g) = ran g") prefer 2 apply (simp add: Img_def dom_def ran_def) apply fast
apply clarsimp
apply (subgoal_tac "(Invv (Invv f)) ++ (Invv (Invv g)) = f ++ g")
  apply (subgoal_tac "Invv ((Invv f ++ Invv g)) = (Invv (Invv f)) ++ (Invv (Invv g))", clarsimp)
  apply (rule Invv_map_add_disj) 
    apply (subgoal_tac "dom (Invv f) = ran f") 
      apply (subgoal_tac "dom (Invv g) = ran g") 
      apply clarsimp
    apply (rule inj_finMap_dom_invv_ran) apply (simp add: inj_finMap_def)
  apply (rule inj_finMap_dom_invv_ran) apply (simp add: inj_finMap_def)
apply (subgoal_tac "inj_finMap (Invv f)") apply (simp add: inj_finMap_def) 
  apply (rule inj_finMap_Invv) apply (simp add: inj_finMap_def) 
apply (subgoal_tac "inj_finMap (Invv g)") apply (simp add: inj_finMap_def) 
  apply (rule inj_finMap_Invv) apply (simp add: inj_finMap_def) 
apply (subgoal_tac "Img (Invv f) (dom (Invv f)) = ran (Invv f)") prefer 2 apply (simp add: Img_def dom_def ran_def) apply fast
  apply (subgoal_tac "Img (Invv g) (dom (Invv g)) = ran (Invv g)") prefer 2 apply (simp add: Img_def dom_def ran_def) apply fast
  apply clarsimp
  apply (subgoal_tac "ran (Invv f) = dom f")
    apply (subgoal_tac "ran (Invv g) = dom g")
      apply clarsimp
    apply (rule inj_finMap_ran_invv_dom) apply (simp add: inj_finMap_def)
  apply (rule inj_finMap_ran_invv_dom) apply (simp add: inj_finMap_def) 
apply (subgoal_tac "Invv (Invv f) = f")
  apply (subgoal_tac "Invv (Invv g) = g")
     apply clarsimp
  apply (erule inj_on_Invv_SelfInverse) 
apply (erule inj_on_Invv_SelfInverse) 
done

end