theory Ren = AbstrState + InjFinMap:

types Mapping = "(Colour ~=> Colour)"

(*the penultimate condition enforces that colours that are merged are of identical type*)
consts MapN::"(MP \<times> Mapping \<times> MP) set"
inductive MapN intros
MapN_emp:"(emp, f, emp) : MapN"
MapN_cons:"\<lbrakk> N = M[x\<mapsto>y]; x\<notin> AL_dom M; (M,f,fM):MapN; f x = Some fx; 
            \<forall> z . fM\<down>fx = Some z \<longrightarrow> z=y; fN=fM[fx\<mapsto>y]\<rbrakk>
           \<Longrightarrow> (N,f,fN): MapN"

lemma MapN_AL_Aux1:"(H, f, K) \<in> MapN \<Longrightarrow> K:AL"
apply (erule MapN.induct)
apply clarsimp apply (rule AL_EMP)
apply clarsimp apply (rule_tac M="delete fM fx" and x=fx and y=y in AL_CONS) 
                 apply (erule AL_DEL) 
                 apply (simp add: AL_dom_delete)
                 apply (simp add: upd_def) apply (rule  AL_delete_idempotent)
done

lemma MapN_AL_Aux2[rule_format]:"(H, f, K) \<in> MapN \<Longrightarrow> K:AL \<longrightarrow> H:AL"
apply (erule MapN.induct)
apply clarsimp
apply clarsimp
  apply (drule MapN_AL_Aux1)
  apply clarsimp
  apply (rule_tac M="delete M x" and x=x and y=y in AL_CONS) apply (erule AL_DEL)
                 apply (simp add: AL_dom_delete)
                 apply (simp add: upd_def) apply (rule  AL_delete_idempotent)
done

lemma MapN_AL_Aux3:"(H, f, K) \<in> MapN \<Longrightarrow> H:AL \<and> K:AL"
apply (frule MapN_AL_Aux1) apply simp
apply (erule MapN_AL_Aux2, assumption)
done

lemma MapN_AL_Aux4[rule_format]:"(N,f,M):MapN \<Longrightarrow>  AL_dom M = Img f (AL_dom N)"
apply (erule MapN.induct)
apply (simp add: Img_emptyset emp_def AL_dom_def)
apply clarsimp
  apply (simp add: AL_dom_upd1)
  apply (simp add: Img_def)
  apply (rule, clarsimp) 
    apply rule apply (rule_tac x=x in exI, simp)
    apply (rule, clarsimp) apply (rule_tac x=s in exI, simp)
  apply (rule, clarsimp) 
    apply (erule disjE, clarsimp)apply (erule_tac x=s in allE, clarsimp)
done

lemma MapN_AL[rule_format]:"(N,f,M):MapN \<Longrightarrow> N:AL \<and> M:AL \<and> AL_dom M = Img f (AL_dom N)"
apply (frule MapN_AL_Aux3, simp)
apply (frule MapN_AL_Aux4, simp)
done

lemma MapN_AL_delete[rule_format]:"(N,f,M):MapN \<Longrightarrow> inj_on f (dom f) \<longrightarrow> (\<forall> x fx. f x = Some fx \<longrightarrow> (delete N x,f,delete M fx):MapN)"
apply (erule MapN.induct)
apply clarsimp
  apply (subgoal_tac "(emp::MP) = delete (emp::MP) x", clarsimp)
  apply (subgoal_tac "(emp::MP) = delete (emp::MP) fx", clarsimp)
    apply (rule MapN_emp)
  apply (simp add: emp_def)
  apply (simp add: emp_def)
apply clarsimp
  apply (case_tac "xa=x", clarsimp)
    apply (simp add: AL_delete_upd_same)
  apply (frule_tac z=xa and M=M and y=y in AL_delete_upd_diff) apply clarsimp
  apply (subgoal_tac "fxa \<noteq> fx")
    apply (frule_tac z=fxa and M=fM and y=y in AL_delete_upd_diff) apply clarsimp
    apply (erule_tac x=xa in allE, clarsimp)
    apply (rule MapN_cons) apply fastsimp apply (simp add: AL_dom_delete) apply assumption apply assumption
       apply clarsimp apply (drule AL_delete6) apply clarsimp
       apply simp
  apply clarsimp
    apply (drule_tac x=x and y=xa in inj_onD) apply simp apply (erule domI) apply (erule domI) apply simp
done

lemma MapN_AL_delete2:"\<lbrakk>(N,f,M):MapN; x \<notin> AL_dom N\<rbrakk> \<Longrightarrow> (delete N x,f,M):MapN"
apply (subgoal_tac "delete N x = N", clarsimp)
apply (erule AL_notdomD_delete) 
done

lemma MapN_dom:"(N,f,M):MapN \<Longrightarrow> AL_dom M = Img f (AL_dom N)"
apply (erule MapN.induct)
  apply (simp add: AL_dom_emp Img_emptyset) 
apply clarsimp
  apply (simp add: Img_def AL_dom_upd1)
  apply (rule, clarsimp)
    apply rule apply (rule_tac x=x in exI, simp)
    apply clarsimp  apply (rule_tac x=s in exI, simp)
  apply clarsimp 
    apply (erule disjE, clarsimp)
    apply (rule_tac x=s in exI, simp)
done

lemma MapN_D_right[rule_format]:"(N,f,M):MapN \<Longrightarrow> M\<down>fc = Some x \<longrightarrow> (\<exists> c . f c = Some fc \<and> N\<down>c = Some x)"
apply (erule MapN.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (rotate_tac -2, erule thin_rl)
  apply (drule AL_upd_split)
  apply (erule disjE)  apply clarsimp
    apply (rule_tac x=xa in exI, simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "xa=c", clarsimp)
    apply (rule_tac x=c in exI, simp)
    apply (erule AL_update5) apply assumption
done

lemma MapN_D_left[rule_format]:"(N,f,M):MapN \<Longrightarrow> N\<down>c = Some x \<longrightarrow> (\<exists> fc . f c = Some fc \<and> M\<down>fc = Some x)"
apply (erule MapN.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE) 
    apply (rotate_tac 2, erule thin_rl) apply clarsimp apply (rule AL_update1) 
  apply (erule conjE)
    apply (erule impE, assumption)
    apply (erule exE, erule conjE) 
    apply (rule_tac x=fc in exI, simp)
    apply (case_tac "fx=fc", clarsimp) apply (simp add: AL_update1)
    apply (erule AL_update5) apply assumption
done

(*in version 12, we had only proven the following weaker result:*)
lemma MapN_D_left1[rule_format]:"(N,f,M):MapN \<Longrightarrow> inj_on f (AL_dom N) \<longrightarrow> N\<down>c = Some x \<longrightarrow> (\<exists> fc . f c = Some fc \<and> M\<down>fc = Some x)"
apply (erule MapN.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (rotate_tac -3, erule thin_rl)
  apply (erule impE) apply (erule inj_on_subset) apply (rule AL_dom_upd)
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "fx=fc", clarsimp)
    prefer 2 apply (erule AL_update5) apply assumption
    apply (drule_tac x=xa and y=c in inj_onD) apply simp apply (rule AL_dom_upd2)
      apply (drule AL_domI) apply (simp add: AL_dom_upd1)
      apply clarsimp
done

lemma MapN_D_left2[rule_format]:"(N,f,M):MapN \<Longrightarrow> inj_on f (dom f) \<longrightarrow> N\<down>c = Some x \<longrightarrow> (\<exists> fc . f c = Some fc \<and> M\<down>fc = Some x)"
apply (erule MapN.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (rotate_tac -4, erule thin_rl)
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "fx=fc", clarsimp)
    prefer 2 apply (erule AL_update5) apply assumption
    apply (drule_tac x=xa and y=c in inj_onD) apply simp apply (erule domI) apply (erule domI)
      apply clarsimp
done

lemma MapN_D_left3[rule_format]:"(N,f,M):MapN \<Longrightarrow> \<forall> x . x:AL_dom N \<longrightarrow> (\<exists> fx . f x = Some fx \<and> fx:AL_dom M)"
apply (erule MapN.induct)
  apply clarsimp
  apply (simp add: AL_dom_emp) 
apply clarsimp
  apply (rotate_tac -2, erule thin_rl)
  apply (simp add: AL_dom_upd1)
  apply (erule disjE, clarsimp) 
  apply (erule_tac x=xa in allE, clarsimp)
done

constdefs TypePreserving::"Mapping \<Rightarrow> MP \<Rightarrow> MP \<Rightarrow> bool"
"TypePreserving f N M \<equiv> \<forall> c x fc . N\<down>c = Some x \<longrightarrow> f c = Some fc \<longrightarrow> M\<down>fc = Some x"

lemma TypePreserving_D:"\<lbrakk> TypePreserving f N M; N\<down>c = Some x; f c = Some fc\<rbrakk> \<Longrightarrow> M\<down>fc = Some x"
apply (unfold TypePreserving_def) apply (erule_tac x=c in allE, clarsimp)
done

lemma MapN_TypePreserving: "(N,f,M):MapN \<Longrightarrow> TypePreserving f N M"
apply (erule MapN.induct)
apply (simp add: TypePreserving_def, clarsimp)
  apply (simp add: AL_emp1)
apply clarsimp
apply (simp (no_asm) add: TypePreserving_def, clarsimp)
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) apply (rule AL_update1)
  apply clarsimp
  apply (drule TypePreserving_D) apply assumption apply assumption 
  apply (case_tac "fx=fc", clarsimp) apply (rule AL_update1)
  apply (erule AL_update5) apply assumption
done

lemma MapN_D_left4[rule_format]:
  "(N,f,M):MapN \<Longrightarrow> inj_on f (ObjIDs N) \<longrightarrow> 
                (\<forall> x C . N\<down>x = Some (classTP C) \<longrightarrow> (\<exists> fx . f x = Some fx \<and> M\<down>fx = Some(classTP C)))"
apply (erule MapN.induct)
  apply clarsimp
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (simp add: AL_update1)
  apply clarsimp
    apply (erule impE) apply (erule inj_on_subset) apply (simp add: ObjIDs_def) apply clarsimp 
      apply (case_tac "x=xb", clarsimp) apply (drule_tac d=xb in AL_domI) apply fast
      apply (rule, erule AL_update5) apply assumption
    apply (erule_tac x=xa in allE, clarsimp)
    apply (case_tac "fx=fxa", clarsimp) apply (rule AL_update1)
    apply (erule AL_update5) apply assumption
done
(*
lemma MapN_exists[rule_format]:
"N:AL \<Longrightarrow> AL_dom N \<subseteq> dom f \<longrightarrow> (\<exists> M . (N,f,M):MapN)"
apply (erule AL.induct)
apply clarsimp
  apply (rule_tac x=emp in exI)  apply (rule MapN_emp)
apply clarsimp
  apply (erule impE) apply (simp add: AL_dom_upd1)
  apply clarsimp
    apply (subgoal_tac "x:dom f")
      apply (drule domD, clarsimp)
      apply (rule, rule MapN_cons) 
        apply fastsimp
        apply assumption
        apply assumption
        apply assumption
        apply simp
    apply (simp add: AL_dom_upd1)
done

lemma MapN_exists_with_Props:
  "\<lbrakk>N:AL; AL_dom N \<subseteq> dom f; inj_on f (AL_dom N)\<rbrakk>
 \<Longrightarrow> \<exists> M . (N,f,M):MapN \<and> M:AL \<and>
           (\<forall> c x. N\<down>c = Some x \<longrightarrow> (\<exists> fc . f c = Some fc \<and> M\<down>fc = Some x)) \<and>
           (\<forall> fc x . M\<down>fc = Some x \<longrightarrow> (\<exists> c . f c = Some fc \<and> N\<down>c = Some x))"
apply (frule_tac f=f in MapN_exists)  apply assumption apply (erule exE)
apply (rule, rule, assumption)
apply (frule MapN_AL, clarsimp)
apply (rule, clarsimp) apply (drule MapN_D_left) apply assumption apply assumption apply assumption
apply clarsimp apply (drule MapN_D_right) apply assumption apply assumption 
done
*)

lemma MapN_unique[rule_format]:
  "\<forall> T . (S,f,T):MapN \<longrightarrow> (\<forall> R . (S,f,R):MapN \<longrightarrow> R=T)"
apply (induct S)
apply clarsimp
  apply (erule MapN.elims, clarsimp)
  apply (erule MapN.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MapN.elims, simp add: upd_def emp_def) apply (simp add: upd_def emp_def) 
apply clarsimp
  apply (erule MapN.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MapN.elims, simp add: upd_def emp_def)
     apply (simp add: upd_def emp_def) apply clarsimp
     apply (frule_tac N=M in AL_notdomD_delete)
     apply (frule_tac N=Ma in AL_notdomD_delete) apply clarsimp
     apply (erule_tac x=fMa in allE, clarsimp)
done

consts Map_Fields::"((Field,Colour) AssList \<times> Mapping \<times> (Field,Colour) AssList)set"
inductive Map_Fields intros
Map_Fields_emp:"(emp,f,emp):Map_Fields"
Map_Fields_cons:"\<lbrakk>GG=FF[fld\<mapsto>c]; fld \<notin> AL_dom FF; (FF,f,F):Map_Fields; f c = Some fc; G=F[fld\<mapsto>fc]\<rbrakk> \<Longrightarrow> (GG,f,G):Map_Fields"

lemma Map_Fields_dom:
"(G,f,F):Map_Fields \<Longrightarrow> AL_dom F=AL_dom G"
apply (erule Map_Fields.induct)
  apply (simp add: AL_dom_emp)
  apply (simp add: AL_dom_upd1)
done

lemma Map_Fields_AL[rule_format]:"(N,f,M):Map_Fields \<Longrightarrow> (N:AL \<and> M:AL)"
apply (erule Map_Fields.induct)
  apply (rule, rule AL_EMP, rule AL_EMP) 
apply clarsimp
apply rule
  apply (rule_tac M="delete FF fld" and x=fld and y=c in AL_CONS)
    apply (erule AL_DEL)
               apply (simp add: AL_dom_delete) 
               apply (simp add: upd_def) 
               apply (rule AL_delete_idempotent)
apply (rule_tac M="delete F fld" and x=fld and y=fc in AL_CONS)
    apply (erule AL_DEL)
               apply (simp add: AL_dom_delete) 
               apply (simp add: upd_def) 
               apply (rule AL_delete_idempotent)
done

lemma Map_Fields_D_right[rule_format]:"(N,f,M):Map_Fields \<Longrightarrow> M\<down>fld = Some fc \<longrightarrow> (\<exists> c . f c = Some fc \<and> N\<down>fld = Some c)"
apply (erule Map_Fields.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (rule_tac x=c in exI, simp add: AL_update1)
  apply clarsimp
    apply (rule_tac x=ca in exI, simp)
    apply (erule AL_update5) apply assumption
done

lemma Map_Fields_D_left[rule_format]: 
  "(s,f,t):Map_Fields \<Longrightarrow> (\<forall> x c . s\<down>x = Some c \<longrightarrow> (\<exists> fc . f c = Some fc \<and> t\<down>x = Some fc))"
apply (erule Map_Fields.induct)
apply (simp add: AL_emp1)
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp)
    apply (simp add: AL_update1)
  apply clarsimp
    apply (erule_tac x=x in allE, clarsimp)
    apply (erule AL_update5)  apply assumption
done

lemma Map_Fields_exists[rule_format]:
"N:AL \<Longrightarrow> AL_cod N \<subseteq> dom f \<longrightarrow> (\<exists> M . (N,f,M):Map_Fields \<and> M:AL)"
apply (erule AL.induct)
apply clarsimp
  apply (rule_tac x=emp in exI)  apply (rule, rule Map_Fields_emp) apply (rule AL_EMP)
apply clarsimp
  apply (erule impE) apply (rule SetTriv101) prefer 2 apply assumption apply (simp add: AL_cod_def) apply clarsimp
    apply (frule AL_domI) apply (rule, erule AL_update5) apply fast 
  apply clarsimp
    apply (subgoal_tac "y:dom f")
      apply (drule domD, clarsimp)
      apply (rule, rule, rule Map_Fields_cons) 
        apply fastsimp
        apply assumption
        apply assumption
        apply assumption
        apply simp
      apply (erule_tac M=Ma and x=x in AL_CONS) apply (drule Map_Fields_dom) apply simp apply simp
    apply (rule SetTriv) apply (rule_tac M=M and x=x in AL_cod_upd2) apply assumption
done

lemma Map_Fields_exists_with_Props:
  "\<lbrakk>N:AL; AL_cod N \<subseteq> dom f\<rbrakk>
 \<Longrightarrow> \<exists> M . (N,f,M):Map_Fields \<and> M:AL \<and>
           (\<forall> flds c . N\<down>fld = Some c \<longrightarrow> (\<exists> fc . f c = Some fc \<and> M\<down>fld = Some fc)) \<and> 
           (\<forall> fld fc . M\<down>fld = Some fc \<longrightarrow> (\<exists> c . f c = Some fc \<and> N\<down>fld = Some c))"
apply (frule_tac f=f in Map_Fields_exists)  apply assumption apply (erule exE, erule conjE)
apply (rule, rule, assumption)
apply (rule, assumption)
apply (rule, clarsimp) apply (erule Map_Fields_D_left) apply assumption 
apply clarsimp apply (erule Map_Fields_D_right) apply assumption 
done

lemma Map_Fields_unique[rule_format]:
  "\<forall> G . (F,f,G):Map_Fields \<longrightarrow> (\<forall> E . (F,f,E):Map_Fields \<longrightarrow> E=G)"
apply (induct F)
apply clarsimp
  apply (erule Map_Fields.elims, clarsimp)
  apply (erule Map_Fields.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule Map_Fields.elims, simp add: upd_def emp_def) apply (simp add: upd_def emp_def) 
apply clarsimp
  apply (erule Map_Fields.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule Map_Fields.elims, simp add: upd_def emp_def)
     apply (simp add: upd_def emp_def) apply clarsimp
     apply (frule_tac N=FF in AL_notdomD_delete)
     apply (frule_tac N=FFa in AL_notdomD_delete) apply clarsimp
     apply (erule_tac x=Fa in allE, clarsimp)
done

lemma Map_Fields_unique_Fun_eq[rule_format]:
  "\<forall> G . (F,f,G):Map_Fields \<longrightarrow> (\<forall> c . c:AL_cod F \<longrightarrow> g c = f c) \<longrightarrow> (\<forall> E . (F,g,E):Map_Fields \<longrightarrow> E=G)"
apply (induct F)
apply clarsimp
  apply (erule Map_Fields.elims, clarsimp)
  apply (erule Map_Fields.elims, clarsimp) apply (erule thin_rl) apply clarsimp apply (simp add: upd_def emp_def)
  apply (erule Map_Fields.elims, simp add: upd_def emp_def) apply (simp add: upd_def emp_def) 
apply clarsimp
  apply (erule Map_Fields.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule Map_Fields.elims, simp add: upd_def emp_def)
     apply (simp add: upd_def emp_def) apply clarsimp
     apply (frule_tac N=FF in AL_notdomD_delete)
     apply (frule_tac N=FFa in AL_notdomD_delete) apply clarsimp
     apply (erule_tac x=Fa in allE, clarsimp)
     apply (erule impE, clarsimp) apply (erule_tac x=c in allE, erule mp) 
       apply (case_tac "c=ca", clarsimp) apply (simp add: AL_cod_def) apply (rule_tac x=flda in exI, clarsimp)
       apply (simp add: AL_cod_def, clarsimp)
       apply (rule_tac x=a in exI, clarsimp)
       apply (subgoal_tac "(delete FFa a)\<down>a = Some c") apply (subgoal_tac "(delete FFa a)\<down>a = None", fastsimp) 
          apply (rule AL_delete1) apply simp
   apply (erule_tac x=Fb in allE, clarsimp)
     apply (erule_tac x=ca in allE, erule impE) apply (simp add: AL_cod_def) apply (rule_tac x=flda in exI, simp)
     apply simp
done

lemma Map_Fields_cod_Img1: "(S,f,T):Map_Fields \<Longrightarrow> AL_cod T \<subseteq> Img f (AL_cod S)"
apply (erule Map_Fields.induct)
apply (simp add: AL_cod_emp Img_emptyset)
apply clarsimp apply (simp add: AL_cod_def)
  apply clarsimp apply (drule AL_upd_split)
    apply (erule disjE, clarsimp) apply (simp add: Img_def) 
      apply (rule_tac x=c in exI, clarsimp)
      apply (rule_tac x=fld in exI, simp add: AL_update1)
    apply clarsimp apply (simp add: Img_def)
      apply (subgoal_tac "x : {b. \<exists>a. F\<down>a = Some b}")
      apply (drule_tac x=x in SetTriv) apply assumption apply clarsimp
      apply (rule_tac x=s in exI, clarsimp)
      apply (rule_tac x=aa in exI, rule AL_update5, assumption) apply (drule_tac H=FF in AL_domI) apply fast
    apply fast
done

lemma Map_Fields_cod_Img2: "(S,f,T):Map_Fields \<Longrightarrow> Img f (AL_cod S) \<subseteq> AL_cod T"
apply (simp add: AL_cod_def Img_def)
apply clarsimp
apply (drule Map_Fields_D_left) apply assumption apply clarsimp apply (rule, assumption)
done

lemma Map_Fields_cod_Img: "(S,f,T):Map_Fields \<Longrightarrow> AL_cod T = Img f (AL_cod S)"
apply rule
  apply (erule Map_Fields_cod_Img1)
  apply (erule Map_Fields_cod_Img2)
done

lemma Map_Fields_Invv[rule_format]:
 "(S,f,T):Map_Fields \<Longrightarrow> inj_finMap f \<longrightarrow> (T,Invv f,S):Map_Fields"
apply (erule Map_Fields.induct)
apply clarsimp
  apply (rule Map_Fields_emp)
apply clarsimp
  apply (rule Map_Fields_cons) apply fastsimp apply (drule Map_Fields_dom, simp) apply assumption apply (erule Invv_f_f, assumption) apply simp
done

lemma MapFields_Fun_eqOn: "(S, f, T) \<in> Map_Fields \<Longrightarrow> (\<forall> c . c:Y \<longrightarrow> g c= f c) \<longrightarrow> AL_cod S \<subseteq> Y \<longrightarrow> (S, g, T) \<in> Map_Fields"
apply (erule Map_Fields.induct)
apply clarsimp
  apply (rule Map_Fields_emp)
apply clarsimp
  apply (erule impE) apply (rule SetTriv101) prefer 2 apply assumption apply (erule AL_cod_upd3)
  apply (rule Map_Fields_cons) apply fastsimp apply assumption apply assumption defer apply simp
  apply (erule_tac x=c in allE, erule impE) apply (rule SetTriv) prefer 2 apply assumption apply (rule AL_cod_upd2)
    apply fastsimp
done

consts MapHP::"(HP \<times> Mapping \<times> HP) set"
inductive MapHP intros
MapHP_emp:"(emp, f, emp) : MapHP"
MapHP_cons:"\<lbrakk> N = M[c\<mapsto>(C,F)]; c \<notin> AL_dom M; (M,f,fM):MapHP;
              (F,f,G):Map_Fields; 
              f c = Some fc; fN=fM[fc\<mapsto>(C,G)]\<rbrakk> \<Longrightarrow> (N,f,fN): MapHP"

lemma MapHP_AL[rule_format]:
"(H,f,K):MapHP \<Longrightarrow> HP_good H X \<longrightarrow> AL_dom H \<subseteq> X \<longrightarrow> (H:AL \<and> K:AL \<and> AL_dom K = Img f (AL_dom H))"
apply (erule MapHP.induct)
apply clarsimp
  apply (rule, rule AL_EMP) apply (simp add: Img_emptyset emp_def AL_dom_def)
apply clarsimp
  apply (erule impE) apply (erule HP_good_subset) apply (erule contained_upd) apply simp
  apply (erule impE) apply (rule SetTriv101) apply (rule AL_dom_upd) apply assumption 
  apply clarsimp
  apply (rule, erule AL_UPD)
  apply (rule, erule AL_UPD)
  apply (simp add: AL_dom_upd1) 
  apply (simp add: Img_def, clarsimp)
  apply (rule, clarsimp) 
    apply rule apply (rule_tac x=c in exI, simp)
    apply clarsimp apply (rule_tac x=s in exI, simp)
  apply clarsimp 
    apply (erule disjE, clarsimp)
    apply (rule_tac x=s in exI, simp)
done

lemma MapHP_AL2[rule_format]:
"(H,f,K):MapHP \<Longrightarrow> HP_good H X \<longrightarrow> AL_dom H \<subseteq> X \<longrightarrow> 
   (H:AL \<and> K:AL \<and> AL_dom K = Img f (AL_dom H) \<and> HP_good K (Img f X))"
apply (erule MapHP.induct)
apply clarsimp
  apply (rule, rule AL_EMP) apply (simp add: Img_emptyset emp_def AL_dom_def)
  apply (simp add: HP_good_def)
apply clarsimp
  apply (erule impE) apply (erule HP_good_subset) apply (erule contained_upd) apply simp
  apply (erule impE) apply (rule SetTriv101) apply (rule AL_dom_upd) apply assumption 
  apply clarsimp
  apply (rule, erule AL_UPD)
  apply (rule, erule AL_UPD)
  apply rule
    apply (simp add: AL_dom_upd1) 
    apply (simp add: Img_def, clarsimp)
    apply (rule, clarsimp) 
      apply rule apply (rule_tac x=c in exI, simp)
      apply clarsimp apply (rule_tac x=s in exI, simp)
    apply clarsimp 
      apply (erule disjE, clarsimp)
      apply (rule_tac x=s in exI, simp)
  apply (simp add: HP_good_def, clarsimp)
     apply (drule AL_upd_split) 
     apply (erule disjE) apply clarsimp  
       apply (erule_tac x=c in allE, simp add: AL_update1)
       apply (rule, drule Map_Fields_AL) apply clarsimp
       apply (rule SetTriv101, erule Map_Fields_cod_Img1) apply (simp add: Img_def, clarsimp) 
         apply (rule_tac x=s in exI, simp) apply (erule SetTriv) apply assumption
     apply clarsimp
done

lemma MapHP_AL_delete[rule_format]:"(N,f,M):MapHP \<Longrightarrow> inj_on f (dom f) \<longrightarrow> (\<forall> x fx. f x = Some fx \<longrightarrow> (delete N x,f,delete M fx):MapHP)"
apply (erule MapHP.induct)
apply clarsimp
  apply (subgoal_tac "(emp::HP) = delete (emp::HP) x", clarsimp)
  apply (subgoal_tac "(emp::HP) = delete (emp::HP) fx", clarsimp)
    apply (rule MapHP_emp)
  apply (simp add: emp_def)
  apply (simp add: emp_def)
apply clarsimp
  apply (case_tac "x=c", clarsimp)
    apply (simp add: AL_delete_upd_same)
  apply (frule_tac z=x and M=M and y="(C,F)" in AL_delete_upd_diff) apply clarsimp
  apply (subgoal_tac "fx \<noteq> fc")
    apply (frule_tac z=fx and M=fM and y="(C, G)" in AL_delete_upd_diff) apply clarsimp
    apply (erule_tac x=x in allE, clarsimp)
    apply (rule MapHP_cons) apply fastsimp apply (simp add: AL_dom_delete) apply assumption apply assumption apply assumption apply simp
  apply clarsimp
    apply (drule_tac x=x and y=c and A="dom f" in inj_onD) apply simp apply (erule domI) apply (erule domI) apply simp
done

lemma MapHP_AL_delete2:"\<lbrakk>(N,f,M):MapHP; x \<notin> AL_dom N\<rbrakk> \<Longrightarrow> (delete N x,f,M):MapHP"
apply (subgoal_tac "delete N x = N", clarsimp)
apply (erule AL_notdomD_delete) 
done

lemma MapHP_dom:"(N,f,M):MapHP \<Longrightarrow> AL_dom M = Img f (AL_dom N)"
apply (erule MapHP.induct)
  apply (simp add: AL_dom_emp Img_emptyset) 
apply clarsimp
  apply (simp add: Img_def AL_dom_upd1)
  apply (rule, clarsimp)
    apply rule apply (rule_tac x=c in exI, simp)
    apply clarsimp  apply (rule_tac x=s in exI, simp)
  apply clarsimp 
    apply (erule disjE, clarsimp)
    apply (rule_tac x=s in exI, simp)
done

lemma MapHP_D_right[rule_format]:
  "(N,f,M):MapHP \<Longrightarrow> M\<down>fc = Some (C,G) \<longrightarrow> 
   (\<exists> c F . f c = Some fc \<and> N\<down>c = Some (C,F) \<and> (F,f,G):Map_Fields)"
apply (erule MapHP.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (rule_tac x=c in exI, simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "c=ca", clarsimp)
    apply (rule_tac x=ca in exI, simp)
    apply (rule, rule, erule AL_update5) apply assumption apply assumption
done

lemma MapHP_D_left[rule_format]:
  "(N,f,M):MapHP \<Longrightarrow> inj_on f (AL_dom N) \<longrightarrow> N\<down>c = Some (C,F) \<longrightarrow> 
    (\<exists> fc G . f c = Some fc \<and> M\<down>fc = Some (C,G) \<and> (F,f,G):Map_Fields)"
apply (erule MapHP.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (erule impE) apply (erule_tac X="AL_dom M[ca\<mapsto>(Ca, Fa)]" in inj_on_subset) apply (rule AL_dom_upd)
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "fca=fc", clarsimp)
    prefer 2 apply (rule, rule, erule AL_update5a) apply assumption apply assumption
    apply (drule_tac x=ca and y=c and A="AL_dom M[ca\<mapsto>(Ca, Fa)]" in inj_onD) apply simp apply (rule AL_dom_upd2) 
      apply (drule AL_domI) apply (simp add: AL_dom_upd1)
      apply clarsimp
done

lemma MapHP_D_left2[rule_format]:
  "(N,f,M):MapHP \<Longrightarrow> inj_on f (dom f) \<longrightarrow> N\<down>c = Some (C,F) \<longrightarrow>
    (\<exists> fc G . f c = Some fc \<and> M\<down>fc = Some (C,G) \<and> (F,f,G):Map_Fields)"
apply (erule MapHP.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (simp add: AL_update1)
  apply clarsimp   
    apply (case_tac "fca=fc", clarsimp)
    prefer 2 apply (rule, rule, erule AL_update5a) apply assumption apply assumption
    apply (drule_tac x=ca and y=c and A="dom f" in inj_onD) apply simp apply (erule domI) apply (erule domI)
      apply clarsimp
done

lemma MapHP_exists[rule_format]:
"N:AL \<Longrightarrow> HP_good N X \<longrightarrow> AL_dom N \<union> X \<subseteq> dom f \<longrightarrow> (\<exists> M . (N,f,M):MapHP)"
apply (erule AL.induct)
apply clarsimp
  apply (rule_tac x=emp in exI)  apply (rule MapHP_emp)
apply clarsimp
  apply (erule impE) apply (erule HP_good_subset) apply (erule contained_upd) apply simp
  apply (erule impE) apply (simp add: AL_dom_upd1)
  apply clarsimp
    apply (drule_tac c=x and C=a and F=b in HP_goodD) apply (simp add: AL_update1) apply clarsimp 
    apply (drule_tac N=b and f=f in Map_Fields_exists) apply fast
    apply (subgoal_tac "x:dom f")
      apply (drule domD, clarsimp)
      apply (rule, rule_tac M="delete M x" and c=x and C=a and F=b and fM=Ma in MapHP_cons) 
        apply (simp add: upd_def) apply (rule AL_delete_idempotent)
        apply (simp add: AL_dom_delete) 
        apply (drule AL_notdomD_delete) apply simp
        apply assumption
        apply assumption
        apply simp
    apply (simp add: AL_dom_upd1)
done

lemma MapHP_exists_with_Props:
  "\<lbrakk>N:AL; HP_good N X; AL_dom N \<subseteq> X; X \<subseteq> dom f; inj_on f X\<rbrakk>
 \<Longrightarrow> \<exists> M . (N,f,M):MapHP \<and> M:AL \<and>
           (\<forall> c C F . N\<down>c = Some (C,F) \<longrightarrow> (\<exists> fc G . f c = Some fc \<and> M\<down>fc = Some (C,G) \<and> (F,f,G):Map_Fields)) \<and>
           (\<forall> fc C G . M\<down>fc = Some (C,G) \<longrightarrow> (\<exists> c F . f c = Some fc \<and> N\<down>c = Some (C,F) \<and> (F,f,G):Map_Fields))"
apply (frule_tac f=f in MapHP_exists)  apply assumption apply fast apply (erule exE)
apply (rule, rule, assumption)
apply (frule MapHP_AL, assumption)  apply assumption 
apply (rule, clarsimp)
apply (rule, clarsimp) apply (drule MapHP_D_left) apply (erule inj_on_subset) apply assumption apply assumption apply simp
apply clarsimp apply (drule MapHP_D_right) apply assumption apply simp
done

lemma MapHP_exists_with_allProps:
  "\<lbrakk>N:AL; HP_good N X; AL_dom N \<subseteq> X; X \<subseteq> dom f; inj_on f X\<rbrakk>
 \<Longrightarrow> \<exists> M . (N,f,M):MapHP \<and> M:AL \<and>
           (\<forall> c C F . N\<down>c = Some (C,F) \<longrightarrow> (\<exists> fc G . f c = Some fc \<and> M\<down>fc = Some (C,G) \<and> (F,f,G):Map_Fields \<and> F:AL \<and> G:AL \<and>
                                                      (\<forall> flds d . F\<down>fld = Some d \<longrightarrow> (\<exists> fd . f d = Some fd \<and> G\<down>fld = Some fd)) \<and> 
                                                      (\<forall> fld fd . G\<down>fld = Some fd \<longrightarrow> (\<exists> d . f d = Some fd \<and> F\<down>fld = Some d)))) \<and>
           (\<forall> fc C G . M\<down>fc = Some (C,G) \<longrightarrow> (\<exists> c F . f c = Some fc \<and> N\<down>c = Some (C,F) \<and> (F,f,G):Map_Fields \<and> F:AL \<and> G:AL \<and>
                                                      (\<forall> flds d . F\<down>fld = Some d \<longrightarrow> (\<exists> fd . f d = Some fd \<and> G\<down>fld = Some fd)) \<and> 
                                                      (\<forall> fld fd . G\<down>fld = Some fd \<longrightarrow> (\<exists> d . f d = Some fd \<and> F\<down>fld = Some d))))"
apply (frule_tac f=f in MapHP_exists)  apply assumption apply fast apply (erule exE)
apply (rule, rule, assumption)
apply (frule MapHP_AL, assumption)  apply assumption 
apply (rule, clarsimp)
apply (rule, clarsimp) apply (frule MapHP_D_left) apply (erule inj_on_subset) apply assumption apply assumption 
  apply clarsimp 
  apply (frule Map_Fields_AL) 
  apply clarsimp
  apply (rule, clarsimp) apply (erule Map_Fields_D_left) apply assumption 
  apply (rule, clarsimp) apply (erule Map_Fields_D_right) apply assumption
apply clarsimp apply (drule MapHP_D_right) apply assumption 
  apply clarsimp 
  apply (rule_tac x=c in exI, clarsimp) 
  apply (frule Map_Fields_AL) 
  apply clarsimp
  apply (rule, clarsimp) apply (erule Map_Fields_D_left) apply assumption 
  apply (rule, clarsimp) apply (erule Map_Fields_D_right) apply assumption
done

lemma MapHP_unique[rule_format]:
  "\<forall> G . (F,f,G):MapHP \<longrightarrow> (\<forall> E . (F,f,E):MapHP \<longrightarrow> E=G)"
apply (induct F)
apply clarsimp
  apply (erule MapHP.elims, clarsimp)
  apply (erule MapHP.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MapHP.elims, simp add: upd_def emp_def) apply (simp add: upd_def emp_def) 
apply clarsimp
  apply (erule MapHP.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MapHP.elims, simp add: upd_def emp_def)
     apply (simp add: upd_def emp_def) apply clarsimp
     apply (frule_tac N=M in AL_notdomD_delete)
     apply (frule_tac N=Ma in AL_notdomD_delete) apply clarsimp
     apply (erule_tac x=fMa in allE, clarsimp)
     apply (erule Map_Fields_unique) apply assumption
done

lemma MapHP_Invv[rule_format]:
 "(H,f,K):MapHP \<Longrightarrow> inj_finMap f \<longrightarrow> (K,Invv f,H):MapHP"
apply (erule MapHP.induct)
apply clarsimp
  apply (rule MapHP_emp)
apply clarsimp
  apply (simp add: inj_finMap_def)
  apply (rule MapHP_cons) 
    apply fastsimp 
    apply (drule MapHP_dom, simp) 
      apply (simp add: Img_def, clarsimp) apply (drule_tac x=c and y=s in inj_onD) apply simp apply (erule domI) apply (erule domI) apply simp 
    apply assumption 
    apply (erule Map_Fields_Invv) apply (simp add: inj_finMap_def)
    apply (rule Invv_f_f) apply (simp add: inj_finMap_def) apply assumption
    apply simp 
done

lemma MapHP_AbsFields_Img: "(H, \<beta>, K) \<in> MapHP \<Longrightarrow> AbsFields K \<subseteq> Img \<beta> (AbsFields H)"
apply (simp add: AbsFields_def AbsFieldsOf_def AbsFieldsAt_def, clarsimp) 
             apply (drule_tac N=H in MapHP_D_right) apply assumption apply clarsimp
             apply (drule Map_Fields_D_right) apply assumption apply clarsimp
             apply (simp add: Img_def) apply (rule_tac x=ca in exI, clarsimp) apply fast
done

lemma MapHP_AbsFields_Img2: "\<lbrakk>(H, \<beta>, K) \<in> MapHP; inj_on \<beta> (AL_dom H)\<rbrakk> \<Longrightarrow> AbsFields K = Img \<beta> (AbsFields H)"
apply (simp add: AbsFields_def AbsFieldsOf_def AbsFieldsAt_def) 
apply (rule, rule, clarsimp)
             apply (drule_tac N=H in MapHP_D_right) apply assumption apply clarsimp
             apply (drule Map_Fields_D_right) apply assumption apply clarsimp
             apply (simp add: Img_def) apply (rule_tac x=ca in exI, clarsimp) apply fast
apply (rule, simp add: Img_def, clarsimp)
             apply (drule_tac N=H in MapHP_D_left) apply assumption apply assumption apply clarsimp 
               apply (rule_tac x=fc in exI, clarsimp) 
               apply (drule Map_Fields_D_left) apply assumption apply clarsimp  apply (rule, assumption)
done

lemma MapHP_Fun_eqOn: "(H, f, K) \<in> MapHP \<Longrightarrow> (\<forall> c . c:Y \<longrightarrow> g c= f c) \<longrightarrow> AL_dom H \<subseteq> Y \<longrightarrow> HP_good H Y \<longrightarrow> (H, g, K) \<in> MapHP"
apply (erule MapHP.induct)
apply clarsimp
  apply (rule MapHP_emp)
apply clarsimp
  apply (erule impE) apply (simp add: AL_dom_upd1) 
  apply (erule impE) apply (erule HP_good_subset) apply (erule contained_upd) apply simp
  apply (subgoal_tac "AL_cod F \<subseteq> Y")
    apply (rule MapHP_cons) apply fastsimp apply assumption apply assumption prefer 3 apply fastsimp  
      apply (drule_tac Y=Y and g=g in MapFields_Fun_eqOn) apply clarsimp 
      apply (erule_tac x=c in allE, erule impE) apply (rule SetTriv) prefer 2 apply assumption apply (rule AL_dom_upd2) apply simp
  apply (drule HP_goodD) apply (rule AL_update1) apply fast
done

consts MAP_OPSTACK::"(Colour list \<times> Mapping \<times> (Colour list)) set"
inductive MAP_OPSTACK intros
MAP_OPSTACK_emp: "([],f,[]):MAP_OPSTACK"
MAP_OPSTACK_cons: "\<lbrakk>ops = d #ops'; f d = Some fd; (ops',f,ops1'):MAP_OPSTACK; ops1=fd#ops1'\<rbrakk> \<Longrightarrow> (ops,f,ops1):MAP_OPSTACK"

lemma MAP_OPSTACK_lookupR[rule_format]: 
  "(S,f,T):MAP_OPSTACK \<Longrightarrow> (\<forall> n c . n < length T \<longrightarrow> T!n = c \<longrightarrow> (\<exists> d . S!n = d \<and> f d = Some c))"
apply (erule MAP_OPSTACK.induct)
apply clarsimp
apply clarsimp
  apply (case_tac n, clarsimp) 
  apply clarsimp
done

lemma MAP_OPSTACK_lookupL[rule_format]: 
  "(S,f,T):MAP_OPSTACK \<Longrightarrow> (\<forall> n c . n < length S \<longrightarrow> S!n = c \<longrightarrow> (\<exists> d . f c = Some d \<and> T!n = d))"
apply (erule MAP_OPSTACK.induct)
apply clarsimp
apply clarsimp
  apply (case_tac n, clarsimp) 
  apply clarsimp
done

lemma MAP_OPSTACK_set_Img: "(S,f,T):MAP_OPSTACK \<Longrightarrow> set T = Img f (set S)"
apply (erule MAP_OPSTACK.induct)
apply (simp add: Img_def)
apply (simp add: Img_def) 
  apply (rule, rule, clarsimp) 
    apply (erule disjE, clarsimp)  apply (rule_tac x=d in exI, simp)
    apply clarsimp  apply (rule_tac x=s in exI, clarsimp) 
  apply (rule, clarsimp) 
    apply (erule disjE, clarsimp) 
    apply (erule_tac x=s in allE, clarsimp)
done

lemma MAP_OPSTACK_length: "(S,f,T):MAP_OPSTACK \<Longrightarrow> length T = length S"
apply (erule MAP_OPSTACK.induct)
apply simp
apply simp
done

lemma MAP_OPSTACK_exists[rule_format]: "set S \<subseteq> dom f \<longrightarrow> (\<exists> T . (S,f,T):MAP_OPSTACK)"
apply (induct S)
apply clarsimp
  apply (rule, rule MAP_OPSTACK_emp)
apply clarsimp
  apply (rule, rule MAP_OPSTACK_cons) apply fastsimp apply assumption apply assumption apply simp
done

lemma MAP_OPSTACK_unique[rule_format]:
  "\<forall> T . (S,f,T):MAP_OPSTACK \<longrightarrow> (\<forall> R . (S,f,R):MAP_OPSTACK \<longrightarrow> R=T)"
apply (induct S)
apply clarsimp
  apply (erule MAP_OPSTACK.elims, clarsimp)
  apply (erule MAP_OPSTACK.elims, clarsimp) apply clarsimp
  apply (erule MAP_OPSTACK.elims, clarsimp) apply clarsimp
apply clarsimp
  apply (erule MAP_OPSTACK.elims, clarsimp) apply clarsimp
  apply (erule MAP_OPSTACK.elims, clarsimp) apply clarsimp
done

lemma MAP_OPSTACK_Invv[rule_format]:
 "(S,f,T):MAP_OPSTACK \<Longrightarrow> inj_finMap f \<longrightarrow> (T,Invv f,S):MAP_OPSTACK"
apply (erule MAP_OPSTACK.induct)
apply clarsimp
  apply (rule MAP_OPSTACK_emp)
apply clarsimp
  apply (rule MAP_OPSTACK_cons) apply fastsimp apply (erule Invv_f_f, assumption) apply assumption apply simp
done

lemma MAP_OPSTACK_Fun_eqOn: "(S, f, T) \<in> MAP_OPSTACK \<Longrightarrow> (\<forall> c . c:Y \<longrightarrow> g c= f c) \<longrightarrow> set S \<subseteq> Y \<longrightarrow> (S, g, T) \<in> MAP_OPSTACK"
apply (erule MAP_OPSTACK.induct)
apply clarsimp
  apply (rule MAP_OPSTACK_emp)
apply clarsimp
  apply (rule MAP_OPSTACK_cons) apply fastsimp defer apply assumption apply simp
    apply fastsimp
done

consts MAP_STORE::"(VarContext \<times> Mapping \<times> VarContext) set"
inductive MAP_STORE intros
MAP_STORE_emp: "(emp,f,emp):MAP_STORE"
MAP_STORE_cons: "\<lbrakk>s = ss[x\<mapsto>c]; x\<notin>AL_dom ss; f c = Some fc; (ss,f,tt):MAP_STORE; t=tt[x\<mapsto>fc]\<rbrakk> \<Longrightarrow> (s,f,t):MAP_STORE"

lemma MAP_STORE_dom:
"(s,f,t):MAP_STORE \<Longrightarrow> AL_dom t = AL_dom s"
apply (erule MAP_STORE.induct)
  apply (simp add: AL_dom_emp)
  apply (simp add: AL_dom_upd1)
done

lemma MAP_STORE_AL[rule_format]:"(s,f,t):MAP_STORE \<Longrightarrow> s:AL \<and> t:AL"
apply (erule MAP_STORE.induct)
  apply (rule, rule AL_EMP, rule AL_EMP) 
apply clarsimp
apply rule
  apply (rule_tac M="delete ss x" and x=x and y=c in AL_CONS)
    apply (erule AL_DEL)
               apply (simp add: AL_dom_delete) 
               apply (simp add: upd_def) 
               apply (rule AL_delete_idempotent)
apply (rule_tac M="delete tt x" and x=x and y=fc in AL_CONS)
    apply (erule AL_DEL)
               apply (simp add: AL_dom_delete) 
               apply (simp add: upd_def) 
               apply (rule AL_delete_idempotent)
done

lemma MAP_STORE_D_right[rule_format]:
  "(s,f,t):MAP_STORE \<Longrightarrow> t\<down>x = Some fc \<longrightarrow> (\<exists> c . f c = Some fc \<and> s\<down>x = Some c)"
apply (erule MAP_STORE.induct)
  apply (simp add: AL_emp1) 
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp) 
    apply (rule_tac x=c in exI, simp add: AL_update1)
  apply clarsimp
    apply (rule_tac x=ca in exI, simp)
    apply (erule AL_update5) apply assumption
done

lemma MAP_STORE_D_left[rule_format]: 
  "(s,f,t):MAP_STORE \<Longrightarrow> (\<forall> x c . s\<down>x = Some c \<longrightarrow> (\<exists> fc . f c = Some fc \<and> t\<down>x = Some fc))"
apply (erule MAP_STORE.induct)
apply (simp add: AL_emp1)
apply clarsimp
  apply (drule AL_upd_split)
  apply (erule disjE, clarsimp)
    apply (simp add: AL_update1)
  apply clarsimp
    apply (erule_tac x=xa in allE, clarsimp)
    apply (erule AL_update5)  apply assumption
done

lemma MAP_STORE_exists[rule_format]:
"s:AL \<Longrightarrow> AL_cod s \<subseteq> dom f \<longrightarrow> (\<exists> t . (s,f,t):MAP_STORE)"
apply (erule AL.induct)
apply clarsimp
  apply (rule_tac x=emp in exI)  apply (rule MAP_STORE_emp) 
apply clarsimp
  apply (erule impE) apply (rule SetTriv101) prefer 2 apply assumption apply (simp add: AL_cod_def) apply clarsimp
    apply (frule AL_domI) apply (rule, erule AL_update5) apply fast 
  apply clarsimp
    apply (subgoal_tac "y:dom f")
      apply (drule domD, clarsimp)
      apply (rule, rule MAP_STORE_cons) 
        apply fastsimp
        apply assumption
        apply assumption
        apply assumption
        apply simp
    apply (rule SetTriv) apply (rule_tac M=M and x=x in AL_cod_upd2) apply assumption
done

lemma MAP_STORE_exists2[rule_format]:
"s:AL \<Longrightarrow> AL_cod s \<subseteq> dom f \<longrightarrow> (\<exists> t . (s,f,t):MAP_STORE \<and> t:AL)"
apply (erule AL.induct)
apply clarsimp
  apply (rule_tac x=emp in exI)  apply (rule, rule MAP_STORE_emp) apply (rule AL_EMP)
apply clarsimp
  apply (erule impE) apply (rule SetTriv101) prefer 2 apply assumption apply (simp add: AL_cod_def) apply clarsimp
    apply (frule AL_domI) apply (rule, erule AL_update5) apply fast 
  apply clarsimp
    apply (subgoal_tac "y:dom f")
      apply (drule domD, clarsimp)
      apply (rule, rule, rule MAP_STORE_cons) 
        apply fastsimp
        apply assumption
        apply assumption
        apply assumption
        apply simp
      apply (erule_tac M=t in AL_CONS) prefer 2 apply fastsimp apply (drule MAP_STORE_dom) apply simp
    apply (rule SetTriv) apply (rule_tac M=M and x=x in AL_cod_upd2) apply assumption
done

lemma MAP_STORE_exists_with_Props:
  "\<lbrakk>s:AL; AL_cod s \<subseteq> dom f\<rbrakk>
 \<Longrightarrow> \<exists> t . (s,f,t):MAP_STORE \<and> t:AL \<and>
           (\<forall> x c . s\<down>x = Some c \<longrightarrow> (\<exists> fc . f c = Some fc \<and> t\<down>x = Some fc)) \<and> 
           (\<forall> x fc . t\<down>x = Some fc \<longrightarrow> (\<exists> c . f c = Some fc \<and> s\<down>x = Some c))"
apply (frule_tac f=f in MAP_STORE_exists2)  apply assumption apply (erule exE) apply (erule conjE)
apply (rule, rule, assumption)
apply (rule, assumption)
apply (rule, clarsimp) apply (erule MAP_STORE_D_left) apply assumption 
apply clarsimp apply (erule MAP_STORE_D_right) apply assumption 
done

lemma MAP_STORE_unique[rule_format]:
  "\<forall> T . (S,f,T):MAP_STORE \<longrightarrow> (\<forall> R . (S,f,R):MAP_STORE \<longrightarrow> R=T)"
apply (induct S)
apply clarsimp
  apply (erule MAP_STORE.elims, clarsimp)
  apply (erule MAP_STORE.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MAP_STORE.elims, simp add: upd_def emp_def) apply (simp add: upd_def emp_def) 
apply clarsimp
  apply (erule MAP_STORE.elims, clarsimp) apply (simp add: upd_def emp_def)
  apply (erule MAP_STORE.elims, simp add: upd_def emp_def)
     apply (simp add: upd_def emp_def) apply clarsimp
     apply (frule_tac N=ss in AL_notdomD_delete)
     apply (frule_tac N=ssa in AL_notdomD_delete) apply clarsimp
     apply (erule_tac x=tt in allE, clarsimp)
done

lemma MAP_STORE_cod_Img1: "(S,f,T):MAP_STORE \<Longrightarrow> AL_cod T \<subseteq> Img f (AL_cod S)"
apply (erule MAP_STORE.induct)
apply (simp add: AL_cod_emp Img_emptyset)
apply clarsimp apply (simp add: AL_cod_def)
  apply clarsimp apply (drule AL_upd_split)
    apply (erule disjE, clarsimp) apply (simp add: Img_def) 
      apply (rule_tac x=c in exI, clarsimp)
      apply (rule_tac x=x in exI, simp add: AL_update1)
    apply clarsimp apply (simp add: Img_def)
      apply (subgoal_tac "xa : {b. \<exists>a. tt\<down>a = Some b}")
      apply (drule_tac x=xa in SetTriv) apply assumption apply clarsimp
      apply (rule_tac x=s in exI, clarsimp)
      apply (rule_tac x=aa in exI, rule AL_update5, assumption) apply (drule_tac H=ss in AL_domI) apply fast
    apply fast
done

lemma MAP_STORE_cod_Img2: "(S,f,T):MAP_STORE \<Longrightarrow> Img f (AL_cod S) \<subseteq> AL_cod T"
apply (simp add: AL_cod_def Img_def)
apply clarsimp
apply (drule MAP_STORE_D_left) apply assumption apply clarsimp apply (rule, assumption)
done

lemma MAP_STORE_cod_Img: "(S,f,T):MAP_STORE \<Longrightarrow> AL_cod T = Img f (AL_cod S)"
apply rule
  apply (erule MAP_STORE_cod_Img1)
  apply (erule MAP_STORE_cod_Img2)
done

lemma MAP_STORE_Invv[rule_format]:
 "(S,f,T):MAP_STORE \<Longrightarrow> inj_finMap f \<longrightarrow> (T,Invv f,S):MAP_STORE"
apply (erule MAP_STORE.induct)
apply clarsimp
  apply (rule MAP_STORE_emp)
apply clarsimp
  apply (rule MAP_STORE_cons) apply fastsimp apply (drule MAP_STORE_dom, simp) apply (erule Invv_f_f, assumption) apply assumption apply simp
done

lemma MAP_STORE_Fun_eqOn: "(S, f, T) \<in> MAP_STORE \<Longrightarrow> (\<forall> c . c:Y \<longrightarrow> g c= f c) \<longrightarrow> AL_cod S \<subseteq> Y \<longrightarrow> (S, g, T) \<in> MAP_STORE"
apply (erule MAP_STORE.induct)
apply clarsimp
  apply (rule MAP_STORE_emp)
apply clarsimp
  apply (erule impE) apply (rule SetTriv101) prefer 2 apply assumption apply (erule AL_cod_upd3)
  apply (rule MAP_STORE_cons) apply fastsimp apply assumption defer apply assumption apply simp
  apply (erule_tac x=c in allE, erule impE) apply (rule SetTriv) prefer 2 apply assumption apply (rule AL_cod_upd2)
    apply fastsimp
done

constdefs Renames_STATE::"STATE \<Rightarrow> Mapping \<Rightarrow> STATE \<Rightarrow> bool"
"Renames_STATE \<Sigma> f \<Pi> \<equiv>
     case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow> 
     case \<Pi> of (T,\<Delta>,K) \<Rightarrow> (S,f,T):MAP_OPSTACK \<and> (\<Gamma>,f,\<Delta>):MAP_STORE \<and> (H,f,K):MapHP"

lemma Renames_STATE_Invv: "\<lbrakk>Renames_STATE \<Sigma> f \<Pi>; inj_finMap f\<rbrakk> \<Longrightarrow> Renames_STATE \<Pi> (Invv f) \<Sigma>"
apply (simp add: Renames_STATE_def)
apply (case_tac \<Sigma>, clarsimp)
apply (case_tac \<Pi>, clarsimp)
apply (drule MAP_OPSTACK_Invv)  apply assumption
apply (drule MAP_STORE_Invv)  apply assumption
apply (drule MapHP_Invv)  apply assumption
apply simp
done

lemma Renames_STATE_Invv_expand:
  "\<lbrakk>Renames_STATE (S, \<Gamma>, H) f (T, \<Delta>, K); inj_finMap f\<rbrakk> \<Longrightarrow> Renames_STATE (T, \<Delta>, K) (Invv f) (S, \<Gamma>, H)"
apply (simp add: Renames_STATE_def, clarsimp)
apply (rule, erule MAP_OPSTACK_Invv, assumption) 
apply (rule, erule MAP_STORE_Invv, assumption) 
apply (erule MapHP_Invv, assumption) 
done

lemma Renames_STATE_setOPS: "Renames_STATE (S,\<Gamma>,H) f (T,\<Delta>,K) \<Longrightarrow> set T = Img f (set S)"
apply (simp add: Renames_STATE_def, clarsimp)
apply(erule MAP_OPSTACK_set_Img)
done

lemma Renames_STATE_codSTORE:"Renames_STATE (S,\<Gamma>,H) f (T,\<Delta>,K) \<Longrightarrow> AL_cod \<Delta> = Img f (AL_cod \<Gamma>)"
apply (simp add: Renames_STATE_def, clarsimp)
apply (erule MAP_STORE_cod_Img)
done

lemma Renames_STATE_domHP:"Renames_STATE (S,\<Gamma>,H) f (T,\<Delta>,K) \<Longrightarrow> AL_dom K = Img f (AL_dom H)"
apply (simp add: Renames_STATE_def, clarsimp)
apply (erule MapHP_dom)
done

lemma Renames_STATE_AbsFields:
"\<lbrakk>Renames_STATE (S,\<Gamma>,H) f (T,\<Delta>,K);inj_on f (AL_dom H)\<rbrakk> \<Longrightarrow> AbsFields K = Img f (AbsFields H)"
apply (simp add: Renames_STATE_def, clarsimp)
apply (erule MapHP_AbsFields_Img2) apply assumption
done

(*
constdefs disjoint_from::"HRSD => Colour set \<Rightarrow> bool"
"disjoint_from \<gamma> X \<equiv> case \<gamma> of (\<Sigma>,N,\<Pi>) \<Rightarrow> (AL_dom N) \<inter> X = {}"
*)

lemma sum_sum_State_welldef:
  "\<lbrakk> State_welldef (fhS, fh\<Gamma>, HH) NN; AL_sum fN Ma NN; 
     AL_sum fN' Ma MM; contained fN fN'; AL_dom NN \<subseteq> AL_dom MM;
     AL_sum fH gHa HH; AL_sum fH' gHa HH'; AL_dom fH \<subseteq> AL_dom fH';
     State_welldef (fS', f\<Gamma>', HH') MM\<rbrakk>
  \<Longrightarrow> State_welldef (fhS, fh\<Gamma>, HH') MM"
apply (simp add: State_welldef_def, clarsimp)
apply rule apply (erule SetTriv101, assumption)
apply rule apply (erule SetTriv101, assumption)
apply rule defer
  apply (simp add: State_welldef_def State_good_def)
apply clarsimp
apply (case_tac "x \<in> AbsFields HH'")
  apply (rule SetTriv) prefer 2 apply assumption apply clarsimp
apply clarsimp
  apply (subgoal_tac "x:ObjIDs NN")
    apply (subgoal_tac "x: AL_dom HH") prefer 2 apply (rule SetTriv) prefer 2 apply assumption apply fast
    apply (drule AL_sum_dom_split, assumption)
      apply (rotate_tac -1, erule disjE)
      apply (drule_tac X="AL_dom fH" in SetTriv, assumption)
      apply (erule AL_sum3a, assumption)
    apply (erule AL_sum4a, assumption)
  apply (simp add: ObjIDs_def, clarsimp)
    apply (subgoal_tac "x:AL_dom NN") prefer 2 apply (erule disjE, erule SetTriv, assumption) apply (erule SetTriv, assumption)
    apply (drule AL_domD, clarsimp)
    apply (drule_tac N=NN in AL_sum_split, assumption)
      apply (rotate_tac -1, erule disjE)
      apply (drule contained3, assumption)
      apply (drule_tac H=fN' in AL_sum3, assumption)
      apply clarsimp
    apply (drule_tac H=fN' in AL_sum4, assumption)
      apply clarsimp
done


(*The following is a sub-property of State_welldef*)
constdefs State_noDangColours::"STATE \<Rightarrow> MP \<Rightarrow> bool"
"State_noDangColours \<Sigma> N \<equiv>
     (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow>
        (set S) \<union> (AL_cod \<Gamma>) \<union> (AL_dom H) \<union> AbsFields H \<subseteq> AL_dom N \<and>
     (AL_dom H) \<subseteq> ObjIDs N \<and>
     State_good \<Sigma> N \<and> N:AL)"


constdefs HPdom_is::"HP \<Rightarrow> Mapping \<Rightarrow> HP \<Rightarrow> MP \<Rightarrow> bool"
"HPdom_is H f K N \<equiv> AL_dom H = {c . \<exists> x . x: AL_dom K \<and> f c = Some x \<and> c:ObjIDs N}"

constdefs renUniState::"(STATE \<times> MP) \<Rightarrow> Mapping \<Rightarrow> (STATE \<times> MP) \<Rightarrow> bool"
"renUniState \<gamma> f \<gamma>' \<equiv> case \<gamma> of (\<Sigma>, N) \<Rightarrow> 
                   case \<gamma>' of (\<Sigma>', N') \<Rightarrow>
                   (N,f,N'):MapN \<and> Renames_STATE \<Sigma> f \<Sigma>' \<and> 
                   AL_dom N = dom f \<and>
                   inj_on f (ObjIDs N) \<and> State_noDangColours \<Sigma> N"

lemma MapN_f_inj_Invv[rule_format]:
  "(N, f, M) \<in> MapN \<Longrightarrow> inj_on f (dom f) \<longrightarrow> (M, Invv f, N) \<in> MapN"
apply (erule MapN.induct)
apply (rule, rule MapN_emp)
apply rule 
  apply (erule impE, assumption)
  apply (rule MapN_cons, assumption) 
    apply clarsimp apply (drule MapN_dom, clarsimp) apply (simp add: Img_def, clarsimp)
      apply (drule_tac x=x and y=s in inj_onD)  apply simp apply (erule domI) apply (erule domI) apply simp 
    apply assumption 
    prefer 3 apply assumption
    apply (subgoal_tac "fx : ran f") 
      apply (drule f_Invv_f) apply clarsimp
      apply (erule inj_onD)  apply simp apply (erule domI) apply (erule domI) 
      apply (erule ranI)
    apply clarsimp apply (drule AL_domI) apply simp
done

end
