theory StateInterpretation = ValuesRep + InjFinMap:

consts typecorrect::"Val \<Rightarrow> TP \<Rightarrow> bool"
primrec
"typecorrect (IVal i) tp = (tp = intTP)"
"typecorrect (RVal r) tp = (\<exists> C . tp = classTP C)"

constdefs Opstack_typecorrect::"MP \<Rightarrow> Colour list \<Rightarrow> OpStack \<Rightarrow> bool"
"Opstack_typecorrect N S ops \<equiv> 
   \<forall> n . n < length S \<longrightarrow> (\<exists> tp . N\<down>(S!n) = Some tp \<and> typecorrect (ops!n) tp)"

constdefs Store_typecorrect::"MP \<Rightarrow> VarContext \<Rightarrow> Store \<Rightarrow> bool"
"Store_typecorrect N \<Gamma> s \<equiv> 
   \<forall> x c . \<Gamma>\<down>x = Some c \<longrightarrow> (\<exists> tp v. N\<down>c = Some tp \<and> s\<down>x = Some v \<and> typecorrect v tp)"

constdefs Fields_typecorrect::"MP \<Rightarrow> ((Field,Colour) AssList) \<Rightarrow> ((Field,Val) AssList) \<Rightarrow> bool"
"Fields_typecorrect N F G \<equiv> 
   \<forall> f c . F\<down>f = Some c \<longrightarrow> (\<exists> tp v. N\<down>c = Some tp \<and> G\<down>f = Some v \<and> typecorrect v tp)"

lemma Fields_tc_emp_emp:" Fields_typecorrect N emp emp"
by (simp add: Fields_typecorrect_def AL_emp1)
lemma Fields_typecorrect_MP_extensional: 
 "\<lbrakk> Fields_typecorrect N F G; \<forall> x c. (N\<down>x = Some c) = (NN\<down>x = Some c)\<rbrakk> \<Longrightarrow> Fields_typecorrect NN F G"
by (simp add: Fields_typecorrect_def)
  
constdefs HeapAbstractsAt::"MP \<Rightarrow> HP \<Rightarrow> Heap \<Rightarrow> Colour \<Rightarrow> Class \<Rightarrow> AddrMapping \<Rightarrow> bool"
"HeapAbstractsAt N H h c C I \<equiv> 
  (\<forall> D F a. H\<down>c = Some(D,F) \<longrightarrow> I c = Some a \<longrightarrow>
            (subclass_of D C \<and> (\<exists> E G . subclass_of E D \<and> h\<down>a = Some(E,G) \<and> Fields_typecorrect N F G)))"

constdefs Heap_typecorrect::"MP \<Rightarrow> HP \<Rightarrow> Heap \<Rightarrow> AddrMapping \<Rightarrow> bool"
"Heap_typecorrect N H h I \<equiv> 
   \<forall> c . c:AL_dom H \<longrightarrow> (\<exists> C . N\<down>c = Some(classTP C) \<and> HeapAbstractsAt N H h c C I \<and> c:dom I)"

constdefs Object_typecorrect::"MP \<Rightarrow> OBJECT \<Rightarrow> Object \<Rightarrow> bool"
"Object_typecorrect N OBJ obj \<equiv> 
   (case OBJ of (D,F) \<Rightarrow>
     (case obj of (E,G) \<Rightarrow> subclass_of E D \<and> Fields_typecorrect N F G))"

lemma Heap_typecorrect_expand:
"Heap_typecorrect N H h I = 
   (\<forall> c OBJ. H\<down>c = Some OBJ  \<longrightarrow> (\<exists> C a obj. N\<down>c = Some(classTP C) \<and> subclass_of (fst OBJ) C \<and>
                                              I c = Some a \<and> h\<down>a = Some obj \<and> 
                                              Object_typecorrect N OBJ obj))"
apply (simp add: Heap_typecorrect_def Object_typecorrect_def)
apply (rule, clarsimp)
  apply (erule_tac x=c in allE, erule impE, erule AL_domI)
  apply clarsimp
  apply (simp add: HeapAbstractsAt_def, clarsimp)
apply clarsimp
  apply (drule AL_domD, clarsimp)
  apply (erule_tac x=c in allE, clarsimp)
  apply (simp add: HeapAbstractsAt_def) apply (erule domI)
done

lemma Heap_typecorrect_lookup_domHeap:"\<lbrakk>Heap_typecorrect N H h I; x:AL_dom H\<rbrakk> \<Longrightarrow> x:AL_dom N"
apply (simp add: Heap_typecorrect_def)
apply (erule_tac x=x in allE, clarsimp)
apply (erule AL_domI)
done

lemma Heap_tc_AddrMappingEqOnN:
  "\<lbrakk>Heap_typecorrect N H h I; \<forall>c. c \<in> ObjIDs N \<longrightarrow> J c = I c\<rbrakk> \<Longrightarrow> Heap_typecorrect N H h J"
apply (simp add: Heap_typecorrect_def, clarsimp)
apply (erule_tac x=c in allE, clarsimp)
apply (erule_tac x=c in allE, erule impE, simp add: ObjIDs_def)
apply (simp add: HeapAbstractsAt_def)
apply (frule AL_domI) apply clarsimp
done

lemma Opstack_typecorrect_FreshInsertInOPS:
 "\<lbrakk>Opstack_typecorrect N S ops; typecorrect v tp; c \<notin> set S\<rbrakk>
  \<Longrightarrow>  Opstack_typecorrect N[c\<mapsto>tp] (c # S) (v # ops)"
apply (simp add: Opstack_typecorrect_def, clarsimp)
    apply (case_tac n, clarsimp) apply (simp add: AL_update1)
    apply clarsimp apply (erule_tac x=nat in allE, clarsimp)
      apply (rule_tac x=tpa in exI, clarsimp)
      apply (erule AL_update5)  apply fastsimp
done

lemma Store_typecorrect_FreshInsert:
 "\<lbrakk>Store_typecorrect N \<Gamma> s; typecorrect v tp; c \<notin> AL_cod \<Gamma>\<rbrakk>
  \<Longrightarrow> Store_typecorrect N[c\<mapsto>tp] \<Gamma> s"
apply (simp add: Store_typecorrect_def, clarsimp)
apply (erule_tac x=x in allE, clarsimp)
apply (rule_tac x=tpa in exI, clarsimp)
apply (erule AL_update5) apply (drule AL_codI) apply fast
done

lemma Heap_typecorrect_FreshInsert:
 "\<lbrakk> Heap_typecorrect N H h I; c \<notin> AL_dom H; c \<notin> AbsFields H\<rbrakk>
    \<Longrightarrow> Heap_typecorrect N[c\<mapsto>tp] H h I"
apply (simp add: Heap_typecorrect_def, clarsimp)
apply (erule_tac x=ca in allE, clarsimp)
apply (rule_tac x=C in exI)
apply (rule, erule AL_update5) apply fast
apply (simp add: HeapAbstractsAt_def)
apply (rule, clarsimp)
  apply (simp add: Fields_typecorrect_def, clarsimp)
  apply (erule_tac x=f in allE, clarsimp)
  apply (rule_tac x=tpa in exI, clarsimp)
  apply (erule AL_update5) apply (drule AbsFieldsI, assumption) apply fast
apply (erule domI)
done

constdefs HeapAbstracts::"HP \<Rightarrow> MP \<Rightarrow> Heap \<Rightarrow> AddrMapping \<Rightarrow> bool"
"HeapAbstracts H N h I \<equiv> Img I (AL_dom H) \<subseteq> AL_dom h"

lemma HeapAbstracts_contained:
  "\<lbrakk>HeapAbstracts H N h I;  contained N M\<rbrakk> \<Longrightarrow> HeapAbstracts H M h I"
by (simp add: HeapAbstracts_def)

constdefs State_typecorrect::"MP \<Rightarrow> STATE \<Rightarrow> State \<Rightarrow> AddrMapping \<Rightarrow> bool"
"State_typecorrect N \<Sigma> \<sigma> I \<equiv>
   (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow> case \<sigma> of (ops,s,h) \<Rightarrow>
          Opstack_typecorrect N S ops \<and> Store_typecorrect N \<Gamma> s \<and> Heap_typecorrect N H h I)"

lemma State_typecorrect_HP_extensional:
 "\<lbrakk> State_typecorrect N  (S,\<Gamma>,H) (ops,s,h) I; \<forall> x c. (H\<down>x = Some c) = (HH\<down>x = Some c)\<rbrakk>
  \<Longrightarrow> State_typecorrect N (S,\<Gamma>,HH) (ops,s,h) I"
apply (simp add: State_typecorrect_def, clarsimp)
apply (simp add: Heap_typecorrect_def)
apply (rule, clarsimp)
  apply (erule_tac x=c in allE, erule impE, simp add: AL_dom_def) apply clarsimp
  apply (simp add: HeapAbstractsAt_def)
  apply (erule domI)
done

lemma Opstack_typecorrect_contained:
 "\<lbrakk>contained N M; Opstack_typecorrect N S ops\<rbrakk> \<Longrightarrow> Opstack_typecorrect M S ops"
apply (simp add: Opstack_typecorrect_def, clarsimp)
apply (erule_tac x=n in allE, clarsimp)
apply (drule contained3, assumption)
apply clarsimp
done

lemma Store_typecorrect_contained:
 "\<lbrakk>contained N M; Store_typecorrect N \<Gamma> s\<rbrakk> \<Longrightarrow> Store_typecorrect M \<Gamma> s"
apply (simp add: Store_typecorrect_def, clarsimp)
apply (erule_tac x=x in allE, clarsimp)
apply (drule contained3, assumption)
apply clarsimp
done

lemma Fields_typecorrect_contained:
  "\<lbrakk>contained N M; Fields_typecorrect N F f\<rbrakk> \<Longrightarrow> Fields_typecorrect M F f"
apply (simp add: Fields_typecorrect_def, clarsimp)
apply (erule_tac x=fa in allE, clarsimp)
apply (drule contained3, assumption)
apply clarsimp
done

lemma HeapAbstractsAt_contained:
  "\<lbrakk>contained N M; HeapAbstractsAt N H h c C I\<rbrakk> \<Longrightarrow> HeapAbstractsAt M H h c C I"
apply (simp add: HeapAbstractsAt_def, clarsimp)
apply (erule Fields_typecorrect_contained) apply assumption
done

lemma Heap_typecorrect_contained:
  "\<lbrakk>contained N M; Heap_typecorrect N H h I\<rbrakk> \<Longrightarrow> Heap_typecorrect M H h I"
apply (simp add: Heap_typecorrect_def, clarsimp)
apply (erule_tac x=c in allE, clarsimp)
apply (frule contained3, assumption)
apply clarsimp
apply (rule, erule  HeapAbstractsAt_contained) apply assumption
apply (erule domI)
done

lemma State_typecorrect_MP_contained:
 " \<lbrakk>State_typecorrect N \<Sigma> \<sigma> I; contained N NN\<rbrakk> \<Longrightarrow> State_typecorrect NN \<Sigma> \<sigma> I"
apply (case_tac \<Sigma>, clarsimp)
apply (case_tac \<sigma>, clarsimp)
apply (simp add: State_typecorrect_def, clarsimp)
apply (rule, erule Opstack_typecorrect_contained, assumption)
apply (rule, erule Store_typecorrect_contained, assumption)
apply (erule Heap_typecorrect_contained, assumption)
done

lemma State_typecorrect_contained:
  "\<lbrakk>contained N M; State_typecorrect N (S, \<Gamma>, H) (ops,s,h) I\<rbrakk> \<Longrightarrow> State_typecorrect M (S, \<Gamma>, H) (ops,s,h) I"
apply (erule State_typecorrect_MP_contained) apply assumption
done

lemma Opstack_typecorrect_Tail:
"Opstack_typecorrect N (c # S) (v # ops) \<Longrightarrow> Opstack_typecorrect N S ops"
apply (simp add: Opstack_typecorrect_def, clarsimp)
apply (erule_tac x="Suc n" in allE, clarsimp)
done

lemma State_typecorrect_Tail:
"State_typecorrect N (c # S, \<Gamma>, H) (v # ops, s, h) I 
    \<Longrightarrow> State_typecorrect N (S, \<Gamma>, H) (ops, s, h) I"
apply (simp add: State_typecorrect_def, clarsimp)
apply (erule Opstack_typecorrect_Tail)
done

lemma State_typecorrect_MP_extensional:
 "\<lbrakk> State_typecorrect N (ops,s,h) (S,\<Gamma>,H)  I; \<forall> x c. (N\<down>x = Some c) = (NN\<down>x = Some c)\<rbrakk>
  \<Longrightarrow> State_typecorrect NN (ops,s,h) (S,\<Gamma>,H) I"
apply (simp add: State_typecorrect_def, clarsimp)
apply rule
  apply (simp add: Opstack_typecorrect_def)
apply rule
  apply (simp add: Store_typecorrect_def)
apply (simp add: Heap_typecorrect_def, clarsimp)
  apply (erule_tac x=c in allE, clarsimp)
  apply (simp add: HeapAbstractsAt_def)
  apply (rule, clarsimp)
    apply (erule Fields_typecorrect_MP_extensional) apply clarsimp
  apply (erule domI)
done

lemma State_typecorrect_VR_notEmpty: "\<lbrakk>State_typecorrect N \<Sigma> \<sigma> I; c:DOM \<Sigma>\<rbrakk> \<Longrightarrow> Values_Represented c \<Sigma> \<sigma> I \<noteq> {}"
apply (simp add: State_typecorrect_def DOM_def Values_Represented_def)
apply (case_tac \<Sigma>, clarsimp)
apply (case_tac \<sigma>, clarsimp)
apply (erule disjE, simp add: Opstack_typecorrect_def set_conv_nth OpsValues_Represented_def, clarsimp)
   apply (erule_tac x="ab!i" in allE, erule_tac x=i in allE, clarsimp) apply (erule_tac x=i in allE, clarsimp)
apply (erule disjE, simp add: Store_typecorrect_def StoreValues_Represented_def)
  apply (drule AL_codD, clarsimp)
   apply (erule_tac x=ad in allE, erule_tac x=c in allE, clarsimp) 
apply (erule disjE, simp add: Heap_typecorrect_def AddrValues_Represented_def)
  apply (erule_tac x=c in allE, clarsimp)
  apply (simp add: HeapAbstractsAt_def)
  apply (drule AL_domD, clarsimp)
  apply (erule_tac x="RVal (Loc y)" in allE, simp)
apply (simp add: AbsFields_expand  FieldValues_Represented_def Heap_typecorrect_def, clarsimp)
  apply (erule_tac x=d in allE, erule impE, erule AL_domI) apply clarsimp
  apply (simp add: HeapAbstractsAt_def Fields_typecorrect_def, clarsimp)
  apply (erule_tac x=f in allE, clarsimp)
  apply (erule_tac x=c in allE,simp)
done

constdefs uniqueValues::"STATE \<Rightarrow> State \<Rightarrow> AddrMapping \<Rightarrow> bool"
"uniqueValues \<Sigma> \<sigma> I \<equiv> \<forall> c . c:DOM \<Sigma> \<longrightarrow> (\<exists> v . ValuesRepresented c \<Sigma> \<sigma> I = {v})"

lemma uniqueValues_Tail: 
  "\<lbrakk>uniqueValues (c # S, \<Gamma>, H) (v # ops, s, h) I; 
    State_typecorrect N (S, \<Gamma>, H) (ops, s, h) I\<rbrakk>
  \<Longrightarrow> uniqueValues (S, \<Gamma>, H) (ops, s, h) I"
apply (simp add: uniqueValues_def) 
apply clarsimp
apply (drule_tac c=ca in State_typecorrect_VR_notEmpty) apply assumption
apply (subgoal_tac "\<exists>v. Values_Represented ca (S, \<Gamma>, H) (ops, s, h) I = {v}")
  apply (simp add: ValRepEQ)
apply (subgoal_tac "\<exists>v. v: Values_Represented ca (S, \<Gamma>, H) (ops, s, h) I")
prefer 2 apply fast
apply clarsimp
apply (rule_tac x=va in exI)
apply rule prefer 2 apply clarsimp
apply (erule_tac x=ca in allE, erule impE) apply (erule SetTriv) apply (rule DOM_cons)
apply clarsimp
apply (subgoal_tac "Values_Represented ca (c # S, \<Gamma>, H) (v # ops, s, h) I = {vaa}")
prefer 2 apply (simp add: ValRepEQ)
apply (rotate_tac -2, erule thin_rl)
apply (drule SetTriv, rule_tac c=c and v=v in Values_Represented_CONS)
apply (drule SetTriv, rule_tac c=c and v=v in  Values_Represented_CONS)
  apply fastsimp
done

constdefs StateSatisfiesAbstraction::"State \<Rightarrow> STATE \<Rightarrow> MP \<Rightarrow> AddrMapping \<Rightarrow> bool"
"StateSatisfiesAbstraction \<sigma> \<Sigma> N I \<equiv> 
  State_typecorrect N \<Sigma> \<sigma> I \<and>
  uniqueValues \<Sigma> \<sigma> I \<and>
  (case \<sigma> of (ops,s,h) \<Rightarrow> 
    (case \<Sigma> of (S,\<Gamma>,H) \<Rightarrow>
           length ops = length S \<and> 
           (DOM \<Sigma> \<inter> ObjIDs N) \<subseteq> dom I \<and> inj_on I (AL_dom H)))"

lemma StateSatis_inj1:
  "\<lbrakk>StateSatisfiesAbstraction (ops, s, h) (S, \<Gamma>, H) N I;  
    I c = I d; c \<in> AL_dom H; d \<in> AL_dom H; AL_dom H \<subseteq> ObjIDs N\<rbrakk>  \<Longrightarrow> c = d"
apply (simp add: StateSatisfiesAbstraction_def State_typecorrect_def Heap_typecorrect_def, clarsimp)
apply (drule inj_on_triv2) apply assumption(* apply (erule SetTriv, assumption) apply (erule SetTriv, assumption) apply assumption *)
apply assumption+
done

lemma StateSatisfiesAbstraction_MPextensional:
  "\<lbrakk>StateSatisfiesAbstraction \<sigma> \<Sigma> N I; \<forall> x c. (N\<down>x = Some c) = (NN\<down>x = Some c)\<rbrakk> \<Longrightarrow> StateSatisfiesAbstraction \<sigma> \<Sigma> NN I"
apply (simp add: StateSatisfiesAbstraction_def)
apply (case_tac \<sigma>, clarsimp)
apply (case_tac \<Sigma>, clarsimp)
apply (frule ObjIDs_extensional,clarsimp)
apply (erule State_typecorrect_MP_extensional) apply clarsimp
done

lemma StateSatisfiesAbstraction_HP_extensional:
  "\<lbrakk>StateSatisfiesAbstraction \<sigma> (S,\<Gamma>,H) N I;
    \<forall> x c. (H\<down>x = Some c) = (HH\<down>x = Some c)\<rbrakk> 
  \<Longrightarrow> StateSatisfiesAbstraction \<sigma> (S,\<Gamma>,HH) N I"
apply (case_tac \<sigma>, clarsimp)
apply (subgoal_tac "DOM (S, \<Gamma>, H) = DOM (S, \<Gamma>, HH)")
prefer 2 apply (rule DOM_HP_extensional) apply clarsimp
apply (simp add: StateSatisfiesAbstraction_def, clarsimp)
apply rule apply (erule State_typecorrect_HP_extensional) apply clarsimp
apply rule apply (simp add: uniqueValues_def) apply clarsimp
  apply (erule_tac x=c in allE, clarsimp) apply (rule_tac x=v in exI) 
      apply (subgoal_tac " ValuesRepresented c (S, \<Gamma>, H) (a, aa, b) I =  ValuesRepresented c (S, \<Gamma>, HH) (a, aa, b) I", clarsimp) 
       apply (rule ValuesRepresented_HPextensional) apply clarsimp
apply (simp (no_asm) add: inj_on_def, clarsimp) 
apply (drule inj_onD) apply assumption  apply (simp add: AL_dom_def)  apply (simp add: AL_dom_def) apply simp 
done

lemma StateSatisfiesAbstraction_freshInt:
  "\<lbrakk>StateSatisfiesAbstraction (ops,s,h) (S, \<Gamma>, H) N I; typecorrect v intTP;
    State_welldef (S, \<Gamma>, H) N; c \<notin> AL_dom N\<rbrakk>
  \<Longrightarrow> StateSatisfiesAbstraction (v # ops,s,h) (c # S, \<Gamma>, H) N[c\<mapsto>intTP] I"
apply (simp add: StateSatisfiesAbstraction_def, clarsimp)
apply (rule, simp add: State_typecorrect_def, clarsimp)
  apply (rule, erule Opstack_typecorrect_FreshInsertInOPS) apply assumption
    apply (simp add: State_welldef_def, clarsimp) apply fast
apply (rule, erule Store_typecorrect_FreshInsert) apply assumption
    apply (simp add: State_welldef_def, clarsimp) apply fast
apply (erule Heap_typecorrect_FreshInsert)
    apply (simp add: State_welldef_def, clarsimp) apply fast
    apply (simp add: State_welldef_def, clarsimp) apply fast
apply rule apply (simp add: uniqueValues_def) apply clarsimp
  apply (subgoal_tac "ValuesRepresented ca (c # S, \<Gamma>, H) (v # ops, s, h) I = ValuesRepresented ca ([c], emp, H) ([v], emp, h) I \<union> ValuesRepresented ca (S, \<Gamma>, H) (ops, s, h) I", clarsimp)
    prefer 2 apply (rule ValuesRepresented14)
  apply (rotate_tac -1, erule thin_rl)
  apply (frule_tac ops=ops and s=s and h=h and I=I in State_welldef_notinMP_ValuesRepresented_Empty) apply assumption
  apply (case_tac "c:DOM (S, \<Gamma>, H)")
    apply (drule SetTriv, erule State_welldef_DOM_ALdom) apply fast
  apply (case_tac "ca=c", clarsimp)
    apply (rotate_tac 4, erule thin_rl)
    apply (rule_tac x=v in exI)
    apply (subgoal_tac "Values_Represented c ([c], emp, H) ([v], emp, h) I = {v}")
    apply (simp add: ValRepEQ)
    apply (subgoal_tac "Values_Represented c (S, \<Gamma>, H) (ops, s, h) I = {}")
    prefer 2 apply (simp add: ValRepEQ)
    apply (simp add: Values_Represented_def) apply clarsimp
    apply (simp add: OpsValues_Represented_def)
    apply (simp add: StoreValues_Represented_def AL_emp1)
    apply (erule_tac x=ca in allE, erule impE, simp add: DOM_def) 
    apply clarsimp apply (subgoal_tac "Values_Represented ca (S, \<Gamma>, H) (ops, s, h) I = {va}")
                   apply (subgoal_tac "x \<in> Values_Represented ca ([c], emp, H) ([v], emp, h) I")
                   apply (simp add: Values_Represented_def)
                   apply (erule disjE) apply (simp add: OpsValues_Represented_def)
                   apply (erule disjE) apply (simp add: StoreValues_Represented_def AL_emp1)
                   apply (erule disjE) apply (simp add: AddrValues_Represented_def) apply clarsimp
                   apply (simp add: FieldValues_Represented_def) apply clarsimp apply fast
          apply (simp add: ValRepEQ)
          apply (simp add: ValRepEQ)
apply (rule SetTriv101) prefer 2 apply assumption
  apply (simp add: ObjIDs_def DOM_def)
  apply (rule, clarsimp) apply (drule AL_upd_split) apply clarsimp
  apply clarsimp apply (drule AL_upd_split) apply clarsimp
done

lemma StateSatisfiesAbstraction_MP_contained:
 "\<lbrakk> StateSatisfiesAbstraction \<sigma> \<Sigma> N I; contained N NN; State_welldef \<Sigma> N\<rbrakk> \<Longrightarrow> StateSatisfiesAbstraction \<sigma> \<Sigma> NN I"
apply (simp add: StateSatisfiesAbstraction_def, clarsimp)
apply (rule, erule State_typecorrect_MP_contained, assumption)
apply (case_tac \<sigma>, clarsimp)
apply (case_tac \<Sigma>, clarsimp)
apply (subgoal_tac "x:DOM (ab, ac, ba) \<inter> ObjIDs N", drule_tac x=x in SetTriv, assumption) apply clarsimp
apply clarsimp apply (drule SetTriv, erule State_welldef_DOM_ALdom) apply (simp add: ObjIDs_def AL_dom_def, clarsimp)
  apply (drule contained3, assumption) apply clarsimp
done

lemma StateSatisfiesAbstraction_Tail:
  "StateSatisfiesAbstraction (v # ops, s, h) (c # S, \<Gamma>, H) N I
  \<Longrightarrow> StateSatisfiesAbstraction (ops, s, h) (S, \<Gamma>, H) N I"
apply (simp add: StateSatisfiesAbstraction_def, clarsimp)
apply (frule State_typecorrect_Tail)
apply rule apply assumption
apply (rule, erule uniqueValues_Tail) apply assumption
apply (rule SetTriv101) prefer 2 apply assumption apply (insert DOM_cons) apply fast
done

(*The following lemma is use for rules NewNew and New, and therfore included in this file*)
lemma StateSatisfiesAbstraction_NewNew:
  "\<lbrakk> StateSatisfiesAbstraction (ops, s, h) (S, \<Gamma>, HH) NN I; 
     State_welldef (S, \<Gamma>, HH) NN; h\<down>a = None;c \<notin> AL_dom NN\<rbrakk>
  \<Longrightarrow> StateSatisfiesAbstraction (RVal (Loc a) # ops, s, h[a\<mapsto>(C, emp)]) (c # S, \<Gamma>, HH[c\<mapsto>(C, emp)])
        NN[c\<mapsto>(classTP C)] (\<lambda>x. if x = c then Some a else (restrict_map I (AL_dom NN)) x)"
apply (simp add: StateSatisfiesAbstraction_def, clarsimp)
apply (rule, simp add: State_typecorrect_def, clarsimp)
  apply rule apply (erule Opstack_typecorrect_FreshInsertInOPS) apply simp
    apply (simp add: State_welldef_def, clarsimp) apply fast
  apply rule apply (erule_tac v="RVal (Loc a)" in Store_typecorrect_FreshInsert) apply simp
    apply (simp add: State_welldef_def, clarsimp) apply fast
  apply (simp add: Heap_typecorrect_def, clarsimp)
    apply (frule AL_domD, clarsimp)
    apply (drule AL_upd_split)
    apply (erule disjE, clarsimp)
      apply (simp add: AL_update1)
      apply (rotate_tac -2, erule thin_rl, erule thin_rl)
      apply (rule, simp add: HeapAbstractsAt_def AL_update1) 
        apply rule apply (simp add : subclass_of_refl)
        apply (rule Fields_tc_emp_emp)
      apply (rule domI) apply simp
    apply clarsimp apply (erule_tac x=ca in allE, erule impE, erule AL_domI)
      apply clarsimp apply (frule_tac H=NN in AL_domI)
      apply (rule_tac x=Ca in exI)
      apply (rule, erule AL_update5) apply assumption
      apply (rule, simp add: HeapAbstractsAt_def, clarsimp) 
        apply (rule, clarsimp)
        apply clarsimp apply (drule AL_update3a) apply assumption
          apply clarsimp 
          apply (rule_tac x=E in exI, clarsimp)
          apply (rule_tac x=G in exI)
          apply rule apply (rule AL_update5, assumption) 
              apply (case_tac "a = y", clarsimp, assumption)
          apply (simp add: Fields_typecorrect_def, clarsimp)
            apply (erule_tac x=f in allE, clarsimp)
            apply (rule_tac x=tp in exI, clarsimp)
(*            apply (rule_tac x=pa in exI*) apply (rule AL_update5, assumption)
            apply (drule_tac d=caa in AL_domI) apply fast
       apply (rule domI) apply (subgoal_tac "ca \<noteq> c", clarsimp) apply simp apply fast
apply (rule, simp add: uniqueValues_def, clarsimp)
  apply (subgoal_tac " ValuesRepresented ca (c # S, \<Gamma>, HH[c\<mapsto>(C, emp)]) (RVal (Loc a) # ops, s, h[a\<mapsto>(C, emp)])
                 (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) =  
          ValuesRepresented ca ([c], emp, HH[c\<mapsto>(C, emp)]) ([RVal (Loc a)], emp, h[a\<mapsto>(C, emp)])
                 (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) \<union>  
          ValuesRepresented ca (S, \<Gamma>, HH[c\<mapsto>(C, emp)]) (ops, s, h[a\<mapsto>(C, emp)])
                 (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x)", simp)
    prefer 2 apply (rule ValuesRepresented14)
  apply (rotate_tac -1, erule thin_rl)
  apply (case_tac "ca=c", clarsimp)
    apply (rotate_tac 4, erule thin_rl)
    apply (rule_tac x="RVal (Loc a)" in exI)
    apply (subgoal_tac "Values_Represented c ([c], emp, HH[c\<mapsto>(C, emp)]) ([RVal (Loc a)], emp, h[a\<mapsto>(C, emp)])
        (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) \<union>
       Values_Represented c (S, \<Gamma>, HH[c\<mapsto>(C, emp)]) (ops, s, h[a\<mapsto>(C, emp)]) (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) =
       {RVal (Loc a)}", simp add: ValRepEQ)
    apply (simp add: Values_Represented_def StoreValues_Represented_def AL_emp1)
    apply (rule, clarsimp)
      apply (rule, simp add: OpsValues_Represented_def)
      apply (rule, simp add: AddrValues_Represented_def) apply (simp add: AL_dom_upd1)
      apply (rule, simp add: FieldValues_Represented_def, clarsimp)
        apply (case_tac "d=c", clarsimp) apply (simp add: AL_update1 AL_emp1) 
        apply clarsimp apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1) apply clarsimp
          apply (frule State_welldef_Field_lookup) apply assumption apply assumption apply clarsimp
          apply (drule_tac d=c in AL_domI) apply simp
      apply (rule, simp add: OpsValues_Represented_def) apply clarsimp
        apply (drule State_welldef_OPS_lookup_dom) apply assumption  apply simp
      apply (rule, clarsimp) apply (drule State_welldef_Store_lookup_dom) apply assumption  apply simp
      apply (rule, simp add: AddrValues_Represented_def, clarsimp) 
      apply (simp add: FieldValues_Represented_def, clarsimp)
        apply (case_tac "d=c", clarsimp) apply (simp add: AL_update1 AL_emp1) 
        apply clarsimp apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1) apply clarsimp
          apply (frule State_welldef_Field_lookup_dom) apply assumption apply assumption apply clarsimp
    apply clarsimp 
      apply (simp add: OpsValues_Represented_def AddrValues_Represented_def FieldValues_Represented_def)
  apply (subgoal_tac "ca \<in> DOM (S, \<Gamma>, HH)")
  prefer 2 apply (simp add: DOM_def) apply (simp add: AL_dom_upd1 AbsFields_expand) apply clarsimp
           apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1) apply clarsimp
  apply (erule_tac x=ca in allE, clarsimp)
  apply (rule_tac x=v in exI)
    apply (subgoal_tac "Values_Represented ca (S, \<Gamma>, HH) (ops, s, h) I = {v}")
    prefer 2 apply (simp add: ValRepEQ)
    apply (rotate_tac -2, erule thin_rl)
    apply (subgoal_tac "Values_Represented ca ([c], emp, HH[c\<mapsto>(C, emp)]) ([RVal (Loc a)], emp, h[a\<mapsto>(C, emp)])
               (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) \<union>
              Values_Represented ca (S, \<Gamma>, HH[c\<mapsto>(C, emp)]) (ops, s, h[a\<mapsto>(C, emp)]) (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) =
              {v}", simp add: ValRepEQ)
    apply (simp add: Values_Represented_def StoreValues_Represented_def AL_emp1)
    apply (rule, clarsimp)
      apply (rule, simp add: OpsValues_Represented_def, clarsimp)
      apply (rule, simp add: AddrValues_Represented_def, clarsimp) apply (simp add: AL_dom_upd1) apply (drule_tac d=ca in AL_domD, clarsimp)
          apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp
      apply (rule, simp add: FieldValues_Represented_def, clarsimp) 
        apply (case_tac "d=c", clarsimp) apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1)
            apply (simp add: AL_update1 AL_emp1) 
        apply clarsimp apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1) apply clarsimp 
                       apply (drule AL_upd_split) apply (erule disjE, simp add: AL_emp1) apply clarsimp 
                       apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp 
        apply (subgoal_tac "x: OpsValues_Represented ca S ops \<union> {v. \<exists>x. \<Gamma>\<down>x = Some ca \<and> s\<down>x = Some v} \<union> AddrValues_Represented ca HH I \<union>
                {v. \<exists>d D F. HH\<down>d = Some (D, F) \<and> (\<exists>f. F\<down>f = Some ca \<and> (\<exists>C G a. I d = Some a \<and> h\<down>a = Some (C, G) \<and> G\<down>f = Some v))}") 
          apply (drule_tac x=x in SetTriv104, assumption) apply clarsimp 
        apply (rule_tac X="{v. \<exists>d D F. HH\<down>d = Some (D, F) \<and> (\<exists>f. F\<down>f = Some ca \<and> (\<exists>C G a. I d = Some a \<and> h\<down>a = Some (C, G) \<and> G\<down>f = Some v))}" in SetTriv) 
          apply clarsimp apply (rule_tac x=d in exI, clarsimp) apply (rule_tac x=f in exI, clarsimp)  
        apply fast 
      apply (rule, fast)
      apply (rule, fast)
      apply (rule, rule_tac Y="AddrValues_Represented ca HH I" in SetTriv101) 
        apply (simp add: AddrValues_Represented_def AL_dom_upd1) apply clarsimp
          apply (drule_tac d=ca in AL_domD, clarsimp) 
          apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp
        apply fast
      apply (rule_tac Y="FieldValues_Represented ca HH I h" in SetTriv101)
        apply (simp add: FieldValues_Represented_def, clarsimp)
          apply (case_tac "d=c", clarsimp)
            apply (simp add: AL_update1 AL_emp1)
          apply clarsimp apply (drule AL_upd_split) apply (erule disjE) apply (simp add: AL_emp1) apply clarsimp
            apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp
            apply (drule AL_upd_split) apply (erule disjE) apply (simp add: AL_emp1) apply clarsimp
            apply (rule_tac x=d in exI, clarsimp) apply (rule_tac x=f in exI, clarsimp)  
        apply fast 
      apply clarsimp 
         apply (subgoal_tac "v:OpsValues_Represented ca S ops \<union> {v. \<exists>x. \<Gamma>\<down>x = Some ca \<and> s\<down>x = Some v} \<union> AddrValues_Represented ca HH I \<union> FieldValues_Represented ca HH I h") prefer 2 apply fast  
         apply (erule thin_rl, clarsimp)
         apply (erule disjE) apply clarsimp 
         apply (erule disjE) apply (drule_tac x=v and Y="AddrValues_Represented ca HH[c\<mapsto>(C, emp)] (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x)" in SetTriv) 
                             apply (simp add: AddrValues_Represented_def AL_dom_upd1, clarsimp) apply (drule_tac d=ca in AL_domD, clarsimp)
                               apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp 
                           apply simp
         apply (drule_tac x=v and Y="FieldValues_Represented ca HH[c\<mapsto>(C, emp)] (\<lambda>x. if x = c then Some a else (I |` AL_dom NN) x) h[a\<mapsto>(C, emp)]" in SetTriv)
           apply (simp add: FieldValues_Represented_def, clarsimp)
             apply (rule_tac x=d in exI, rule)
               apply clarsimp apply (erule_tac x=c in allE, simp add: AL_update1) 
                               apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp 
               apply clarsimp apply (rule, rule, rule, erule AL_update5a, assumption)
                 apply (rule, rule, assumption)
                 apply (frule State_welldef_Heap_lookup_dom) apply assumption apply clarsimp 
                 apply (rule, rule, rule, erule AL_update5a) apply (simp add: State_typecorrect_def Heap_typecorrect_def, clarsimp)
                   apply (erule_tac x=d in allE, erule impE, erule AL_domI) apply (simp add: HeapAbstractsAt_def) apply clarsimp
                     apply (erule_tac x=F in allE, clarsimp)
                 apply assumption
          apply simp
apply rule 
    apply (rule, clarsimp)
    apply (simp add: DOM_def ObjIDs_insertObj)
    apply (rule_tac m="(I |` AL_dom NN)" in domD)  
    apply (simp add: DOM_def ObjIDs_insertObj)
      apply (subgoal_tac "x:AL_dom NN", simp) 
        apply (rule SetTriv) prefer 2 apply assumption 
        apply clarsimp apply (simp add: AL_dom_upd1) apply clarsimp
        apply (simp add: AbsFields_expand, clarsimp) apply (drule AL_upd_split, clarsimp) 
                                                   apply (erule disjE, clarsimp) apply (simp add: AL_emp1)
                                                   apply clarsimp
      apply (erule SetTriv) apply (rule ObjIDs_in_Dom) 
  apply (subgoal_tac "c \<notin> ObjIDs NN")
  prefer 2 apply clarsimp apply (drule SetTriv) apply (rule ObjIDs_in_Dom) apply simp
    apply (simp (no_asm) add: inj_on_def, clarsimp)
      apply (rule, clarsimp) 
        apply (subgoal_tac "y :AL_dom HH") prefer 2 apply (simp add: AL_dom_upd1)
        apply (simp add:State_typecorrect_def Heap_typecorrect_expand, clarsimp)
        apply (drule_tac H=HH in AL_domD, clarsimp)
        apply (erule_tac x=y in allE, clarsimp) 
        apply (subgoal_tac "y :AL_dom NN", clarsimp) apply (erule AL_domI)
    apply clarsimp
    apply (rule, clarsimp) apply (simp add: AL_dom_upd1)
        apply (simp add:State_typecorrect_def Heap_typecorrect_expand, clarsimp)
        apply (drule_tac H=HH in AL_domD, clarsimp)
        apply (erule_tac x=x in allE, clarsimp) 
        apply (subgoal_tac "x :AL_dom NN", clarsimp) apply (erule AL_domI)
    apply clarsimp apply (simp add: AL_dom_upd1)
        apply (subgoal_tac "x :AL_dom NN", clarsimp) 
        apply (subgoal_tac "y :AL_dom NN", clarsimp) 
          apply (erule inj_onD, simp) apply assumption apply assumption
        apply (erule State_welldef_lookupHeap1) apply assumption
      apply (erule State_welldef_lookupHeap1) apply assumption
done

lemma StateSatisfiesAbstraction_HD:
  "StateSatisfiesAbstraction (v # os, s, h) (c # S, \<Gamma>, H) N I
  \<Longrightarrow> StateSatisfiesAbstraction ([v], emp, h) ([c], emp, H) N I"
apply (simp add: StateSatisfiesAbstraction_def)
apply (subgoal_tac "State_typecorrect N ([c], emp, H) ([v], emp, h) I", rule, assumption)
prefer 2 apply (simp add: State_typecorrect_def, clarsimp)
         apply (rule, simp add: Opstack_typecorrect_def)
           apply (erule_tac x=0 in allE, clarsimp)
         apply (simp add: Store_typecorrect_def AL_emp1)
apply (rule, simp add: uniqueValues_def, clarsimp)
  apply (erule_tac x=ca in allE, erule impE, simp add: DOM_def AL_cod_emp) apply fast apply (erule exE)
  apply (subgoal_tac "\<exists> w . w:ValuesRepresented ca ([c], emp, H) ([v], emp, h) I", erule exE)
  prefer 2 apply (subgoal_tac "ValuesRepresented ca ([c], emp, H) ([v], emp, h) I \<noteq> {}", fast)
           apply (subgoal_tac "Values_Represented ca ([c], emp, H) ([v], emp, h) I \<noteq> {}", simp add: ValRepEQ)
           apply (erule State_typecorrect_VR_notEmpty) apply assumption
    apply (rule_tac x=w in exI, rule) prefer 2 apply clarsimp 
    apply (subgoal_tac "w=va", clarsimp)
      apply (drule_tac x=x in SetTriv) apply (rule SetTriv101) apply (rule_tac \<Gamma>=\<Gamma> and s=s in ValuesRepresented13) 
                                                               apply (rule_tac S=S and ops=os in ValuesRepresented_HD) apply clarsimp
      apply (drule_tac x=w in SetTriv) apply (rule SetTriv101) apply (rule_tac \<Gamma>=\<Gamma> and s=s in ValuesRepresented13) 
                                                               apply (rule_tac S=S and ops=os in ValuesRepresented_HD) apply clarsimp 
apply clarsimp apply (rule_tac m=I in domD)
  apply (rule SetTriv) prefer 2 apply assumption apply (simp add: DOM_def AL_cod_emp) apply fastsimp
done

constdefs unisat::"State \<Rightarrow> STATE \<Rightarrow> MP \<Rightarrow> AddrMapping \<Rightarrow> bool"
"unisat \<sigma> \<Sigma> N I \<equiv>
  State_welldef \<Sigma> N \<and> StateSatisfiesAbstraction \<sigma> \<Sigma> N I"

lemma unisat_MPextensional:
"\<lbrakk>unisat \<sigma> \<Sigma> N I; \<forall> x c . (N\<down>x = Some c) = (NN\<down>x = Some c); NN \<in> AL\<rbrakk> \<Longrightarrow>
 unisat \<sigma> \<Sigma> NN I"
apply (simp add: unisat_def, clarsimp)
apply (rule, erule State_welldef_MPextensional) apply simp apply assumption
apply (erule StateSatisfiesAbstraction_MPextensional) apply simp
done

lemma unisat_HPextensional:
 "\<lbrakk> unisat \<sigma> (S,\<Gamma>,H) N I; 
    \<forall> x c . (H\<down>x = Some c) = (HH\<down>x = Some c); HH \<in> AL\<rbrakk> 
 \<Longrightarrow> unisat \<sigma> (S,\<Gamma>,HH) N I"
apply (simp add: unisat_def, clarsimp)
apply (rule, erule State_welldef_HPextensional) apply simp apply assumption
apply (case_tac \<sigma>, clarsimp)
apply (erule StateSatisfiesAbstraction_HP_extensional) apply simp
done

lemma unisat_MP_contained: "\<lbrakk>unisat \<sigma> \<Sigma> N I; contained N NN; NN \<in> AL\<rbrakk> \<Longrightarrow> unisat \<sigma> \<Sigma> NN I"
apply (simp add: unisat_def, clarsimp)
apply rule apply (case_tac \<Sigma>, clarsimp) apply (erule State_welldef_MP_contained) apply assumption+
apply (erule StateSatisfiesAbstraction_MP_contained, assumption+)
done

lemma unisat_Tail:"unisat (v # ops, s, h) (c#S, \<Gamma>, H) N I
       \<Longrightarrow> unisat (ops, s, h) (S, \<Gamma>, H) N I"
apply (unfold unisat_def, clarsimp)
apply (rule, erule State_welldef_Tail)
apply (erule StateSatisfiesAbstraction_Tail)
done

(*
(-The class IDs for which a concrete operand stack contains a reference value-)
constdefs OpstackRefIDs::"Colour list \<Rightarrow> OpStack \<Rightarrow> Colour set"
"OpstackRefIDs S ops == { c . \<exists> n r . n < length S \<and> c = S!n \<and> ops!n = RVal r}"

(-The class IDs for which a concrete store contains a reference value-)
constdefs StoreRefIDs::"VarContext \<Rightarrow> Store \<Rightarrow> Colour set"
"StoreRefIDs \<Gamma> s == { c . \<exists> x r . s\<down>x = Some(RVal r) \<and> \<Gamma>\<down>x = Some c}"

(-The class IDs for which a concrete object field contains a reference value-)
constdefs FieldRefIDs::"HP \<Rightarrow> Heap \<Rightarrow> AddrMapping \<Rightarrow> Colour set"
"FieldRefIDs H h I == { d . \<exists> c a C F f r . I c = Some a \<and> h\<down>a = Some(C,F) \<and> F\<down>f = Some(RVal r) \<and>
                                          d : AbsFieldsAt H c f}"
*)

end